/*
 * Copyright 2020 - 2021 IoT.bzh
 *
 * author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
#define _GNU_SOURCE

#ifndef CONTROL_CONFIG_PATH
    #define CONTROL_CONFIG_PATH "@CONTROL_CONFIG_PATH@"
#endif

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "badge-reader-binding.h"
#include <ctl-config.h>
#include <filescan-utils.h>

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            GLOBAL VARIABLES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

char *events_badges[MAX_EVENT_LIST] = {GOOD_BADGE_EVENT, BAD_BADGE_EVENT};

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PROTOTYPE FUNCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

static int FamilyReaderConfig(afb_api_t api, CtlSectionT *section, json_object *fReadersJ);
static int DatabaseConfig(afb_api_t api, CtlSectionT *section, json_object *databaseJ);
static int CommandConfig(afb_api_t api, CtlSectionT *section, json_object *commandsJ);
static void ping_verb_cb (afb_req_t request);
static void info_verb_cb (afb_req_t request);
static void list_verb_cb (afb_req_t request);
static void subscribe_verb_cb (afb_req_t request);
static void unsubscribe_verb_cb (afb_req_t request);
static void sqldb_add_badge (afb_req_t request);
static void sqldb_get_badge (afb_req_t request);
static void sqldb_update_badge (afb_req_t request);
static void sqldb_list_badge (afb_req_t request);
static void sqldb_delete_badge (afb_req_t request);

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            VARIABLES GLOBAL
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * ctrlSections - Array of callbacks linked to config file sections
 */
static CtlSectionT ctrlSections[] = {
    { .key = "plugins",         .loadCB = PluginConfig, .handle= fReaderPointerFunctionsRegister},
    { .key = "onload",          .loadCB = OnloadConfig },
    { .key = "database",        .loadCB = DatabaseConfig },
    { .key = "commands",        .loadCB = CommandConfig },
    { .key = "family-reader",   .loadCB = FamilyReaderConfig },
    { .key = NULL } /* marker for end of the array */
};

/**
 * CtrlApiVerb - Array of callbacks linked to verbs
 */
static afb_verb_t CtrlApiVerbs[] = {
    /* VERB'S NAME              FUNCTION TO CALL                    SHORT DESCRIPTION */
    { .verb = "ping",           .callback = ping_verb_cb,           .info = "badge-reader API ping test"},
    { .verb = "info",           .callback = info_verb_cb,           .info = "Gives info of Badge Reader Binding"},
    { .verb = "list",           .callback = list_verb_cb,           .info = "List all Badge Reader"},
    { .verb = "subscribe",      .callback = subscribe_verb_cb,      .info = "Subscribe to all Badge Reader event"},
    { .verb = "unsubscribe",    .callback = unsubscribe_verb_cb,    .info = "Unsubscribe from all Badge Reader event"},
    { .verb = "add_badge",      .callback = sqldb_add_badge,        .info = "Add a badge into database"},
    { .verb = "get_badge",      .callback = sqldb_get_badge,        .info = "Get a badge from the database"},
    { .verb = "update_badge",   .callback = sqldb_update_badge,     .info = "Update a badge into database"},
    { .verb = "list_badge",     .callback = sqldb_list_badge,       .info = "List all badges from database"},
    { .verb = "delete_badge",   .callback = sqldb_delete_badge,     .info = "Delete a badge from database"},
    { .verb = NULL} /* marker for end of the array */
};

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            API VERBS FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * ping_verb_cb - Callback of the ping verb
 * 
 * @param request afb request
 */
static void ping_verb_cb (afb_req_t request) {
    static int count=0;
    char response[32];
    json_object *queryJ =  afb_req_json(request);

    snprintf (response, sizeof(response), "Pong=%d", count++);
    AFB_API_NOTICE(request->api, "badge-reader:ping count=%d query=%s", count, json_object_get_string(queryJ));
    afb_req_success_f(request,json_object_new_string(response), NULL);

    return;
}

/**
 * info_verb_cb - Callback of the info verb
 * 
 * @param request afb request
 */
static void info_verb_cb (afb_req_t request) {
    int err = 0;
    int index = 0;
    int verbose = 0;
    family_reader_t * fReader = (family_reader_t *) afb_req_get_vcbdata(request);
    json_object * queryJ =  afb_req_json(request);
    json_object * responseJ = NULL;

    if (json_object_is_type(queryJ, json_type_object)) {
        err = wrap_json_unpack(queryJ, "{s?i !}", "verbose", &verbose);
        if (err) goto OnJsonUnwrapError;
    }

    json_object * groupJ = json_object_new_array();

    /* GENERAL group */

    json_object *generalJ = NULL;
    json_object *general_verbsJ = json_object_new_array();

    // Ping verb
    json_object *general_pingJ = NULL;
    err += wrap_json_pack(&general_pingJ, "{ss ss ss}"
            , "uid", "ping"
            , "info", "Ping the binding"
            , "verb", "ping");
    json_object_array_add(general_verbsJ, general_pingJ);

    // Info verb
    json_object *general_infoJ = NULL;
    err += wrap_json_pack(&general_infoJ, "{ss ss ss}"
            , "uid", "info"
            , "info", "Info of the binding"
            , "verb", "info");
    json_object_array_add(general_verbsJ, general_infoJ);

    // List verb
    json_object *general_listJ = NULL;
    err += wrap_json_pack(&general_listJ, "{ss ss ss}"
            , "uid", "list"
            , "info", "List all readers"
            , "verb", "list");
    json_object_array_add(general_verbsJ, general_listJ);

    // Subscribe verb
    json_object *general_subJ = NULL;
    err += wrap_json_pack(&general_subJ, "{ss ss ss}"
            , "uid", "subscribe"
            , "info", "Subscribe to reader event"
            , "verb", "subscribe");
    json_object_array_add(general_verbsJ, general_subJ);

    // Unsubscribe verb
    json_object *general_unsubJ = NULL;
    err += wrap_json_pack(&general_unsubJ, "{ss ss ss}"
            , "uid", "unsubscribe"
            , "info", "Subscribe to reader event"
            , "verb", "unsubscribe");
    json_object_array_add(general_verbsJ, general_unsubJ);

    err += wrap_json_pack(&generalJ, "{ss ss* so}",
                "uid", "general",
                "info", "General command to send to the badge-reader binding",
                "verbs", general_verbsJ);
    json_object_array_add(groupJ, generalJ);

    /* SQL Group */

    json_object * sqldbJ = NULL;
    json_object *sql_verbsJ = json_object_new_array();

    // sql adding verb
    json_object *sql_addJ = NULL;
    json_object *sql_samples_addJ = json_object_new_array();
    json_object *sql_sample_addJ = NULL;
    err += wrap_json_pack(&sql_sample_addJ, "{ss ss}"
            , "key", "11223344"
            , "name", "username");
    json_object_array_add(sql_samples_addJ,sql_sample_addJ);
    err += wrap_json_pack(&sql_addJ, "{ss ss ss so*}"
            , "uid", "add_badge"
            , "info", "Adding badge into database"
            , "verb", "add_badge"
            , "sample", sql_samples_addJ);
    json_object_array_add(sql_verbsJ, sql_addJ);

    // sql getting verb
    json_object *sql_getJ = NULL;
    json_object *sql_samples_getJ = json_object_new_array();
    json_object *sql_sample_getJ = NULL;
    err += wrap_json_pack(&sql_sample_getJ, "{ss}"
            , "key", "11223344");
    json_object_array_add(sql_samples_getJ,sql_sample_getJ);
    err += wrap_json_pack(&sql_getJ, "{ss ss ss so*}"
            , "uid", "get_badge"
            , "info", "Get the badge from database"
            , "verb", "get_badge"
            , "sample", sql_samples_getJ);
    json_object_array_add(sql_verbsJ, sql_getJ);

    // sql list verb
    json_object *sql_listJ = NULL;
    err += wrap_json_pack(&sql_listJ, "{ss ss ss}"
            , "uid", "list_badge"
            , "info", "List badge from database"
            , "verb", "list_badge");
    json_object_array_add(sql_verbsJ, sql_listJ);

    // sql update verb
    json_object *sql_updateJ = NULL;
    json_object *sql_samples_updateJ = json_object_new_array();
    json_object *sql_sample_updateJ = NULL;
    err += wrap_json_pack(&sql_sample_updateJ, "{ss ss}"
            , "key", "11223344"
            , "control", "000000");
    json_object_array_add(sql_samples_updateJ,sql_sample_updateJ);
    err += wrap_json_pack(&sql_updateJ, "{ss ss ss so*}"
            , "uid", "update_badge"
            , "info", "update badge into database"
            , "verb", "update_badge"
            , "sample", sql_samples_updateJ);
    json_object_array_add(sql_verbsJ, sql_updateJ);

    // sql delete verb
    json_object *sql_deleteJ = NULL;
    json_object *sql_samples_deleteJ = json_object_new_array();
    json_object *sql_sample_deleteJ = NULL;
    err += wrap_json_pack(&sql_sample_deleteJ, "{ss}"
            , "key", "11223344");
    json_object_array_add(sql_samples_deleteJ,sql_sample_deleteJ);
    err += wrap_json_pack(&sql_deleteJ, "{ss ss ss so*}"
            , "uid", "delete_badge"
            , "info", "delete badge into database"
            , "verb", "delete_badge"
            , "sample", sql_samples_deleteJ);
    json_object_array_add(sql_verbsJ, sql_deleteJ);

    err += wrap_json_pack(&sqldbJ, "{ss ss* so}",
                    "uid", "database",
                    "info", "Administrate sqlite database",
                    "verbs", sql_verbsJ);

    json_object_array_add(groupJ, sqldbJ);

    // build global info page for developper dynamic HTML5 page
    json_object * globalJ, * fReaderJ, * readersJ;
    CtlConfigT * ctlConfig = (CtlConfigT *) afb_api_get_userdata(afb_req_get_api(request));
    err = wrap_json_pack (&globalJ, "{ss ss* ss* ss*}",
            "uid", ctlConfig->api,
            "info", ctlConfig->info,
            "version", ctlConfig->version,
            "author", ctlConfig->author);
    if (err) goto OnJsonPackFamilyError; 

    for (index = 0; fReader[index].uid; index++) {
        // Prepare array to hold all Reader verbs
        readersJ = json_object_new_array();
        // Create group object with infos and readers list
        FamilyReaderReadersId(&fReader[index], 3, readersJ);
        err += wrap_json_pack(&fReaderJ, "{ss ss* so}",
                "uid", fReader[index].uid,
                "info", fReader[index].info,
                "verbs", readersJ);
        if (err) goto OnJsonPackReadersError;
        json_object_array_add(groupJ, fReaderJ);
    }

    err = wrap_json_pack(&responseJ, "{so so}",
            "metadata", globalJ,
            "groups", groupJ);
    if (err) goto OnJsonPackGlobalError;

    afb_req_success(request, responseJ, NULL);
    
    return;

OnJsonUnwrapError:
    afb_req_fail_f(request, "BadgeReaderAdmin", "InfoReader: invalid 'json query' query=%s", json_object_get_string(queryJ));
    goto OnErrorExit;

OnJsonPackFamilyError:
    AFB_DEBUG("InfoReader: Failed to wrap Json binding metaddata");
    goto OnErrorExit;

OnJsonPackReadersError:
    AFB_DEBUG ("InfoReader: Fail to wrap json readers info FamilyReader=%s", fReader[index].uid);
    goto OnErrorExit;

OnJsonPackGlobalError:
    AFB_DEBUG("InfoReader: Failed to wrap Json Global response");
    goto OnErrorExit;

OnErrorExit:
    return;
}

/**
 * subscribe_verb_cb - Callback of the subscribe verb
 * 
 * @param request afb request
 */
static void subscribe_verb_cb (afb_req_t request) {
    afb_api_t api = afb_req_get_api(request);
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(api);
    user_data_t * user_data = (user_data_t *) ctlConfig->external;
    char response[32];
    int ret = 0;
    json_object *queryJ =  afb_req_json(request);

    // For each possible events:
    for (int index = 0; index < MAX_EVENT_LIST; index ++) {
        // Create event if not already created
        if (!afb_event_is_valid(user_data->event[index])) {
            user_data->event[index] = afb_api_make_event(api, events_badges[index]);
            if (!user_data->event[index]) {
                AFB_API_ERROR(api, "[subscribe_verb_cb] fail to create the global event: %s", events_badges[index]);
                afb_req_fail_f(request, "query-error", "Failed create event for subscription");
                return;
            }
        }
        // Subscribe to the global event
        ret = afb_req_subscribe(request, user_data->event[index]);
        if (ret < 0) {
            AFB_API_ERROR(api, "[subscribe_verb_cb] fail to subscribe to the event %s", events_badges[index]);
            afb_req_fail_f(request, "query-error", "Failed subscribe to event");
            return;
        }
    }

    snprintf (response, sizeof(response), "Subscribed to all event");
    AFB_API_NOTICE(request->api, "badge-reader:subscribe query=%s", json_object_get_string(queryJ));
    afb_req_success_f(request,json_object_new_string(response), NULL);

    return;
}

/**
 * unsubscribe_verb_cb - Callback of the unsubscribe verb
 * 
 * @param request afb request
 */
static void unsubscribe_verb_cb (afb_req_t request) {
    afb_api_t api = afb_req_get_api(request);
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(api);
    user_data_t * user_data = (user_data_t *) ctlConfig->external;
    char response[32];
    int ret = 0;
    json_object *queryJ =  afb_req_json(request);

    // Unsubscribe to the global event
    for (int index = 0; index < MAX_EVENT_LIST; index ++) {
        if (afb_event_is_valid(user_data->event[index])) {
            ret = afb_req_unsubscribe(request, user_data->event[index]);
            if (ret < 0) {
                AFB_API_ERROR(api, "[unsubscribe_verb_cb] fail to subscribe to global event ");
                afb_req_fail_f(request, "query-error", "Failed subscribe to event");
                return;
            }
        }
    }

    snprintf (response, sizeof(response), "Unsubscribed to all event");
    AFB_API_NOTICE(api, "badge-reader:unsubscribe query=%s", json_object_get_string(queryJ));
    afb_req_success_f(request,json_object_new_string(response), NULL);

    return;
}

/**
 * sqldb_add_badge - Callback of the add badge verb
 * 
 * @param request afb request
 */
static void sqldb_add_badge (afb_req_t request) {
    char response[64];
    int ret = 0;
    char * key = NULL;
    char * name = NULL;
    char * errMsg = NULL;
    json_object *queryJ =  afb_req_json(request);
    
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(afb_req_get_api(request));
    user_data_t * user_data = (user_data_t *) ctlConfig->external;
    sqldbData_t *sql_data = user_data->sql_data;

    if (json_object_get_type(queryJ) != json_type_object) {
        AFB_API_ERROR(afb_req_get_api(request), "[sqldb_add_badge] Query is in wrong type reader");
        afb_req_fail_f(request, "query-error", "sqldb_add_badge: invalid type of the query");
        return;
    }
    ret= wrap_json_unpack(queryJ, "{ss ss}", "key", &key, "name", &name);
    if (ret < 0) {
        afb_req_fail_f(request, "query-error", "sqldb_add_badge: invalid 'json' need to be {'key'='keyID, 'name='userName'}");
        return;
    }

    // Get the time now
    // Date variables
    time_t t = time(NULL);
    struct tm tm = *localtime(&t);
    char * date = NULL;

    // Get the actual date
    ret = asprintf(&date, "%d%02d%02d", tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday);
    if (ret < 0) {
        afb_req_fail_f(request, "query-error", "sqldb_add_badge: error to getting data");
        AFB_API_ERROR(afb_req_get_api(request), "[sqldb_add_badge] error to getting date !");
        return;
    }

    ret = sqldb_add(sql_data, key, date, name, &errMsg);
    if (ret < 0) {
        if (ret == -SQLDB_ERROR_ADDING) {
            AFB_API_ERROR(afb_req_get_api(request), "[sqldb_add_badge] Get error[%i] to add badge[%s]: %s", ret, key, errMsg);
            afb_req_fail_f(request, "query-error", "sqldb_add_badge: Badge[%s] already exists", key);
        } else {
            AFB_API_ERROR(afb_req_get_api(request), "[sqldb_add_badge] Get error[%i] to add badge[%s]: %s", ret, key, errMsg);
            afb_req_fail_f(request, "query-error", "sqldb_add_badge: Failed to add badge[%s]: %s", key, errMsg);
        }
        free(errMsg);
        free(date);
        return ;
    }
    free(date);
    snprintf (response, sizeof(response), "Badge=%s has been added to database", key);
    AFB_API_NOTICE(request->api, "badge-reader: query=%s", json_object_get_string(queryJ));
    afb_req_success_f(request,json_object_new_string(response), NULL);
    return;
}

/**
 * sqldb_get_badge - Callback of the get badge verb
 * 
 * @param request afb request
 */
static void sqldb_get_badge (afb_req_t request) {
    char response[64];
    int ret = 0;
    char * key = NULL;
    char * keyID = NULL;
    char * control = NULL;
    char * name = NULL;
    char * errMsg = NULL;
    json_object *queryJ =  afb_req_json(request);
    
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(afb_req_get_api(request));
    user_data_t * user_data = (user_data_t *) ctlConfig->external;
    sqldbData_t *sql_data = user_data->sql_data;

    if (json_object_get_type(queryJ) != json_type_object) {
        AFB_API_ERROR(afb_req_get_api(request), "[sqldb_get_badge] Query is in wrong type reader");
        afb_req_fail_f(request, "query-error", "sqldb_get_badge: invalid type of the query");
        return;
    }
    ret= wrap_json_unpack(queryJ, "{ss}", "key", &key);
    if (ret < 0) {
        afb_req_fail_f(request, "query-error", "sqldb_get_badge: invalid 'json' need to be {'key'='keyID'}");
        return;
    }

    ret = sqldb_get(sql_data, key, &keyID, &control, &name, &errMsg);
    if (ret < 0) {
        if (ret == -SQLDB_ERROR_GET) {
            AFB_API_ERROR(afb_req_get_api(request), "[sqldb_get_badge] Get error[%i] to get badge[%s]: %s", ret, key, errMsg);
            afb_req_fail_f(request, "query-error", "sqldb_get_badge: Badge[%s] hasn't been found", key);
        } else {
            AFB_API_ERROR(afb_req_get_api(request), "[sqldb_get_badge] Get error[%i] to get badge[%s]: %s", ret, key, errMsg);
            afb_req_fail_f(request, "query-error", "sqldb_get_badge: Failed to get badge[%s]: %s", key, errMsg);
        }
        free(errMsg);
        return ;
    }

    snprintf (response, sizeof(response), "Keys [%s] : name = %s, control =%s!", key, name, control);
    AFB_API_NOTICE(request->api, "badge-reader:ping query=%s", json_object_get_string(queryJ));
    afb_req_success_f(request,json_object_new_string(response), NULL);

    return;
}

/**
 * sqldb_update_badge - Callback of the update badge verb
 * 
 * @param request afb request
 */
static void sqldb_update_badge (afb_req_t request) {
    char response[64];
    int ret = 0;
    char * key = NULL;
    char * control = NULL;
    char * errMsg = NULL;
    json_object *queryJ =  afb_req_json(request);
    
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(afb_req_get_api(request));
    user_data_t * user_data = (user_data_t *) ctlConfig->external;
    sqldbData_t *sql_data = user_data->sql_data;

    if (json_object_get_type(queryJ) != json_type_object) {
        AFB_API_ERROR(afb_req_get_api(request), "[sqldb_update_badge] Query is in wrong type reader");
        afb_req_fail_f(request, "query-error", "sqldb_update_badge: invalid type of the query");
        return;
    }
    ret= wrap_json_unpack(queryJ, "{ss ss}", "key", &key, "control", &control);
    if (ret < 0) {
        afb_req_fail_f(request, "query-error", "sqldb_update_badge: invalid 'json' need to be {'key': 'keyID', 'control': '210811'}");
        return;
    }

    ret = sqldb_update_control(sql_data, key, control, &errMsg);
    if (ret < 0) {
        if (ret == -SQLDB_ERROR_UPDATE) {
            AFB_API_ERROR(afb_req_get_api(request), "[sqldb_update_badge] Get error[%i] to get badge[%s]: %s", ret, key, errMsg);
            afb_req_fail_f(request, "query-error", "sqldb_update_badge: Badge[%s] hasn't been found", key);
        } else {
            AFB_API_ERROR(afb_req_get_api(request), "[sqldb_update_badge] Get error[%i] to get badge[%s]: %s", ret, key, errMsg);
            afb_req_fail_f(request, "query-error", "sqldb_update_badge: Failed to get badge[%s]: %s", key, errMsg);
        }
        free(errMsg);
        return ;
    }

    snprintf (response, sizeof(response), "Keys [%s] :has been update!", key);
    AFB_API_NOTICE(request->api, "badge-reader:update_badge query=%s", json_object_get_string(queryJ));
    afb_req_success_f(request,json_object_new_string(response), NULL);

    return;
}

/**
 * sqldb_list_badge - Callback of the list badge verb
 * 
 * @param request afb request
 */
static void sqldb_list_badge (afb_req_t request) {
    int ret = 0;
    char **list_key=NULL;
    int list_length = 0;
    char * errMsg = NULL;
    json_object *queryJ =  afb_req_json(request);
    json_object *responseJ = NULL;
    json_object *listJ = json_object_new_array();
    
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(afb_req_get_api(request));
    user_data_t * user_data = (user_data_t *) ctlConfig->external;
    sqldbData_t *sql_data = user_data->sql_data;

    list_key = sqldb_list(sql_data, &list_length, &errMsg);
    if (list_key == NULL) {
        AFB_API_ERROR(afb_req_get_api(request), "[sqldb_list_badge] Get error[%i] to list badges: %s", ret, errMsg);
        afb_req_fail_f(request, "query-error", "sqldb_list_badge: Failed to list badges: %s", errMsg);
        free(errMsg);
        return ;
    }

    for (int index = 0; index < list_length; index ++) {
        ret = json_object_array_add(listJ, json_object_new_string(list_key[index]));
        if (ret < 0) {
            AFB_API_ERROR(afb_req_get_api(request), "[sqldb_list_badge] Get error[%i] to wrap response", ret);
            afb_req_fail_f(request, "query-error", "sqldb_list_badge: Failed to send response");
            return ;
        }
    }
    

    ret = wrap_json_pack(&responseJ, "{so*}", "list", listJ);
    if (ret < 0) {
        AFB_API_ERROR(afb_req_get_api(request), "[sqldb_list_badge] Get error[%i] to add list to response", ret);
        afb_req_fail_f(request, "query-error", "sqldb_list_badge: Failed to add list to response");
        return ;
    }

    AFB_API_NOTICE(request->api, "badge-reader:ping query=%s", json_object_get_string(queryJ));
    afb_req_success_f(request,responseJ, NULL);

    return;
}

/**
 * sqldb_delete_badge - Callback of the delete badge verb
 * 
 * @param request afb request
 */
static void sqldb_delete_badge (afb_req_t request) {
    char response[64];
    int ret = 0;
    char * key = NULL;
    char * errMsg = NULL;
    json_object *queryJ =  afb_req_json(request);
    
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(afb_req_get_api(request));
    user_data_t * user_data = (user_data_t *) ctlConfig->external;
    sqldbData_t *sql_data = user_data->sql_data;

    if (json_object_get_type(queryJ) != json_type_object) {
        AFB_API_ERROR(afb_req_get_api(request), "[sqldb_delete_badge] Query is in wrong type reader");
        afb_req_fail_f(request, "query-error", "sqldb_delete_badge: invalid type of the query");
        return;
    }
    ret= wrap_json_unpack(queryJ, "{ss}", "key", &key);
    if (ret < 0) {
        afb_req_fail_f(request, "query-error", "sqldb_delete_badge: invalid 'json' need to be {'key'='keyID'}");
        return;
    }

    ret = sqldb_delete(sql_data, key, &errMsg);
    if (ret < 0) {
        if (ret == -SQLDB_ERROR_DELETE) {
            AFB_API_ERROR(afb_req_get_api(request), "[sqldb_delete_badge] Get error[%i] to delete badge[%s]: %s", ret, key, errMsg);
            afb_req_fail_f(request, "query-error", "sqldb_delete_badge: Badge[%s] hasn't been found", key);
        } else {
            AFB_API_ERROR(afb_req_get_api(request), "[sqldb_delete_badge] Get error[%i] to add badge[%s]: %s", ret, key, errMsg);
            afb_req_fail_f(request, "query-error", "sqldb_delete_badge: Failed to add badge[%s]: %s", key, errMsg);
        }
        free(errMsg);
        return ;
    }
    snprintf (response, sizeof(response), "Badge=%s has been deleted from database", key);
    AFB_API_NOTICE(request->api, "badge-reader:delete_badge query=%s", json_object_get_string(queryJ));
    afb_req_success_f(request,json_object_new_string(response), NULL);

    return;
}


/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            BINDING FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * info_verb_cb - Callback of the info verb
 * 
 * @param request afb request
 */
static void list_verb_cb (afb_req_t request) {
    family_reader_t * fReader = (family_reader_t *) afb_req_get_vcbdata(request);
    afb_api_t api = afb_req_get_api(request);
    json_object * responseJ = json_object_new_array();
    json_object * familyReaderJ = NULL;
    json_object * readerJ = NULL;
    int ret = 0;
    int indexFamily = 0;
    int indexReaders = 0;

    // Loop on each family reader
    for (indexFamily = 0; fReader[indexFamily].uid; indexFamily++) {

        // Loop on every Reader of the Family
        json_object * readersJ = json_object_new_array();
        for (indexReaders = 0; fReader[indexFamily].readers[indexReaders].uid; indexReaders++) {
            char* stateStr;
            ret = stateToString(fReader[indexFamily].readers[indexReaders].state, &stateStr);
            if (ret < 0) {
                AFB_API_ERROR(api, "[ListReader] Unable to convert state to string for reader %s/%s",
                    fReader[indexFamily].uid, fReader[indexFamily].readers[indexReaders].uid);
                afb_req_fail_f(request, "list error","Unable to convert state to string for reader %s/%s",
                    fReader[indexFamily].uid, fReader[indexFamily].readers[indexReaders].uid);
                return;
            }
            ret = wrap_json_pack(&readerJ, "{ss s?s s?s}",
                "uid", fReader[indexFamily].readers[indexReaders].uid,
                "info", fReader[indexFamily].readers[indexReaders].info,
                "state", stateStr);
            if (ret < 0) {
                AFB_API_ERROR(api, "[ListReader] Unable to wrap Reader %s",
                    fReader[indexFamily].readers[indexReaders].uid);
                afb_req_fail_f(request, "list error","Unable to wrap Reader %s",
                    fReader[indexFamily].readers[indexReaders].uid);
                free(stateStr);
                return;
            }
            free(stateStr);
            json_object_array_add(readersJ, readerJ);
        }

        ret = wrap_json_pack(&familyReaderJ, "{ss s?s s?i s?o*}",
            "uid", fReader[indexFamily].uid,
            "info", fReader[indexFamily].info,
            "countReaders", indexReaders,
            "readers", readersJ);
        if (ret < 0) {
            AFB_API_ERROR(api, "[ListReader] Unable to wrap familyReader %s",
                fReader[indexFamily].uid);
            afb_req_fail_f(request, "list error","Unable to wrap familyReader %s",
                fReader[indexFamily].uid);
            return;
        }
        json_object_array_add(responseJ, familyReaderJ);
    }
    afb_req_success(request, responseJ, NULL);
    AFB_API_NOTICE(api, "[ListReader] Response to list request in success");
}

/**
 * CtrlLoadStaticVerbs - Load all Statics verbs
 * 
 * @param api afb api
 * @param verbs pointer to the list of verbs with their callbacks
 * @param vcbdata context data
 */
static int CtrlLoadStaticVerbs (afb_api_t api, afb_verb_t *verbs, void *vcbdata) {
    int errcount=0;

    for (int idx=0; verbs[idx].verb; idx++) {
        errcount+= afb_api_add_verb(api, CtrlApiVerbs[idx].verb, CtrlApiVerbs[idx].info, CtrlApiVerbs[idx].callback, vcbdata, 0, 0, 0);
    }

    return errcount;
};

/**
 * ReaderDynRequest - CAll dynamically the function which take cares about request for Reader
 * 
 * @param request afb request
 */
static void ReaderDynRequest(afb_req_t request) {
    // retrieve action handle from request and execute the request
    json_object *queryJ = afb_req_json(request);
    reader_t* reader = (reader_t *) afb_req_get_vcbdata(request);

    ReaderRequest(request, reader, queryJ);
}

/**
 * ReaderDynRequest - CAll dynamically the function which take cares about request for Family Reader
 * 
 * @param request afb request
 */
static void FamilyReaderDynRequest(afb_req_t request) {
    // retrieve action handle from request and execute the request
    json_object *queryJ = afb_req_json(request);
    family_reader_t* fReader= (family_reader_t*) afb_req_get_vcbdata(request);
    FamilyReaderRequest(request, fReader, queryJ);
}

/**
 * ReaderLoadOne - Check the config file to Load a reader
 * 
 * @param api afb_api_t 
 * @param reader ptr of reader_t
 * @param readerJ Json of the reader config.
 * 
 * @param return errors
 */
static int ReaderLoadOne(afb_api_t api, family_reader_t * fReader, reader_t * reader, json_object * readerJ, uint16_t count) {
    int ret = 0;
    afb_auth_t *authent=NULL;

    // Should already be allocated
    assert (readerJ);
    memset(reader, 0, sizeof(reader_t));
    reader->fReader = fReader;
    reader->numero = count;
    reader->loa = 2;

    char * rank = NULL;
    char * mode =NULL;

    ret = wrap_json_unpack(readerJ, "{ss,s?s,s?s,s?s,s?i,s?i,s?o}",
                "uid", &reader->uid,
                "info", &reader->info,
                "mode", &mode,
                "rank", &rank,
                "loa", &reader->loa,
                "verbose", &reader->verbose,
                "config", &reader->specificConfig);
    if (ret < 0) {
        AFB_API_ERROR(api, "Fail to Parse Reader configs");
        return -1;
    }

    // Catch Rank
    if (rank && !strncmp(rank, "admin", 4)) {
        reader->rank = READER_RANK_ADMIN;
    } else {
        reader->rank = READER_RANK_STANDARD;
    }

    // Catch mode
    if (mode && !strncmp(mode, "auto", 4)) {
        reader->mode = READER_MODE_AUTO;
    } else {
        reader->mode = READER_MODE_MANUAL;
    }

    // Keep specificConfig as object when defined
    if (reader->specificConfig) 
        json_object_get (reader->specificConfig);

    // Store api into the object
    reader->api = api;

    // Create the apiverb according to the reader
    ret = asprintf ((char**) &reader->apiverb, "%s/%s", fReader->uid, reader->uid);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "[ReaderLoadOne] Errro on asprintf for apiVerb for %s/%s",
            reader->fReader->uid, reader->uid);
        return -1;
    }

    // Add dynamically all verbs
    ret = afb_api_add_verb(api, reader->apiverb, reader->info, ReaderDynRequest, reader, authent, AFB_SESSION_NONE, 0);
    if (ret < 0) {
        AFB_API_ERROR(api, "[ReaderLoadOne] Fail to register API verb=%s", reader->apiverb);
        return -1;
    }

    // If defined call function init callback
    if (fReader->functions->initPtr) {
        ret = fReader->functions->initPtr(reader, readerJ);
        if (ret < 0) {
            AFB_API_ERROR(api, "[ReaderLoadOne] Fail during init callback uid=%s", reader->uid);
            return -1;
        }
    }

    // Initiate the state to disconnected
    reader->state = READER_STATE_DISCONNECTED;

    // Start the monitoring state reader
    int group = 1;
    int force = 0;
    if (reader->rank == READER_RANK_ADMIN) {
        group = 0;
        force = 1;
    }
    ret = reader_monitor_state(reader, group, force, reader->verbose);

    // Store callbacks for data caught by reader
    reader->magic_key_part1.length = 4;
    reader->magic_key_part1.key = (char *)malloc(4*sizeof(char));
    reader->magic_key_part2.length = 4;
    reader->magic_key_part2.key = (char *)malloc(4*sizeof(char));

    return 0;
}

/**
 * Function: FamilyReaderLoadOne - Check the config file to Load a type
 *  of badge reader, create dynamically a verb
 * 
 * @param api: afb_api_t 
 * @param fReader: ptr of family_reader_t
 * @param fReaderJ: Json of the badge-reader config.
 * @return errors
 */
static int FamilyReaderLoadOne(afb_api_t api, family_reader_t * fReader, json_object * fReaderJ) {
    int ret = 0;
    json_object *readersJ;
    const char *format=NULL;
    afb_auth_t *authent=NULL;

    // Should already be allocated
    assert(fReaderJ);
    assert(api);

    memset(fReader, 0, sizeof(family_reader_t));
    ret = wrap_json_unpack(fReaderJ, "{ss,s?s,s?s,s?o}",
            "uid", &fReader->uid,
            "info", &fReader->info,
            "format", &format,
            "readers", &readersJ);
    if (ret < 0) {
        AFB_API_ERROR(api, "[FamilyReaderLoadOne] Fail to parse badge reader JSON : (%s)", json_object_to_json_string(fReaderJ));
        return -1;
    }

    ret = asprintf ((char**)&fReader->adminapi, "%s/%s", fReader->uid, "admin");
    if (ret < 0) {
        AFB_API_ERROR(api, "[FamilyReaderLoadOne] Fail to create apiverb for %s", fReader->uid);
        return -1;
    }
    ret = afb_api_add_verb(api, fReader->adminapi, fReader->info, FamilyReaderDynRequest, fReader, authent, AFB_SESSION_NONE, 0);
    if (ret < 0) {
        AFB_API_ERROR(api, "[FamilyReaderLoadOne] Fail to register API uid=%s verb=%s info=%s", fReader->uid, fReader->adminapi, fReader->info);
        return -1;
    }

    // Find function Callback
    fReader->functions = familyReaderFunctionFind(api, format);
    if (!fReader->functions) {
        AFB_API_ERROR(api, "FamilyReaderLoadOne: missing/invalid Format Type=%s JSON=%s", fReader->uid, json_object_to_json_string(fReaderJ));
        return -1;
    }

    // Loop on readers
    if (json_object_is_type(readersJ, json_type_array)) {
        int count = (int) json_object_array_length(readersJ);
        fReader->readers = (reader_t *) calloc(count +1, sizeof(reader_t));

        for (uint16_t index = 0; index < count; index++) {
            json_object * readerJ = json_object_array_get_idx(readersJ, index);
            ret = ReaderLoadOne(api, fReader, &fReader->readers[index], readerJ, index);
            if (ret < 0) 
                return -1;
        }
    } else {
        fReader->readers = (reader_t *) calloc(2, sizeof(reader_t));
        ret = ReaderLoadOne(api, fReader, &fReader->readers[0], readersJ, 0);
        if (ret < 0) 
            return -1;
    }

    return 0;
}

/**
 * FamilyReaderConfig - Callback of the controller section "badge-reader"
 *  Load and store into struct the API config
 * 
 * @param api: afb_api_t 
 * @param section ptr of CtlSectionT
 * @param fReadersJ Json of the badge-reders config.
 * @return errors
 */
static int FamilyReaderConfig(afb_api_t api, CtlSectionT *section, json_object *fReadersJ) {
    // Everything is done during initial config call
    if (!fReadersJ) return 0;

    int ret = 0;
    char * msgErr = NULL;
    //family_reader_t * fReaders;
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(api);
    user_data_t * user_data = (user_data_t *) ctlConfig->external;

    // Create database
    ret = sqldb_create(user_data->sql_data, &msgErr);
    if (ret < 0) {
        AFB_API_ERROR(api, "Failed to create database: %s", msgErr);
        free(msgErr);
        return -1;
    } else if (ret == 1) {
        AFB_API_DEBUG(api, "Database already open");
    }
    AFB_API_DEBUG(api, "Database created");

    // Initiate familyReader in data user
    if (json_object_is_type(fReadersJ, json_type_array)) {
        int count = (int) json_object_array_length(fReadersJ);
        user_data->familyreader = (family_reader_t *) calloc(count+1, sizeof(family_reader_t));

        for (int index = 0; index < count; index++) {
            json_object * fReaderJ = json_object_array_get_idx(fReadersJ, index);
            ret = FamilyReaderLoadOne(api, &user_data->familyreader[index], fReaderJ);
            if (ret < 0) goto OnErrorExit;
        }
    } else {
        user_data->familyreader = (family_reader_t *) calloc(2, sizeof(family_reader_t));
        ret = FamilyReaderLoadOne(api, &user_data->familyreader[0], fReadersJ);
        if (ret < 0) goto OnErrorExit;
    }

    // add static controls verbs
    ret = CtrlLoadStaticVerbs (api, CtrlApiVerbs, (void*) user_data->familyreader);
    if (ret < 0) {
        AFB_API_ERROR(api, "CtrlLoadOneApi fail to Registry static API verbs");
        goto OnErrorExit;
    }

    return 0;

OnErrorExit:
    AFB_API_ERROR(api, "Fail to Initiate FamilyReader check Json Config");
    return -1;
}

/**
 * DatabaseConfig - Callback of the controller section "database"
 *  Load and store into struct the API config
 * 
 * @param api: afb_api_t 
 * @param section ptr of CtlSectionT
 * @param databaseJ Json of the badge-reders config.
 * @return errors
 */
static int DatabaseConfig(afb_api_t api, CtlSectionT *section, json_object *databaseJ) {
    int ret  = 0;
    // Everything is done during initial config call
    if (!databaseJ) goto OnExit;

    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(api);
    user_data_t * user_data = (user_data_t *) ctlConfig->external;

    ret = wrap_json_unpack(databaseJ, "{ss}"
        , "path", &user_data->sql_data->dbPath
    );
    if (ret < 0) {
        AFB_API_ERROR(afbBindingRoot, "[%s] json unsupported key", __func__);
        goto OnExit;
    }
    
OnExit:
    return ret;
}

/**
 * CommandConfig - Callback of the controller section "commands"
 *  Load and store into struct the API config
 * 
 * @param api: afb_api_t 
 * @param section ptr of CtlSectionT
 * @param commandsJ Json of the badge-reders config.
 * @return errors
 */
static int CommandConfig(afb_api_t api, CtlSectionT *section, json_object *commandsJ) {
    int ret  = 0;
    // Everything is done during initial config call
    if (!commandsJ) goto OnExit;

    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(api);
    user_data_t * user_data = (user_data_t *) ctlConfig->external;

    user_data->cmd_keys = reader_parse_cmd_key(commandsJ);
    if (user_data->cmd_keys == NULL) {
        ret = -1;
    }
    
OnExit:
    return ret;
}

/**
 * CtrlInitOneApi - Init a single API
 * 
 * @param api afb_api_t 
 * @return errors
 */
static int CtrlInitOneApi(afb_api_t api) {
    int ret = 0;
    // retrieve section config from api handle
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(api);
    user_data_t * user_data = (user_data_t *) ctlConfig->external;

    // Init SQLDB data
    if (sqldb_init(user_data->sql_data)) {
        AFB_API_ERROR(api, "Failed to init sqldb data data");
        return -1;
    }

    ret = CtlConfigExec(api, ctlConfig);
    if (ret) {
        AFB_API_ERROR(api, "Error at CtlConfigExec step");
    }

    return ret;
}

/**
 * CtrlLoadOneApi - Load a single API
 * 
 * @param vcbdata ptr of CtlConfigT
 * @param api afb_api_t
 * 
 * @return errors
 */
static int CtrlLoadOneApi(void* vcbdata, afb_api_t api) {
    CtlConfigT* ctlConfig = (CtlConfigT*)vcbdata;
    user_data_t * user_data = (user_data_t *)calloc(1, sizeof(user_data_t));
    user_data->sql_data = sqldb_preinit();
    user_data->event = (afb_event_t *)calloc(MAX_EVENT_LIST, sizeof(afb_event_t));
    ctlConfig->external = (void *) user_data;

    // save closure as api's data context
    afb_api_set_userdata(api, ctlConfig);

    // load section for corresponding API
    int error = CtlLoadSections(api, ctlConfig, ctrlSections);

    // init and seal API function
    afb_api_on_init(api, CtrlInitOneApi);
    afb_api_seal(api);

    return error;    
}

/**
 * afbBindingEntry - Entry point of the binding
 * 
 * @param api afb_api_t 
 * @return status
 */
int afbBindingEntry(afb_api_t api) {
    int status = 0;
    char *searchPath = NULL;
    const char *envConfig = NULL;
    afb_api_t handle;

    AFB_API_NOTICE(api, "Controller in afbBindingEntry");

    envConfig = getenv("CONTROL_CONFIG_PATH");
    if (!envConfig){
        AFB_API_NOTICE(api, "Using default environnement config path : %s", CONTROL_CONFIG_PATH);
        envConfig = CONTROL_CONFIG_PATH;
    }
    else AFB_API_NOTICE(api, "Found environnement config path : %s", envConfig);


    status=asprintf (&searchPath,"%s:%s", envConfig, GetBindingDirPath(api));
    AFB_API_NOTICE(api, "Json config directory : %s", searchPath);

    const char* prefix = "control";
    const char* configPath = CtlConfigSearch(api, searchPath, prefix);
    if (!configPath) {
        AFB_API_ERROR(api, "afbBindingEntry: No %s-%s* config found in %s ", prefix, GetBinderName(), searchPath);
        status = -1;
        free(searchPath);
        return status;
    }

    AFB_API_NOTICE(api, "api will be using config : %s", configPath);

    // load config file and create API
    CtlConfigT* ctrlConfig = CtlLoadMetaData(api, configPath);
    if (!ctrlConfig) {
        AFB_API_ERROR(api, "afbBindingEntry No valid control config file in:\n-- %s", configPath);
        status = -1;
        free((void *)configPath);
        free(searchPath);
        return status;
    }

    AFB_API_NOTICE(api, "Controller API='%s' info='%s'", ctrlConfig->api, ctrlConfig->info);

    // create one API per config file (Pre-V3 return code ToBeChanged)
    handle = afb_api_new_api(api, ctrlConfig->api, ctrlConfig->info, 1, CtrlLoadOneApi, ctrlConfig);
    status = (handle) ? 0 : -1;

    free((void *)configPath);
    free(searchPath);
    return status;
   
}