/*
 * Copyright 2020 - 2021 IoT.bzh
 *
 * author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
#define _GNU_SOURCE

#ifndef _READER_BADGE_BINDING_INCLUDE_
#define _READER_BADGE_BINDING_INCLUDE_

#define  AFB_BINDING_VERSION 3

#define MAGIC_KEY "magicencrypted"

#define MAX_EVENT_LIST 2
#define GOOD_BADGE_EVENT "good-badges"
#define BAD_BADGE_EVENT "bad-badges"

#define BAD_BADGE_ERROR_NOT_FOUND "Not found"
#define BAD_BADGE_ERROR_CLONED "cloned"
#define BAD_BADGE_ERROR_KEY "Bad key"
#define BAD_BADGE_ERROR_MAGIC_KEY "Bad magic key"

#ifndef ERROR
  #define ERROR -1
#endif

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

// usefull classical include
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdbool.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <errno.h>

// Includes for binding dev
#include <afb/afb-binding.h>
#include <afb-timer.h>
#include <wrap-json.h>
#include <ctl-config.h>

// Include own library
#include "./sqldb/sqldb.h"

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            ENUM
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
typedef enum {
  READER_TYPE_UNSET=0,      // Null is not a valid default
  READER_TYPE_CN56,
} reader_type_e;

/**
 * reader_mode_e - define the mode of the reader
 * 
 * AUTO: Check code into config and execute command if id Card found
 * TEST: Don't care about the car deaden and execute the command
 * MANUAL: Just send code to the client and lets decision to the client
 */
typedef enum {
  READER_MODE_MANUAL=0,      // Null is not a valid default
  READER_MODE_AUTO,
  READER_MODE_TEST
} reader_mode_e; 

/**
 * State - Enum which List of different kind of state
 */
typedef enum {
  READER_STATE_CONNECTED = 0,
  READER_STATE_ERROR,
  READER_STATE_RESET,
  READER_STATE_CONNECTING,
  READER_STATE_DISCONNECTED
} reader_state_e;

/**
 * Rank - Enum which List of different kind of rank
 */
typedef enum {
  READER_RANK_STANDARD = 0,
  READER_RANK_ADMIN
} reader_rank_e;

/**
 * Action - Enum which List of different kind of action
 */
typedef enum {
    READER_ACTION_UNKNOWN=0,
    READER_ACTION_LED,
    READER_ACTION_READ,
    READER_ACTION_WRITE,
    READER_ACTION_RESET,
    READER_ACTION_AUTHENTICATION,
    READER_ACTION_TRAILER,
    READER_ACTION_EXT
} reader_action_e;

/**
 * Error - Enum which List of different kind of error
 */
typedef enum  {
  READER_ERROR_DISCONNECTED=1,
  READER_ERROR_ALREADY_CONNECTED,
  READER_ERROR_NULL_ARGS,
  READER_ERROR_SOCKET,
  READER_ERROR_EPOLL,
  READER_ERROR_CREATION_THREAD,
  READER_ERROR_CONNECT,
  READER_ERROR_WRITE,
  READER_ERROR_READ,
  READER_ERROR_CLOSED_SOCKET,
  READER_ERROR_NOT_SUCCESS,
  READER_ERROR_BAD_SEQ,
  READER_ERROR_WRONG_ARGS,
  READER_ERROR_NO_DATA,
  READER_ERROR_WRONG_BLK,
  READER_ERROR_TIMEOUT,
  READER_ERROR_AUTHENTICATE
}reader_error_e;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            STRUCTURES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

typedef struct {
    const char *uid;
    int idx;    // idx 0 = KeyA // idx 1 == KeyB
    uint8_t *value;
    uint8_t len;
} reader_key_t;

typedef struct {
    u_int8_t *acls;
    u_int8_t alen;
    reader_key_t *keyA;
    reader_key_t *keyB;
} reader_trailer_t;

typedef struct {
    const char *uid;
    int group;
    const uint8_t sec; // sector number (0 for mifare)
    const uint8_t blk; // block number for NFC tag-2
    u_int8_t *data;
    ulong dlen;
    const reader_key_t *key;
    reader_action_e action;
    reader_trailer_t *trailer;
} reader_cmd_t;

/**
 * reader_args_cmd_t - Structure of the command arguments from request
 * @param key Uid of the key to use
 * @param sec N° sector
 * @param blk N° block
 * @param dlen Data length
 * @param data Data
 */
typedef struct {
  reader_key_t *key;
  uint8_t sec;
  uint8_t blk;
  ulong dlen;
  uint8_t *data;
} reader_args_cmd_t;

typedef struct {
    reader_key_t *keys;
    size_t keys_nbr;
    reader_cmd_t *cmds;
    size_t cmds_nbr;
} binding_cmd_key_t;

typedef struct family_reader_s family_reader_t;
typedef struct reader_s reader_t;
typedef struct family_reader_func_ptr_s family_reader_func_ptr_t;

/**
 * magic_key_t - Structure of the magic key
 * @param key value of the key
 * @param length length of the key
 */
typedef struct {
    char *key;
    uint8_t length;
} magic_key_t;

typedef struct {
  binding_cmd_key_t *cmd_keys;
  family_reader_t *familyreader;
  sqldbData_t *sql_data;
  afb_event_t *event;
} user_data_t;

struct family_reader_s {
  const char *uid;
  const char *info;
  const char *adminapi;
  reader_t *readers;
  family_reader_func_ptr_t *functions;
};

struct reader_s {
  const char *uid;
  reader_rank_e rank;
  reader_mode_e mode;
  const char *apiverb;
  const char * info;
  int loa; // 0 -> no verification badges, 1 -> check on database, 2 -> execute command groups and check in database
  int verbose;
  family_reader_t *fReader;
  uint16_t numero;
  json_object *specificConfig;
  json_object *sample;
  afb_api_t api;
  afb_event_t event;
  TimerHandleT *timer;
  magic_key_t magic_key_part1;
  magic_key_t magic_key_part2;
  // void (*response_handler) (uint8_t *uuid, const char * cmd_uid, uint8_t * response_buffer, int response_len, void * user_data);
  // void (*bad_authent_handler) (uint8_t *uuid, void * user_data);
  pthread_t thread_auto_id;
  reader_state_e state;
  reader_action_e action;
  void * ctx;
};

/**
 * read_badge_ptr_t - Pointer function calling plugin read function
 * 
 * @param reader Poiter to Reader context
 * @param key Key to use for read, if no-one, use the default key
 * @param sec Sector where to read
 * @param blk block to read
 * @param len Length of data to read
 * @param response Response returned by the reader, need to be allocated
 * 
 * @return 0 in success, <0 otherwise
 */
typedef int (*read_badge_ptr_t) (reader_t *reader, const reader_key_t *key, uint8_t sec, uint8_t blk, uint8_t len, uint8_t *response);

/**
 * write_badge_ptr_t - Pointer function calling plugin write function
 * 
 * @param reader Reader context
 * @param key Key to use for read, if no-one, use the default key
 * @param sec Sector where to write
 * @param blk block to write
 * @param len Data length to write
 * @param buffer Data to write
 * 
 * @return 0 in success, <0 otherwise
 */
typedef int (*write_badge_ptr_t) (reader_t *reader, const reader_key_t *key, uint8_t sec, uint8_t blk, uint8_t len, uint8_t *buffer);

/**
 * trailer_badge_ptr_t - Pointer function calling plugin trailer function
 * 
 * @param reader Reader context
 * @param key Key to use for read, if no-one, use the default key
 * @param sec Sector where to write
 * @param trailer trailer data
 * 
 * @return 0 in success, <0 otherwise
 */
typedef int (*trailer_badge_ptr_t) (reader_t *reader, const reader_key_t *key, uint8_t sec, reader_trailer_t *trailer);

struct family_reader_func_ptr_s {
  const char *uid;
  const char *info;
  reader_type_e type;
  int (*initPtr) (reader_t *reader, json_object *argJ);
  int (*connectPtr) (reader_t *reader);
  int (*catchUuidPtr) (reader_t *reader, uint8_t **uuid);
  read_badge_ptr_t readBadgePtr;
  //int (*readBadgePtr) (reader_t *reader, const cn56_key_t *key, uint8_t sec, uint8_t blk, uint8_t len, uint8_t *response);
  write_badge_ptr_t writeBadgePtr;
  //int (*writeBadgePtr)(reader_t *reader, json_object *inputJ);
  trailer_badge_ptr_t trailerBadgePtr;
  int (*activateExtPtr)(reader_t *reader);
  int (*errorPtr)(reader_t *reader);
  int (*subscribePtr)(reader_t *reader);
  int (*unSubscribePtr)(reader_t *reader);
  int (*onResetPtr)(reader_t *reader);
};

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            DECLARATIONS FUNCTIONS - badge-reader-glue.c
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

int stateToString(reader_state_e state, char **stateStr);
void FamilyReaderReadersId(family_reader_t * fReader, int verbose, json_object *responseJ);
void FamilyReaderRequest(afb_req_t request, family_reader_t * fReader, json_object *queryJ);
void ReaderRequest(afb_req_t request, reader_t * reader, json_object *queryJ);

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            DECLARATIONS FUNCTIONS - badge-reader-callback.c
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

void fReaderPointerFunctionsRegister (char *uid, family_reader_func_ptr_t *functionPtr);
family_reader_func_ptr_t *mvOneFunctionFind (family_reader_func_ptr_t *functions, const char *uid);
family_reader_func_ptr_t * familyReaderFunctionFind (afb_api_t api, const char *uri);
typedef void (*registerCbT) (const char *uid, family_reader_func_ptr_t *encoderCB);

/**
 * bad_authent_handler_cb - Must be called when caught a non-authenticated badge
 * 
 * @param uuid The Unique ID of the reader
 * @param user_data pointer to the user data
 */
void bad_authent_handler_cb(uint8_t *uuid, void *user_data);

/**
 * response_handler_cb - Must be called when read command is called and data are available
 * 
 * @param uuid The Unique ID of the reader
 * @param cmd_uid Uid of the command
 * @param response Buffer containing response
 * @param response_len Length of the response
 * @param user_data pointer to the user data
 */
void response_handler_cb(uint8_t *uuid, const char * cmd_uid, uint8_t * response_buffer, int response_len, void * user_data);


/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            DECLARATIONS FUNCTIONS - badge-reader-config.c
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * reader_parse_cmd_key - Parse the confif of cmd and key
 * 
 * @param configJ config in json object
 * @return pointer to an allocated config object, if NULL = error occurred
 */
binding_cmd_key_t *reader_parse_cmd_key (json_object *configJ);

/**
 * reader_exec_one_cmd - Sending a command to the reader
 * 
 * @param reader Reader object which to send the command
 * @param cmd Command object to send to the reader
 * @param badge_uuid Uuid of the badge caught
 * @return 0 in success, <0 if error
 */
int reader_exec_one_cmd(reader_t *reader, const reader_cmd_t *cmd, uint8_t *badge_uuid);

/**
 * exec_group_cmd - Execute a group of command
 * 
 * @param reader Reader object which to send the command
 * @param config Configuration of the reader
 * @param group N° of group to exec
 * @param force if set Force to continu next comand even if the previous failed
 * @param verbose Verbosity for log
 * @param badge_uuid Uuid of the badge caught
 * 
 * @return 0 in success, <0 if error
 */
int exec_group_cmd (reader_t * reader, binding_cmd_key_t * config, int group, int force, int verbose, uint8_t *badge_uuid);

/** reader_monitor_state - Creat thread which will try to send group command each second
 * 
 * @param reader pointer to reader context
 * @param config Configuration of the reader
 * @param group N° of group to exec in automode
 * @param force if set Force:
 *      1 -> continu next comand even if the previous failed
 *      2 -> continue group command even if the badge stay at the badge reader
 * @param verbose Verbosity for log
 * 
 * @return 0 in success, <0 in error
 */
int reader_monitor_state (reader_t *reader, int group, int force, int verbose);

/** reader_monitor_wait - Wait until auto failled
 * 
 * @param reader pointer to reader context
 * 
 * @return 0 in success, <0 in error
 */
int reader_monitor_wait (reader_t *reader);

/**
 * reader_free_config - free all allocated part of a config object
 * 
 * @param config pointer to an allocated config object
 */
void reader_free_config(binding_cmd_key_t *config);

#endif /*_READER_BADGE_BINDING_INCLUDE_*/