/*
 * Copyright 2020 - 2021 IoT.bzh
 *
 * author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#define _GNU_SOURCE

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "badge-reader-binding.h"

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            STRUCTURES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/** 
 * badgeReaderRegistryS - Contain all callbacks functions and information about the Family Reader
 */
typedef struct badgeReaderRegistryS {
   char *uid;
   struct badgeReaderRegistryS *next;
   family_reader_func_ptr_t *functionPtr;
} badgeReaderRegistryT;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            GLOBAL VARIABLES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

// registry holds a linked list of core+pugins encoders
static badgeReaderRegistryT *registryHead = NULL;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */


/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PUBLIC FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * fReaderPointerFunctionsRegister - Register all functions pointers from plugins
 * 
 * @param uid of the plugin
 * @param functionPtr pointer to the structure containing all functions pointer
 */
void fReaderPointerFunctionsRegister (char *uid, family_reader_func_ptr_t *functionPtr) {
    badgeReaderRegistryT * registryIdx, * registryEntry;

    registryEntry = (badgeReaderRegistryT *) calloc(1, sizeof(badgeReaderRegistryT));
    registryEntry->uid = uid;
    registryEntry->functionPtr = functionPtr;

    if (!registryHead) {
        registryHead = registryEntry;
    } else {
        for (registryIdx = registryHead; registryIdx->next; registryIdx=registryIdx->next);
        registryIdx->next = registryEntry;
    }
}

/**
 * PluginParseURI - PArse the uri and catch plugin and format
 * 
 * @param uri Containing raw information
 * @param plugin pointer to the output plugin uid
 * @param plugin pointer to the output format
 */
static int PluginParseURI (const char* uri, char **plugin, char **format) {
    char *chaine;
    int idx;
    static int prefixlen = sizeof(PLUGIN_ACTION_PREFIX)-1;

    if (strncasecmp(uri, PLUGIN_ACTION_PREFIX, prefixlen)) return -1;

    // break URI in substring ignoring leading tcp:
    chaine= strdup(uri);
    if (!chaine) return -1;
    for (idx=prefixlen; idx < strlen(chaine); idx ++) {
        if (chaine[idx] == '#'){
            chaine[idx] = '\0';
            break;
        }
    }

    asprintf(plugin, "%s", &chaine[prefixlen]);
    asprintf(format, "%s", &uri[idx+1]);

    free(chaine);
    return 0;
}

/**
 * mvOneFunctionFind - Find the function whithin one plugin
 * 
 * @param functions Pointer function to find
 * @param uid The uid of the FamilyReader
 * @return The pointer to the plugin function
 */
family_reader_func_ptr_t *mvOneFunctionFind (family_reader_func_ptr_t *functions, const char *uid) {
    int idx;
    assert (uid);
    assert(functions);

    for (idx=0; functions[idx].uid; idx++) {
        if (!strcasecmp (functions[idx].uid, uid)) break;
    }

    return (&functions[idx]);
}

/**
 * familyReaderFunctionFind - Return the correct family_reader_func_ptr_t
 * for FamilyReader with an uid
 * 
 * @param api
 * @param uri The uid of the FamilyReader
 * @return The pointer to the plugin function
 */
family_reader_func_ptr_t * familyReaderFunctionFind (afb_api_t api, const char *uri) {
    char *pluginuid = NULL, *formatuid = NULL;
    badgeReaderRegistryT *registryIdx;
    family_reader_func_ptr_t *functionPtr;

    int ret = PluginParseURI (uri, &pluginuid, &formatuid);
    if (ret < 0) {
        AFB_API_ERROR(api, "familyReaderFunctionFind: Fail to parse uri plugin");
        return NULL;
    }

    if (!pluginuid) return NULL;

    for (registryIdx= registryHead; registryIdx->next; registryIdx=registryIdx->next) {
        if (registryIdx->uid && !strcasecmp (registryIdx->uid, uri)) break;
    }
    if (!registryIdx) {
        AFB_API_ERROR(api, "familyReaderFunctionFind: Fail to find plugin='%s' format encoder ", formatuid);
        return NULL;
    }

    functionPtr = mvOneFunctionFind(registryIdx->functionPtr, formatuid);
    if (!functionPtr || !functionPtr->uid) {
        AFB_API_ERROR(api, "familyReaderFunctionFind: Fail to find plugin='%s' format='%s' encoder", pluginuid, formatuid);
        return NULL;
    }
    free(pluginuid);
    free(formatuid);

    return (functionPtr);
}
