/*
 * Copyright (C) 2020 - 2021 "IoT.bzh"
 *
 * Author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "badge-reader-binding.h"
#include <stdlib.h>
#include <pthread.h>

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE STRUCT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

typedef struct {
    const char *label;
    const int  value;
} reader_key_enum_t;

typedef struct {
    binding_cmd_key_t * config;
    int group;
    int force;
    int verbose;
    reader_t * reader_ctx;
} reader_state_thread_t;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE ENUM
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

static const reader_key_enum_t reader_action_list[] = {
    {"read"     , READER_ACTION_READ},
    {"write"    , READER_ACTION_WRITE},
    {"reset"    , READER_ACTION_RESET},
    {"trailer"  , READER_ACTION_TRAILER},
    {NULL} // terminator
};

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE FUNCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * _reader_label_to_value - search for key label within key/value array
 */
static int _reader_label_to_value (const reader_key_enum_t *keyvals, const char *label) {
    int value=0;
    if (!label) goto OnDefaultExit;

    for (int idx=0; keyvals[idx].label; idx++) {
        if (!strcasecmp (label,keyvals[ idx].label)) {
            value= keyvals[idx].value;
            break;
        }
    }
    return value;

OnDefaultExit:
    return keyvals[0].value;
}

/**
 * _reader_parse_one_data - Parse data from json object to array of hexadecimal
 * 
 * @param dataJ data in json object
 * @param data output to store the data, need to be free in success
 * @param dataLength length of the data
 * @return 0 in success, <0 in error
 */
static int _reader_parse_one_data (json_object *dataJ, u_int8_t **data, ulong *dataLength) {
    switch (json_object_get_type (dataJ)) {
        const char *byteS, *dataS;
        u_int8_t *valueB;
        int err, byte;
        size_t count;

        case json_type_string:
            dataS= json_object_get_string (dataJ);
            // if no dataLength then use data string len
            if (!*dataLength) {
                *dataLength= strlen(dataS);
                *data=(u_int8_t*) strdup(dataS);
            } else {
                *data= malloc (*dataLength);
                strncpy ((char*)*data, dataS, *dataLength);
            }
            break;

        case json_type_array:
            count = json_object_array_length(dataJ);
            valueB= calloc (count+1, sizeof(u_int8_t));

            for (int idx=0; idx < count; idx++) {
                byteS= json_object_get_string (json_object_array_get_idx(dataJ, idx));
                err= sscanf (byteS, "0x%2x", &byte);
                if (err <0) goto OnErrorExit;
                if (byte > 255) goto OnErrorExit;
                valueB[idx]= (u_int8_t)byte;
            }

            *dataLength= count;
            *data= valueB;
            break;

        default:
            goto OnErrorExit;

    }
    return 0;

OnErrorExit:
    AFB_API_ERROR(afbBindingRoot, "[_reader_parse_one_data] key/cmd data should be asci/string or array of hexadecimal/string");
    return -1;
}

/**
 * _reader_parse_one_key - Parse key from json and store to key object
 * 
 * @param keyJ Key found in JSON
 * @param key key struct output
 * @return 0 in success, <0 in error
 */
static int _reader_parse_one_key(json_object *keyJ, reader_key_t *key) {
    int ret = 0;
    json_object *valueJ=NULL;

    // {"uid":"abc, "idx": 0, "value":"asci value" }
    ret = wrap_json_unpack (keyJ, "{ss,s?i,so !}"
        ,"uid", &key->uid
        ,"idx", &key->idx
        ,"value", &valueJ
    );
    if (ret < 0) {
        AFB_API_ERROR(afbBindingRoot, "[_reader_parse_one_key] json supported keys:[uid,idx,value]");
        goto OnErrorExit;
    }
    // value should be an asci string or an array of hexa valueB
    ulong length;
    ret = _reader_parse_one_data (valueJ, &key->value, &length);
    if (ret < 0) {
        goto OnErrorExit;
    }
    key->len= (uint8_t)length;

    return 0;

OnErrorExit:
    return -1;
}

/**
 * _reader_key_by_uid search a key from its uid
 * 
 * @param config General config
 * @param keyUid uid of the key
 * @return pointer to the key object
 */
static reader_key_t *_reader_key_by_uid (binding_cmd_key_t *config, const char *keyUid) {
    reader_key_t *key=NULL;

    if (config->keys) {
        for (int idx=0; config->keys[idx].uid; idx++) {
            if (!strcasecmp ( config->keys[idx].uid, keyUid)) {
                key= &config->keys[idx];
                break;
            }
        }
    }
    return key;
}

/**
 * _reader_parse_one_trailer - Parse the trailer information from the configuration
 * 
 * @param config General config
 * @param trailerJ trailer json object
 * @param trailer output trailer object
 * @return 0 in success, < 0 in error
 */
static int _reader_parse_one_trailer (binding_cmd_key_t *config, json_object *trailerJ, reader_trailer_t **trailer) {
    int ret = 0;
    json_object *valueJ=NULL;
    const char *keyA, *keyB;
    reader_trailer_t *response= calloc (1, sizeof(reader_trailer_t));

    // "trailer": {"keys": ["key-a","keyb"], "acls":["0xF0","0xF7","0x80","0x00"]}
    ret = wrap_json_unpack (trailerJ, "{ss,ss,so !}"
        ,"keyA", &keyA
        ,"keyB", &keyB
        ,"acls", &valueJ
    );
    if (ret < 0) {
        AFB_API_ERROR(afbBindingRoot, "[_reader_parse_one_trailer] json mandatory keys:[keyA,keyA,acls]");
        goto OnErrorExit;
    }

    response->keyA = _reader_key_by_uid (config, keyA);
    response->keyB = _reader_key_by_uid (config, keyB);
    if (!response->keyA  || !response->keyB) {
        AFB_API_ERROR(afbBindingRoot, "[_reader_parse_one_trailer] KeyA=%s keyB=%s not found", keyA, keyB);
        goto OnErrorExit;
    }

    // value should be an asci string or an array of hexa valueB
    ulong alen;
    ret = _reader_parse_one_data (valueJ, &response->acls, &alen);
    if (ret < 0 || alen != 4) goto OnErrorExit;
    response->alen= (uint8_t)alen;

    *trailer= response;
    return 0;

OnErrorExit:
    free (response);
    return ret;
}

/**
 * _reader_parse_one_cmd - Parse command and store to config
 * 
 * @param config pointer to the general config
 * @param cmdJ command found in JSON
 * @param cmd command struct output
 * @return 0 in success, <0 in error
 */
static int _reader_parse_one_cmd (binding_cmd_key_t *config, json_object *cmdJ, reader_cmd_t *cmd) {
    int ret = 0;
    json_object *dataJ=NULL, *trailerJ=NULL;
    const char *keyUid=NULL, *cmdAction=NULL;

    // {"uid":"zzz", "action":"write", "blk": xx, "key":"kuid","data": ["0xab", "0x01", ....]},
    ret = wrap_json_unpack (cmdJ, "{ss,s?i,ss,s?i,s?i,s?i,s?s,s?o,s?o !}"
        ,"uid", &cmd->uid
        ,"group", &cmd->group
        ,"action", &cmdAction
        ,"sec", &cmd->sec
        ,"blk", &cmd->blk
        ,"len", &cmd->dlen
        ,"key", &keyUid
        ,"data", &dataJ
        ,"trailer", &trailerJ
    );
    if (ret) {
        AFB_API_ERROR(afbBindingRoot, "[_reader_parse_one_cmd] json supported keys:[uid,action,blk,key,data,len]");
        goto OnErrorExit;
    }

    // check action
    cmd->action = _reader_label_to_value (reader_action_list, cmdAction);
    switch (cmd->action) {
        case READER_ACTION_READ:
            if (!cmd->dlen || dataJ) {
                if (!keyUid) {
                    keyUid = config->keys[0].uid;
                }
                AFB_API_ERROR(afbBindingRoot, "[_reader_parse_one_cmd] uid=%s action=%s len:mandatory data:forbiden", cmd->uid,cmdAction);
                goto OnErrorExit;
            }
            break;

        case READER_ACTION_WRITE:
            if (!keyUid) {
                keyUid = config->keys[0].uid;
            }
            if (!dataJ) {
                AFB_API_ERROR(afbBindingRoot, "[_reader_parse_one_cmd] uid=%s action=%s data:mandatory", cmd->uid,cmdAction);
                goto OnErrorExit;
            }
            ret = _reader_parse_one_data (dataJ, &cmd->data, &cmd->dlen);
            if (ret < 0) goto OnErrorExit;
            break;

        case READER_ACTION_RESET:
            if (cmd->dlen || dataJ) {
                AFB_API_ERROR(afbBindingRoot, "[_reader_parse_one_cmd] uid=%s action=%s len:forbiden data:forbiden", cmd->uid,cmdAction);
                goto OnErrorExit;
            }
            break;

        case READER_ACTION_TRAILER:
            if (!keyUid) {
                keyUid = config->keys[0].uid;
            }
            if (dataJ || cmd->dlen || !trailerJ) {
                AFB_API_ERROR(afbBindingRoot, "[_reader_parse_one_cmd] uid=%s action=%s trailer:mandatory len+data:forbiden", cmd->uid,cmdAction);
                goto OnErrorExit;
            }
            ret = _reader_parse_one_trailer(config, trailerJ, &cmd->trailer);
            if (ret < 0) goto OnErrorExit;
            break;

        default:
            AFB_API_ERROR(afbBindingRoot, "[_reader_parse_one_cmd] uid=%s action=%s unknown", cmd->uid,cmdAction);
            goto OnErrorExit;
    }

    // if key is defined search for it
    if (keyUid) {
        cmd->key= _reader_key_by_uid (config, keyUid);
        if (!cmd->key) {
            AFB_API_ERROR(afbBindingRoot, "[_reader_parse_one_cmd] cmd=%s keys=%s non found within defined keys]", cmd->uid, keyUid);
            goto OnErrorExit;
        }
    }

    return 0;

OnErrorExit:
    return -1;
}

static int _update_control_data_badge(reader_t* reader, char *new_control) {
    int ret = 0;
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(reader->api);
    user_data_t* user_data = (user_data_t *) ctlConfig->external;
    binding_cmd_key_t * commands = user_data->cmd_keys;

    // Update data control on badge !
    if (reader->fReader->functions->writeBadgePtr) {
        ret = reader->fReader->functions->writeBadgePtr(reader, (const reader_key_t *) &commands->keys[1], 11, 0, 8, (uint8_t *)new_control);
        if (ret < 0 ) {
            int try_count = 0;
            // If failed try again 
            while(ret < 0 && try_count <= 3) {
                ret = reader->fReader->functions->writeBadgePtr(reader, (const reader_key_t *) &commands->keys[1], 11, 0, 8, (uint8_t *)new_control);
                try_count ++;
            }
            if (ret < 0) {
                AFB_API_WARNING(reader->api, "[_update_control_data_badge] Failed to send write command with error: %i", ret);
            } else {
                AFB_API_DEBUG(reader->api, "[_update_control_data_badge] write command success after %i try", try_count);
            }
            goto OnExit;
        }
    }else {
        AFB_API_ERROR(reader->api, "[_update_control_data_badge] missing plugin write function !\n");
        ret =  -1;
        goto OnExit;
    }

OnExit:
    return ret;
}

static int _update_control_data_db(sqldbData_t * sql_data, char *badge_uid, char *new_control) {
    int ret = 0;
    char *errMsg = NULL;

    // Update data control in database
    ret = sqldb_update_control(sql_data, badge_uid, new_control, &errMsg);
    if (ret < 0) {
        free(errMsg);
        goto OnExit;
    }

OnExit:
    return ret;
}

/**
 * _push_event - Push afb event for subscription client
 * 
 * @param reader reader context
 * @param responseJ Formatted and allocated response in JSON (will be freed if cannot been send)
 * @param event the afb event pointer
 * @param status 0 = success, 1 = failure
 */
static void _push_event(reader_t *reader, json_object *responseJ, afb_event_t* event, int status) {
    bool event_was_pushed = false;
    int event_count = 0;

    if (event[status] && afb_event_is_valid(event[status])) {
        event_count = afb_event_push(event[status], responseJ);
        event_was_pushed = true;
        if (event_count == 0) {
            AFB_API_DEBUG(reader->api, "[_push_event] No more client listen global event %s/%s", reader->fReader->uid, reader->uid);
        }
    }

    // Clear response JSON if not push
    if (!event_was_pushed) {
        json_object_put(responseJ);
    }
}

static int _endpoint_process(reader_t * reader, char *badge_uid, char *control_badge, int control_length) {
    int ret = 0;

    // User data variables
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(reader->api);
    user_data_t* user_data = (user_data_t *) ctlConfig->external;

    // Date variables
    time_t t = time(NULL);
    struct tm tm = *localtime(&t);

    // Event variables
    json_object * responseJ = NULL;

    // Badges informations variables
    char * key_uid = NULL;
    char * name = NULL;
    char * control = NULL;
    char * errMsg = NULL;
    char * new_control = NULL;

    // Get the actual date
    ret = asprintf(&new_control, "%d%02d%02d", tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "[_endpoint_process] error to getting date !");
        goto OnExit;
    }

    // Verify in database if badges exists
    ret = sqldb_get(user_data->sql_data, badge_uid, &key_uid, &control, &name, &errMsg);
    if (ret < 0) {
        if (ret == -SQLDB_ERROR_NOT_FOUND) {

            // If admin badge reader, adding the badge
            if (reader->rank == READER_RANK_ADMIN) {
                // Write control data on the badge
                ret = _update_control_data_badge(reader, new_control);
                if (ret < 0) {
                    AFB_API_WARNING(reader->api, "[_endpoint_process] Error [%i] to update data control in badge [%s]", ret, badge_uid);
                    goto OnExit;
                }

                // Adding badge into database
                ret = sqldb_add(user_data->sql_data, badge_uid, new_control, "user", &errMsg);
                if (ret < 0) {
                    AFB_API_ERROR(reader->api, "[_endpoint_process] Error [%s] to adding badge in database", errMsg);
                    free(errMsg);
                    goto OnExit;
                }

                goto OnExit;
            }

            AFB_API_DEBUG(reader->api, "[_endpoint_process] Key %s is not found\n", badge_uid);
            free(errMsg);

            // Push on the bad badge event if valid
            ret = wrap_json_pack(&responseJ, "{ss ss ss}"
                    , "reader", reader->uid
                    , "badge", badge_uid
                    , "error", BAD_BADGE_ERROR_NOT_FOUND);
            if (ret < 0) {
                AFB_API_ERROR(reader->api, "[_endpoint_process] Failed to wrap response JSON !");
                goto OnExit;
            }
            _push_event(reader, responseJ, user_data->event, 1);

            ret = -SQLDB_ERROR_NOT_FOUND;
            goto OnExit;
        }else {
            AFB_API_ERROR(reader->api, "[_endpoint_process] Error [%s] to getting key in database", errMsg);
            free(errMsg);
            goto OnExit;
        }
    } else {

        // If admin badge reader, update the badge
        if (reader->rank == READER_RANK_ADMIN) {
            // Write control data on the badge
            ret = _update_control_data_badge(reader, new_control);
            if (ret < 0) {
                AFB_API_WARNING(reader->api, "[_endpoint_process] Error [%i] to update data control in badge [%s]", ret, badge_uid);
                goto OnExit;
            }

            // Update in database
            ret = _update_control_data_db(user_data->sql_data, badge_uid, new_control);
            if (ret < 0) {
                AFB_API_ERROR(reader->api, "[_endpoint_process] Failed to update database %i !\n", ret);
                goto OnExit;
            }

            goto OnExit;
        }

        // Check if control data is good, if not => badge was cloned
        if (reader->loa <= 1 || !strncmp(control_badge, control, control_length)) {
            if (reader->loa > 1) {
                // Update controle data into badge          
                ret = _update_control_data_badge(reader, new_control);
                if (ret < 0) {
                    AFB_API_WARNING(reader->api, "[_endpoint_process] Error [%i] to update data control in badge [%s]", ret, badge_uid);
                    goto OnExit;
                }

                // Update in database
                ret = _update_control_data_db(user_data->sql_data, badge_uid, new_control);
                if (ret < 0) {
                    AFB_API_ERROR(reader->api, "[_endpoint_process] Failed to update database %i !", ret);
                    goto OnExit;
                }
            }
            // Prepare response
            ret = wrap_json_pack(&responseJ, "{ss ss si}"
                    , "reader", reader->uid
                    , "badge", badge_uid
                    , "loa", reader->loa);
            if (ret < 0) {
                AFB_API_ERROR(reader->api, "[_endpoint_process] Failed to wrap response JSON !");
                goto OnExit;
            }

            // If subscribed on global event send it
            _push_event(reader, responseJ, user_data->event, 0);

        } else {

            AFB_API_DEBUG(reader->api, "[_endpoint_process] Wrong data Control for %s's badge[%s]! Badge has been cloned %s/%s!!!!", name, badge_uid, control_badge, control);
            // Push on the bad badge event if valid
            ret = wrap_json_pack(&responseJ, "{ss ss ss}"
                    , "reader", reader->uid
                    , "badge", badge_uid
                    , "error", BAD_BADGE_ERROR_CLONED);
            if (ret < 0) {
                AFB_API_ERROR(reader->api, "[_endpoint_process] Failed to wrap response JSON !");
                goto OnExit;
            }
            _push_event(reader, responseJ, user_data->event, 1);
        }
    }

OnExit:
    free(key_uid);
    free(control);
    free(name);
    free(new_control);
    return ret;
}

/**
 * _cleanup_handler_auto - called when auto_thread is cancelled
 * 
 * @param arg Arguments passed during the pthread_cleanup_push() - the ctx
 */
static void _cleanup_handler_auto(void *arg) {
    free(arg);
}

void _reader_response_handler(reader_t *reader, uint8_t *uuid, const char * cmd_uid, uint8_t * response_buffer, int response_len) {
    int ret = 0;

    // Event variables
    json_object * responseJ = NULL;

    if (response_len <=0 || !response_buffer) {
        fprintf(stdout, "\nStatus OK !\n");
    } else {
        // Check the command and store in the right variable
        if (!strcmp(cmd_uid, "read-magic-key-1")) {

            if (response_len != reader->magic_key_part1.length) {
                reader->magic_key_part1.key = (char *) realloc(reader->magic_key_part1.key, response_len +1 );
            }
            reader->magic_key_part1.key = (char *) memcpy(reader->magic_key_part1.key, (char *)response_buffer, response_len);
            reader->magic_key_part1.length = (uint8_t) (response_len);

        } else if (!strcmp(cmd_uid, "read-magic-key-2")) {

            if (response_len != reader->magic_key_part2.length) {
                reader->magic_key_part2.key = (char *) realloc(reader->magic_key_part2.key, response_len + 1);
            }
            reader->magic_key_part2.key = (char *) memcpy(reader->magic_key_part2.key, (char *)response_buffer, response_len);
            reader->magic_key_part2.length = (uint8_t) (response_len);

        } else if (!strcmp(cmd_uid, "verify-data-control")) {

            // Convert uuid into string
            char * uuid_str = (char*) calloc(9, sizeof(char));
            ret = snprintf(uuid_str, 9, "%02x%02x%02x%02x", uuid[0], uuid[1], uuid[2], uuid[3]);
            if (ret < 0) {
                AFB_API_ERROR(reader->api, "Error [%i] to convert uuid in string", ret);
                free(uuid_str);
                return;
            }

            // Which badge reader rank call to verify data ?
            if (reader->rank == READER_RANK_ADMIN) {
                // In admin, just end the process in verifying into database
                ret = _endpoint_process(reader, uuid_str, (char *)response_buffer, response_len);
                if (ret < 0) {
                    AFB_API_DEBUG(reader->api, "Error [%i] to finish process\n", ret);
                    free(uuid_str);
                    if (reader->fReader->functions->errorPtr) {
                        ret = reader->fReader->functions->errorPtr(reader);
                        if (ret < 0) {
                            AFB_API_ERROR(reader->api, "Error [%i] to use error pointer function", ret);
                            return;
                        }
                    }
                    return;
                }

                free(uuid_str);
                if ( reader->fReader->functions->activateExtPtr) {
                    ret = reader->fReader->functions->activateExtPtr(reader);
                    if (ret < 0) {
                        AFB_API_ERROR(reader->api, "Error [%i] to activate Command Ext", ret);
                        return;
                    }
                }
                return;

            } else {
                // check if magic key was informed correctly
                size_t magic_length = (size_t) (reader->magic_key_part1.length + reader->magic_key_part2.length + 1);
                char * magic_key =  (char *)calloc(magic_length, sizeof(char));
                for (int idx = 0; idx < reader->magic_key_part1.length ; idx ++) {
                    magic_key[idx] = reader->magic_key_part1.key[idx];
                }
                for (int idx = 0; idx < reader->magic_key_part2.length ; idx ++) {
                    magic_key[reader->magic_key_part1.length +idx] = reader->magic_key_part2.key[idx];
                }
                if (ret < 0) {
                    AFB_API_ERROR(reader->api, " Error to concat 2 keys with err: [%i]", ret);
                    free(magic_key);
                    return;
                }

                AFB_API_DEBUG(reader->api, "Magic key caught: %s", magic_key);

                if (!strncmp(magic_key, MAGIC_KEY, sizeof(MAGIC_KEY) - 1)) {  

                    // Verify if buffer is not all zeros:
                    for (int index = 0; index < response_len; index++) {
                        if (response_buffer[index] == (uint8_t) 0x00) {
                            AFB_API_ERROR(reader->api, "Read an empty NULL control data \n");
                            free(uuid_str);
                            free(magic_key);
                            return;
                        }
                    }

                    // Call endpoint process to finish reader transactions !
                    ret = _endpoint_process(reader, uuid_str, (char *)response_buffer, response_len);
                    if (ret < 0) {
                        AFB_API_DEBUG(reader->api, "Error [%i] to finish process", ret);
                        free(uuid_str);
                        free(magic_key);
                        // Send error ptr if exists in auto mode
                        if (reader->mode == READER_MODE_AUTO) {
                            if ( reader->fReader->functions->errorPtr) {
                                ret = reader->fReader->functions->errorPtr(reader);
                                if (ret < 0) {
                                    AFB_API_ERROR(reader->api, "Error [%i] to send error", ret);
                                    return;
                                }
                            }
                        }
                        return;
                    }
                    free(uuid_str);

                    // Call the Activation Ext function of the plugin in auto mode
                    if (reader->mode == READER_MODE_AUTO) {
                        if ( reader->fReader->functions->activateExtPtr) {
                            ret = reader->fReader->functions->activateExtPtr(reader);
                            if (ret < 0) {
                                AFB_API_ERROR(reader->api, "Error [%i] to activate Command Ext", ret);
                                free(magic_key);
                                return;
                            }
                        }
                    }
                } else {
                    // Push on the bad badge event if valid
                    ret = wrap_json_pack(&responseJ, "{ss ss ss}"
                            , "reader", reader->uid
                            , "badge", uuid_str
                            , "error", BAD_BADGE_ERROR_MAGIC_KEY);
                    if (ret < 0) {
                        AFB_API_ERROR(reader->api, "[response_handler_cb] Failed to wrap response JSON !");
                        free(magic_key);
                        return;
                    }

                    // Catch user data
                    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(reader->api);
                    user_data_t* user_data = (user_data_t *) ctlConfig->external;

                    _push_event(reader, responseJ, user_data->event, 1);

                    // Call the Error function of the plugin in auto mode
                    if (reader->mode == READER_MODE_AUTO) {
                        if (reader->fReader->functions->errorPtr) {
                            ret = reader->fReader->functions->errorPtr(reader);
                            if (ret < 0) {
                                AFB_API_ERROR(reader->api, "Error [%i] to use error pointer function", ret);
                                free(magic_key);
                                return;
                            }
                        }
                    }
                }
                free(magic_key);
            }
        }
    }
}

void _reader_bad_authent_handler(reader_t *reader, uint8_t *uuid) {
    int ret = 0;

    // User data variables
    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(reader->api);
    user_data_t* binding_data = (user_data_t *) ctlConfig->external;

    // Event variables
    json_object * responseJ = NULL;

    // Convert uuid in string
    char * uuid_str = (char*) calloc(9, sizeof(char));
    ret = snprintf(uuid_str, 9, "%02x%02x%02x%02x", uuid[0], uuid[1], uuid[2], uuid[3]);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "Error [%i] to get uuid in string", ret);
        goto OnExit;
    }
    
    // Push on the bad badge event if valid
    ret = wrap_json_pack(&responseJ, "{ss ss ss}"
            , "reader", reader->uid
            , "badge", uuid_str
            , "error", BAD_BADGE_ERROR_KEY);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "[bad_authent_handler_cb] Failed to wrap response JSON !");
        goto OnExit;
    }

    // Call the Error function of the plugin in auto mode
    if (reader->mode == READER_MODE_AUTO) {
        if ( reader->fReader->functions->errorPtr) {
            ret = reader->fReader->functions->errorPtr(reader);
            if (ret < 0) {
                AFB_API_ERROR(reader->api, "Error [%i] to send error Command", ret);
                return;
            }
        }
    }
    // Push event if subscribed on global event
    _push_event(reader, responseJ, binding_data->event, 1);
    
OnExit:
    free(uuid_str);
    return;
}

/** 
 * _reader_state_thread_func - send each second the command groups
 */
static void *_reader_state_thread_func(void *ctx) {
    reader_state_thread_t * thread_ctx = (reader_state_thread_t *) ctx;
    reader_t * reader = thread_ctx->reader_ctx;
    int ret = 0;
    uint8_t * uuid = NULL;
    //uint8_t last_uuid[4] = {0};

    pthread_cleanup_push(_cleanup_handler_auto, ctx);

    while(1) {
        // Check state 
        switch (reader->state)
        {
        case READER_STATE_RESET:
            AFB_API_DEBUG(reader->api, "[_reader_state_thread_func] Reader %s is resetting", reader->uid);
            sleep(10);
            break;

        case READER_STATE_CONNECTING:
            AFB_API_DEBUG(reader->api, "[_reader_state_thread_func] Reader %s is in connecting process", reader->uid);
            sleep(10);
            break;

        case READER_STATE_DISCONNECTED:
            AFB_API_DEBUG(reader->api, "[_reader_state_thread_func] Reader %s is disconnected -> try to (re)connect", reader->uid);
            if (!reader->fReader->functions->connectPtr) {
                AFB_API_ERROR(reader->api, "[_reader_state_thread_func] Missing connection pointer function for reader %s", reader->uid);
                goto OnExit;
            }
            ret = reader->fReader->functions->connectPtr(reader);
            if (ret < 0) {
                AFB_API_ERROR(reader->api, "[_reader_state_thread_func] Error occurred during connection with reader %s (err = %i)", reader->uid, ret);
            }
            sleep(10);
            break;

        case READER_STATE_ERROR:
            reader->action = READER_ACTION_RESET;
            AFB_API_DEBUG(reader->api, "[_reader_state_thread_func] Reader %s is in an error state", reader->uid);
            if (!reader->fReader->functions->onResetPtr) {
                AFB_API_ERROR(reader->api, "[_reader_state_thread_func] Missing reset pointer function for reader %s", reader->uid);
                goto OnExit;
            }
            ret = reader->fReader->functions->onResetPtr(reader);
            if (ret < 0) {
                AFB_API_ERROR(reader->api, "[_reader_state_thread_func] Error occurred on reseting the reader %s (err = %i)", reader->uid, ret);
            } else {
                AFB_API_DEBUG(reader->api, "[_reader_state_thread_func] Reader %s has been resetting", reader->uid);
            }
            sleep(1);
            reader->state = READER_STATE_DISCONNECTED;
            break;

        case READER_STATE_CONNECTED:
            if (reader->action == READER_ACTION_EXT) {
                usleep(500000);
                continue;
            }

            if (!reader->fReader->functions->catchUuidPtr) {
                AFB_API_ERROR(reader->api, "[_reader_state_thread_func] Missing catch uuid pointer function for reader %s", reader->uid);
                goto OnExit;
            }
            
            ret = reader->fReader->functions->catchUuidPtr(reader, &uuid);
            if (ret < 0) {
                if (ret == -READER_ERROR_NOT_SUCCESS || ret == -READER_ERROR_TIMEOUT) {
                    if (thread_ctx->verbose) {
                        AFB_API_DEBUG(reader->api, "--- [%s] No badge\n", reader->uid);
                    }
                    //if (thread_ctx->force != 2) memset(last_uuid, 0, sizeof(last_uuid));
                }else if (ret == -READER_ERROR_DISCONNECTED) {
                    AFB_API_WARNING(reader->api, "--- [%s] Reader Disconnected\n", reader->uid);
                    sleep(1);
                } else {
                    AFB_API_ERROR(reader->api, "--- [%s] Error %i to getting uuid badge\n", reader->uid, ret);
                }
            } else {

                if (reader->loa <= 1) {
                    // User data variables
                    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(reader->api);
                    user_data_t* user_data = (user_data_t *) ctlConfig->external;
                    json_object * responseJ = NULL;

                    // Convert badge_uuid into string
                    char * uuid_str = (char*) calloc(9, sizeof(char));
                    ret = snprintf(uuid_str, 9, "%02x%02x%02x%02x", uuid[0], uuid[1], uuid[2], uuid[3]);
                    if (ret < 0) {
                        AFB_API_ERROR(reader->api, "Error [%i] to convert uuid in string", ret);
                        free(uuid_str);
                        goto OnExit;
                    }

                    if (reader->loa == 0 ) {
                        // Prepare response
                        ret = wrap_json_pack(&responseJ, "{ss ss si}"
                                , "reader", reader->uid
                                , "badge", uuid_str
                                , "loa", 0);
                        if (ret < 0) {
                            free(uuid_str);
                            AFB_API_ERROR(reader->api, "[_reader_state_thread_func] Failed to wrap response JSON !");
                            goto OnExit;
                        }
                        
                        // If subscribed on global event send it
                        _push_event(reader, responseJ, user_data->event, 0);

                        // call success reader process
                        if (reader->mode == READER_MODE_AUTO) {
                            if ( reader->fReader->functions->activateExtPtr) {
                                ret = reader->fReader->functions->activateExtPtr(reader);
                                if (ret < 0) {
                                    AFB_API_ERROR(reader->api, "Error [%i] to activate Command Ext", ret);
                                }
                            }
                        }
                    } else {
                        // Verify the database
                        ret = _endpoint_process(reader, uuid_str, NULL, 0);
                        if (ret < 0) {
                            // call error reader process
                            if (reader->mode == READER_MODE_AUTO) {
                                if (reader->fReader->functions->errorPtr) {
                                    ret = reader->fReader->functions->errorPtr(reader);
                                    if (ret < 0) {
                                        AFB_API_ERROR(reader->api, "Error [%i] to use error pointer function", ret);
                                    }
                                }
                            }
                        } else {
                            // call success reader process
                            if (reader->mode == READER_MODE_AUTO) {
                                if ( reader->fReader->functions->activateExtPtr) {
                                    ret = reader->fReader->functions->activateExtPtr(reader);
                                    if (ret < 0) {
                                        AFB_API_ERROR(reader->api, "Error [%i] to activate Command Ext", ret);
                                    }
                                }
                            }
                        }
                    }
                } else {
                    ret = exec_group_cmd(reader, thread_ctx->config, thread_ctx->group, thread_ctx->force, thread_ctx->verbose, uuid);
                    if (ret < 0) {
                        if (thread_ctx->verbose)  fprintf(stderr, "--- [%s] Error during the group command %i\n", reader->uid, thread_ctx->group);

                        if (ret == -READER_ERROR_AUTHENTICATE) {
                            _reader_bad_authent_handler(reader, uuid);
                        }
                    }
                }

                free(uuid);
            }
            
            usleep(500000);
            break;
        }
    }

OnExit:
    pthread_cleanup_pop(1);
    return NULL;
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PUBLIC FUNCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

binding_cmd_key_t *reader_parse_cmd_key (json_object *configJ) {
    int ret = 0;
    binding_cmd_key_t *config = calloc(1, sizeof(binding_cmd_key_t));
    json_object *cmdsJ=NULL, *keysJ=NULL;

    ret = wrap_json_unpack(configJ, "{s?o s?o}"
        , "keys", &keysJ
        , "cmds", &cmdsJ
    );
    if (ret < 0) {
        AFB_API_ERROR(afbBindingRoot, "[reader_parse_cmd_key] json unsupported key");
        goto OnErrorExit;
    }
    
    if (keysJ && !cmdsJ) {
        AFB_API_ERROR(afbBindingRoot, "[reader_parse_cmd_key] key 'cmds' mandatory when 'keys' present");
        goto OnErrorExit;
    }
    
    // Parse Keys
    size_t keyCount = 0;
    switch (json_object_get_type(keysJ)) {
        case json_type_object:
            config->keys = calloc(1, sizeof(reader_key_t));
            config->keys_nbr = 1;
            ret = _reader_parse_one_key(keysJ, &config->keys[0]);
            if (ret < 0) goto OnErrorExit;
            break;

        case json_type_array:
            keyCount = json_object_array_length(keysJ);
            config->keys_nbr = keyCount;
            config->keys = calloc(keyCount, sizeof(reader_key_t));
            for (int index = 0; index < keyCount; index ++) {
                json_object *keyJ = json_object_array_get_idx(keysJ, index);
                ret = _reader_parse_one_key(keyJ, &config->keys[index]);
                if (ret < 0) goto OnErrorExit;
            }
            break;

        case json_type_null:
            config->keys_nbr = 0;
            break;

        default:
            AFB_API_ERROR(afbBindingRoot, "[reader_parse_cmd_key] keys should be a json array of a json object");
            goto OnErrorExit;

    }

    // Parse Commands
    size_t cmdCount = 0;
    switch (json_object_get_type (cmdsJ)) {

        case json_type_object:
            config->cmds= calloc (1, sizeof(reader_cmd_t));
            config->cmds_nbr = 1;
            ret = _reader_parse_one_cmd (config, cmdsJ, &config->cmds[0]);
            if (ret < 0) goto OnErrorExit;
            break;

        case json_type_array:
            cmdCount= json_object_array_length (cmdsJ);
            config->cmds= calloc (cmdCount, sizeof(reader_cmd_t));
            config->cmds_nbr = cmdCount;
            for (int index=0; index < cmdCount; index++) {
                json_object *cmdJ= json_object_array_get_idx (cmdsJ, index);
                ret = _reader_parse_one_cmd (config, cmdJ, &config->cmds[index]);
                if (ret < 0) goto OnErrorExit;
            }
            break;

        case json_type_null:
            config->cmds_nbr = 0;
            break;

        default:
            AFB_API_ERROR(afbBindingRoot, "[reader_parse_cmd_key] cmds should be json object or array of object");
            goto OnErrorExit;
    }
    return config;

OnErrorExit:
    free (config);
    return NULL;
}

int reader_exec_one_cmd(reader_t *reader, const reader_cmd_t *cmd, uint8_t *badge_uuid) {
    int ret = 0;

    reader->action = cmd->action;
    uint8_t *response = NULL;
    
    switch (cmd->action) {

        case READER_ACTION_READ:
            response = (uint8_t *) calloc(cmd->dlen, sizeof(uint8_t));

            if (!reader->fReader->functions->readBadgePtr) {
                AFB_API_ERROR(reader->api, "[reader_exec_one_cmd] Missing pointer to read function for reader %s", reader->uid);
                free(response);
                ret = -1;
                break;
            }
            ret = reader->fReader->functions->readBadgePtr(reader, cmd->key, cmd->sec, cmd->blk, (uint8_t) cmd->dlen, response);
            if ( ret < 0) {
                AFB_API_WARNING(reader->api, "[reader_exec_one_cmd] Failed to read with reader %s", reader->uid);
                free(response);
                break;
            }
            _reader_response_handler(reader, badge_uuid, cmd->uid, response, (uint8_t)cmd->dlen);
            free(response);
            break;

        case READER_ACTION_WRITE:
            if (!reader->fReader->functions->writeBadgePtr) {
                AFB_API_ERROR(reader->api, "[reader_exec_one_cmd] Missing pointer to write function for reader %s", reader->uid);
                ret = -1;
                break;
            }
            ret = reader->fReader->functions->writeBadgePtr(reader, cmd->key, cmd->sec, cmd->blk, (uint8_t) cmd->dlen, cmd->data);
            if ( ret < 0) {
                AFB_API_WARNING(reader->api, "[reader_exec_one_cmd] Failed to write with reader %s", reader->uid);
                break;
            }
            break;
        case READER_ACTION_RESET:
            if (!reader->fReader->functions->onResetPtr) {
                AFB_API_ERROR(reader->api, "[reader_exec_one_cmd] Missing pointer to reset function for reader %s", reader->uid);
                ret = -1;
                break;
            }
            ret = reader->fReader->functions->onResetPtr(reader);
            if ( ret < 0) {
                AFB_API_WARNING(reader->api, "[reader_exec_one_cmd] Failed to write with reader %s", reader->uid);
                break;
            }
            break;

        case READER_ACTION_TRAILER:
            if (!reader->fReader->functions->trailerBadgePtr) {
                AFB_API_ERROR(reader->api, "[reader_exec_one_cmd] Missing pointer to trailer function for reader %s", reader->uid);
                ret = -1;
                break;
            }

            if (!cmd->key) {
                uint8_t defaultKey[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
                const reader_key_t key = {
                    .idx = 0,
                    .len = 6,
                    .uid = "default-A",
                    .value = defaultKey,
                };
                ret =reader->fReader->functions->trailerBadgePtr(reader, &key, cmd->sec, cmd->trailer);
            } else {
                ret = reader->fReader->functions->trailerBadgePtr(reader, cmd->key, cmd->sec, cmd->trailer);
            }
            break;

        case READER_ACTION_LED:
            // TODO
            break;

        default:
            ret = -1;
            break;
    }
    return ret;
}

int exec_group_cmd (reader_t * reader, binding_cmd_key_t * config, int group, int force, int verbose, uint8_t *badge_uuid) {
    int jump=0;
    int ret = 0;


    // loop on defined commands
    for (int idx=0; idx < config->cmds_nbr; idx++) {
        const reader_cmd_t *cmd= &config->cmds[idx];

        if (group <= cmd->group*-1  || group == cmd->group) {
            jump=1;
            ret = reader_exec_one_cmd(reader, cmd, badge_uuid);
            if (ret < 0) {
                if (verbose)
                    fprintf (stderr, " -- Fail Executing command uid=%s error= [%i]\n", cmd->uid, ret);
                if (force <= 0) {
                    usleep(1000);
                    goto OnErrorExit;
                }
            }
        } else {
            if (verbose) {
                if (jump) {
                    fprintf (stdout, "\n");
                    jump=0;
                }
                fprintf (stdout, " -- Ignoring cmd=%s group=%d\n", cmd->uid, cmd->group);
            }
        }
    }
    //reader->authenticate.isAuthenticate = false;
    if (verbose) {
        fprintf (stdout,"\t** OK: Cmds/group=%d [done]\n\n", group);
        fprintf(stderr,"\t?? Insert new badges ??\n\n");
    }

    return 0;

OnErrorExit:
    // reader->authenticate.isAuthenticate = false;
    return ret;
}


int reader_monitor_state (reader_t *reader, int group, int force, int verbosity) {
    int ret = 0;

    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(reader->api);
    user_data_t * user_data = (user_data_t *) ctlConfig->external;

    reader_state_thread_t *threadCtx= calloc (1, sizeof(reader_state_thread_t));
    threadCtx->reader_ctx = reader;
    threadCtx->config = user_data->cmd_keys;
    threadCtx->group = group;
    threadCtx->force = force;
    threadCtx->verbose = verbosity;

    AFB_API_DEBUG(reader->api, "Startin new monitoring thread for reader %s", reader->uid);

    ret = pthread_create (&reader->thread_auto_id, NULL, _reader_state_thread_func, (void*) threadCtx);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "[reader_monitor_state] Fail to create state thread for reader=%s. (err=%s)", reader->uid, strerror(errno)) ;
        return ret;
    }
    return 0;
}

int reader_monitor_wait (reader_t *reader) {
    return pthread_join(reader->thread_auto_id, NULL); // infinit wait for thread auto to quit
}

void reader_free_config(binding_cmd_key_t *config) {
    if (config) {
        if (config->keys_nbr > 0) {
            for (int index = 0; index < config->keys_nbr; index++) {
                free(config->keys[index].value);
            }
            free(config->keys);
        }
        if (config->cmds_nbr > 0) {
            for (int index = 0; index < config->cmds_nbr; index++) {
                if (config->cmds[index].trailer) {
                    free(config->cmds[index].trailer->acls);
                    free(config->cmds[index].trailer);
                }
                free(config->cmds[index].data);
            }
            free(config->cmds);
        }
        free (config);
    }
}