/*
 * Copyright 2020 - 2021 IoT.bzh
 *
 * author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#define _GNU_SOURCE

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "badge-reader-binding.h"

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * _parse_args_cmd - Parse command arguments from a json object
 * 
 * @param commands struct with commands and keys configured fr the binding
 * @param inputJ JSON where to search command's informations
 * @param argsCmd Arguments of the command parsed, need to be allocated
 * @return 0 in success <0 otherwise
 */
static int _parse_args_cmd(binding_cmd_key_t *commands, json_object *inputJ, reader_args_cmd_t *argsCmd) {

    const char *keyUID = NULL;
    json_object *dataJ = NULL;

    if (!inputJ || !argsCmd) {
        return -READER_ERROR_NULL_ARGS;
    }

    // Parse the inputJ for the command
    int ret = wrap_json_unpack (inputJ, "{s?i,s?i,s?i,s?s,s?o !}"
        ,"sec", &argsCmd->sec
        ,"blk", &argsCmd->blk
        ,"len", &argsCmd->dlen
        ,"key", &keyUID
        ,"data", &dataJ
    );
    if (ret < 0) {
        return -READER_ERROR_WRONG_ARGS;
    }

    // Check if data is given and store it
    if (dataJ != NULL) {
        switch (json_object_get_type (dataJ)) {
            const char *byteS, *dataS;
            u_int8_t *valueB;
            int err, byte;
            size_t count;

            case json_type_string:
                dataS= json_object_get_string (dataJ);
                // if no dataLength then use data string len
                if (argsCmd->dlen == 0) {
                    argsCmd->dlen= strlen(dataS);
                    argsCmd->data=(uint8_t *) strdup(dataS);
                } else {
                    argsCmd->data = (uint8_t *) malloc (argsCmd->dlen * sizeof(uint8_t));
                    strncpy ((char*) argsCmd->data, dataS, argsCmd->dlen);
                }
                break;

            case json_type_array:
                count = json_object_array_length(dataJ);
                valueB= calloc (count+1, sizeof(u_int8_t));

                for (int idx=0; idx < count; idx++) {
                    byteS= json_object_get_string (json_object_array_get_idx(dataJ, idx));
                    err= sscanf (byteS, "0x%2x", &byte);
                    if (err <0 || byte > 255) {
                        free(valueB);
                        return -READER_ERROR_WRONG_ARGS;
                    }
                    valueB[idx]= (u_int8_t)byte;
                }

                argsCmd->dlen = count;
                argsCmd->data = valueB;
                break;

            default:
                return -READER_ERROR_NO_DATA;

        }
    }

    // Check if key is defined
    if (keyUID != NULL) {
        if (commands->keys) {
            for (int idx=0; commands->keys[idx].uid; idx++) {
                if (!strcasecmp ( commands->keys[idx].uid, keyUID)) {
                    argsCmd->key= &commands->keys[idx];
                    break;
                }
            }
        }
    }

    return 0;
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PUBLIC FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * _stateToString - return name of a specific status
 * 
 * @param state one of reader_state_e
 * @param command Command stringiffy, NEED TO BE FREE BY THE USER
 * @return status of asprintf
 */
int stateToString(reader_state_e state, char **stateStr) {
    int ret = 0;
    switch (state)
    {
    default:
    case READER_STATE_DISCONNECTED:
        ret = asprintf(stateStr, "disconnected");
        break;
    case READER_STATE_CONNECTED:
        ret = asprintf(stateStr, "connected");
        break;
    case READER_STATE_ERROR:
        ret = asprintf(stateStr, "error");
        break;
    case READER_STATE_RESET:
        ret = asprintf(stateStr, "reset");
        break;
    case READER_STATE_CONNECTING:
        ret = asprintf(stateStr, "connecting ...");
    }
    return ret;
}

/**
 * FamilyReaderReadersId - Create Json in respons to the info verb of the api of the family reader
 * 
 * @param fReader ptr of FamilyReader
 * @param verbose level of the verbosity
 * @param responsJ Response in Json to use
 */
void FamilyReaderReadersId(family_reader_t * fReader, int verbose, json_object * responseJ) {
    json_object *elemJ = NULL;
    json_object *actionsJ = NULL;
    json_object *dataJ = NULL;
    json_object *usageJ = NULL;
    json_object *sampleJ = NULL;
    json_object *sampleWriteJ=NULL;
    reader_t *reader;
    int err = 0;
    char verb_uid[64];

    // Loop on every Reader
    for (int index = 0; fReader->readers[index].uid; index++) {
        reader = &fReader->readers[index];
        sprintf(verb_uid, "%s/%s", fReader->uid, reader->uid);
        switch (verbose) {
            default:
            case 1 :
                err += wrap_json_pack(&elemJ, "{ss ss ss}", 
                        "uid", reader->uid,
                        "family", fReader->uid,
                        "info", reader->info);
                break;
            case 2 :
                err += wrap_json_pack(&elemJ, "{ss ss ss ss}", 
                        "uid", reader->uid,
                        "family", fReader->uid,
                        "info", reader->info,
                        "verb", verb_uid);
                break;
            case 3 :
                sampleJ = json_object_new_array();
                err += wrap_json_pack(&dataJ, "{si si si ss ss}", "sec", 11, "blk", 0, "len", 4,"key", "key-a", "data", "test");
                err += wrap_json_pack(&sampleWriteJ, "{ss so}", "action", "write", "data", dataJ);
                dataJ = json_object_get(dataJ);
                err += json_object_array_add(sampleJ, sampleWriteJ);
                err += wrap_json_pack(&actionsJ, "[s s s s s s s s]", "write", "read", "activate", "error", "subscribe", "unsubscribe", "state", "reset");
                err += wrap_json_pack (&usageJ, "{so, so}", "action", actionsJ, "data", dataJ);
                err += wrap_json_pack(&elemJ, "{ss ss ss ss so* so*}", 
                        "uid", reader->uid,
                        "family", fReader->uid,
                        "info", reader->info,
                        "verb", verb_uid,
                        "usage", usageJ,
                        "sample", sampleJ);
                break;
        }

        if (err) {
           AFB_DEBUG ("FamilyReaderReadersId: Fail to wrap json Family Reader info family=%s reader=%s", fReader->uid, reader->uid);
        } else {
           json_object_array_add (responseJ, elemJ);
        }
    }
}

/**
 * FamilyReaderRequest - Callback to the api/verb of a family reader
 * 
 * @param request the api request
 * @param fReader ptr of FamilyReader
 * @param queryJ The query in Json
 */
void FamilyReaderRequest(afb_req_t request, family_reader_t * fReader, json_object *queryJ) {
    int ret;
    const char *action;
    const char *uri=NULL;
    int verbose=0;
    json_object *responseJ=NULL;

    ret= wrap_json_unpack(queryJ, "{ss s?s s?i !}", "action", &action, "verbose", &verbose, "uri", &uri);
    if (ret < 0) {
        afb_req_fail_f (request, "FamilyReaderAdmin", "invalid query FamilyReader=%s query=%s", fReader->uid, json_object_get_string(queryJ));
        return;
    }

    if (!strcasecmp(action, "connect")) {
        AFB_NOTICE("WILL BE AVAILABLE SOON");
    } else if (!strcasecmp(action, "disconnect")) {
        AFB_NOTICE("WILL BE AVAILABLE SOON");
    } else if (!strcasecmp(action, "info")){
        AFB_NOTICE("WILL BE AVAILABLE SOON");
    } else if (!strcasecmp(action, "list")){
        AFB_NOTICE("WILL BE AVAILABLE SOON");
    }
    afb_req_success(request, responseJ, NULL);
    return;
}

/**
 * ReaderEventCreate - Call when subscribe is called, to add Callbacks functions 
 * with EPOLL event from reader sockets
 * 
 * @param reader ptr of Reader
 */
static int ReaderEventCreate(reader_t * reader) {
    family_reader_t * fReader = reader->fReader;

    // If no event create one an attach it to the reader
    if (!afb_event_is_valid(reader->event)) {
        reader->event = afb_api_make_event(reader->api, reader->apiverb);
        if (!reader->event) {
            AFB_API_ERROR(reader->api, "ReaderEventCreate: fail to create event Family=%s reader=%s",
            fReader->uid, reader->uid);
            return 1;
        }
    }
    return 0;
}

/**
 * ReaderRequest - Callback to the api/verb of a Reader
 * 
 * @param request the api request
 * @param fReader ptr of FamilyReader
 * @param queryJ The query in Json
 */
void ReaderRequest(afb_req_t request, reader_t * reader, json_object *queryJ) {
    assert(reader);
    assert(queryJ);
    
    int ret = 0;
    family_reader_t * fReader =  reader->fReader;
    const char * action;
    json_object * dataJ = NULL, * responseJ = NULL;

    CtlConfigT* ctlConfig = (CtlConfigT*)afb_api_get_userdata(afb_req_get_api(request));
    user_data_t * user_data = (user_data_t *) ctlConfig->external;

    if (json_object_get_type(queryJ) != json_type_object) {
        AFB_API_ERROR(reader->api, "[ReaderRequest] Query is in wrong type reader %s/%s", fReader->uid, reader->uid);
        afb_req_fail_f(request, "query-error", "ReaderRequest: invalid type of the query for reader=%s/%s"
            , fReader->uid, reader->uid);
        return;
    }
    ret= wrap_json_unpack(queryJ, "{ss s?o !}", "action", &action, "data", &dataJ);
    if (ret < 0) {
        afb_req_fail_f(request, "query-error", "ReaderRequest: invalid 'json' FamilyReader=%s reader=%s"
            , fReader->uid, reader->uid);
        return;
    }

    if (!strcasecmp(action, "WRITE")) {

        if (reader->state == READER_STATE_DISCONNECTED) goto OnDisconnectedError;
        if (!fReader->functions->writeBadgePtr) goto OnNoPtrFunction;

        // Parse command args
        reader_args_cmd_t *cmdArgs = (reader_args_cmd_t *) calloc(1, sizeof(reader_args_cmd_t));
        ret = _parse_args_cmd(user_data->cmd_keys, dataJ, cmdArgs);
        if (ret < 0) {
            free(cmdArgs);
            goto OnErrorExit;
        }

        // call plugin func point
        ret = (fReader->functions->writeBadgePtr) (reader, cmdArgs->key, cmdArgs->sec, cmdArgs->blk, (uint8_t) cmdArgs->dlen, cmdArgs->data);
        free(cmdArgs);
        if (ret < 0)  goto OnErrorExit;

    } else if (!strcasecmp(action, "READ")) {
        
        if (reader->state == READER_STATE_DISCONNECTED) goto OnDisconnectedError;
        if (!fReader->functions->readBadgePtr) goto OnNoPtrFunction;
        if (!dataJ) {
            goto OnErrorExit;
        }

        // Parse command args
        reader_args_cmd_t *cmdArgs = (reader_args_cmd_t *) calloc(1, sizeof(reader_args_cmd_t));
        ret = _parse_args_cmd(user_data->cmd_keys, dataJ, cmdArgs);
        if (ret < 0) {
            free(cmdArgs);
            goto OnErrorExit;
        }

        // call plugin func point
        uint8_t * response = (uint8_t *) calloc(cmdArgs->dlen, sizeof(uint8_t));
        ret = (fReader->functions->readBadgePtr) (reader, cmdArgs->key, cmdArgs->sec, cmdArgs->blk, (uint8_t) cmdArgs->dlen, response);
        free(cmdArgs);
        if (ret < 0) goto OnErrorExit;
        ret = wrap_json_pack(&responseJ, "{s?y}"
                , "data", response);
        free(response);
        if (ret < 0) {
            afb_req_fail_f(request, "action error","Failed to apply action %s, unable to construct the JSON response for %s/%s",
                action, fReader->uid, reader->uid);
            return;
        }

    } else if (!strcasecmp(action, "ACTIVATE")) {

        if (reader->state == READER_STATE_DISCONNECTED) goto OnDisconnectedError;
        if (!fReader->functions->activateExtPtr) goto OnNoPtrFunction;
        ret = (fReader->functions->activateExtPtr) (reader);
        if (ret < 0) goto OnErrorExit;

    } else if (!strcasecmp(action, "ERROR")) {

        if (reader->state == READER_STATE_DISCONNECTED) goto OnNoPtrFunction;
        if (!fReader->functions->errorPtr) goto OnNoPtrFunction;
        ret = (fReader->functions->errorPtr) (reader);
        if (ret < 0) goto OnErrorExit;

    } else if (!strcasecmp(action, "RESET")) {

        if (reader->state == READER_STATE_DISCONNECTED) goto OnNoPtrFunction;
        if (!fReader->functions->onResetPtr) goto OnNoPtrFunction;
        ret = (fReader->functions->onResetPtr) (reader);
        if (ret < 0) goto OnErrorExit;

    } else if (!strcasecmp(action, "SUBSCRIBE")) {

        if (reader->state == READER_STATE_DISCONNECTED) goto OnDisconnectedError;
        if(!fReader->functions->subscribePtr) goto OnNoPtrFunction;
        ret = (fReader->functions->subscribePtr) (reader);
        AFB_API_DEBUG(reader->api, "Subscribing has processed!");
        if (ret < 0) goto OnErrorExit;
        ret = ReaderEventCreate(reader);
        if (ret < 0) goto OnErrorExit;
        ret = afb_req_subscribe(request, reader->event);
        if (ret < 0) goto OnErrorExit;

    } else if (!strcasecmp(action, "UNSUBSCRIBE")) {

        if (reader->state == READER_STATE_DISCONNECTED) goto OnDisconnectedError;
        if (reader->event) {
            ret = (fReader->functions->unSubscribePtr) (reader);
            if (ret < 0) goto OnErrorExit;
            ret = afb_req_unsubscribe(request, reader->event);
            if (ret < 0) goto OnErrorExit;
        }

    } else if (!strcasecmp(action, "STATE")) {
        char* stateStr;
        ret = stateToString(reader->state, &stateStr);
        if (ret < 0) {
            AFB_API_ERROR(reader->api, "[Action %s] Error to parse state of %s/%s", 
                action, fReader->uid, reader->uid);
            afb_req_fail_f(request, "action error","Error to parse state of reader %s/%s",
                fReader->uid, reader->uid);
            return ;
        }

        ret = wrap_json_pack(&responseJ, "{s?s}", "state", stateStr);
        if (ret < 0) {
            afb_req_fail_f(request, "state error", "ReaderRequest: fail to wrap state FamilyReader=%s reader=%s",
                fReader->uid, reader->uid);
            free(stateStr);
            return;
        }
        free(stateStr);

    } else {
        AFB_API_ERROR(reader->api, "[Action %s] Unknown action for reader %s/%s", 
            action, fReader->uid, reader->uid);
        afb_req_fail_f(request, "action error","This action %s is unknown for reader %s/%s",
            action, fReader->uid, reader->uid);
        return ;
    }

    afb_req_success(request, responseJ, NULL);
    return;

OnDisconnectedError:
    AFB_API_ERROR(reader->api, "[Action %s] Reader %s/%s is disconnected", 
        action, fReader->uid, reader->uid);
    afb_req_fail_f(request, "state error","Cannot apply the action %s for disconnected reader %s/%s",
        action, fReader->uid, reader->uid);
    return;

OnNoPtrFunction:
    AFB_API_ERROR(reader->api, "[Action %s] Reader %s/%s doesn't have this functionality",
        action, fReader->uid, reader->uid);

OnErrorExit:
    afb_req_fail_f(request, "action error","Failed to apply action %s for reader %s/%s",
        action, fReader->uid, reader->uid);
    return;
}