/*
 * Copyright (C) 2020 - 2021 "IoT.bzh"
 *
 * Author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "cn56-core.h"

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            STRUCT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/*
 * Struct: data
 * ------------
 * 
 *  Represent the data field to be send
 * 
 */
typedef struct frame_data_command_s
{
    int size;
    uint8_t cmd;
    uint8_t time;
    uint8_t data[MAX_CN56_BUFFER_LENGTH];
}frame_data_command_t;

typedef struct data_frame_s
{
    uint8_t *buffer;
    int size;
} data_frame_t;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

pthread_cond_t cnd_disconnect = PTHREAD_COND_INITIALIZER;

/**
 * _add_config - Fill in the data tram to configure the Wiegand option
 * 
 * @param frame_data_command Representation of a data to be send
 * @param cn56_config_wiegand configuration option for wiegand
 */
static void _add_config(frame_data_command_t * frame_data_command, cn56_config_wiegand_t * cn56_config_wiegand) {
    frame_data_command->data[0] = 0x00; /* Wiegand_26 */
    frame_data_command->data[1] = cn56_config_wiegand->block_to_read; /* Indicate the Block Number to Read */
    frame_data_command->data[2] = cn56_config_wiegand->request_idle? 0x26 : 0x52;     /* request_idle */
    frame_data_command->data[3] = (uint8_t) (cn56_config_wiegand->open_wiegand<<5 | 
                                    cn56_config_wiegand->reader_enable<<4 |
                                    cn56_config_wiegand->beep_led_enable<<1 |
                                    cn56_config_wiegand->led_buzzer_controle);
    frame_data_command->data[4] = 0x55;
    frame_data_command->data[5] = 0xaa;
    frame_data_command->data[6] = 0x01;
    frame_data_command->data[7] = 0x06; //Card Type
    frame_data_command->data[8] = 0x00;
    frame_data_command->data[9] = 0x00; /* Mifare 1(0x00) ; Mifare Plus(0x01) ; DESFrie EV1(0x02) ; NFC_T2T(0x03) ;  NFC_T4T(0x04) */
    frame_data_command->data[10] = 0x00;
    frame_data_command->data[11] = 0x00;
    frame_data_command->data[12] = 0x00;
}

/**
 * _create_data_command - Create the data buffer to be send to the reader badges
 * 
 * @param frame_data_command Representation of a data to be send
 * @param cmd The command to send
 * @param cn56_add_config configuration option if the cmd need specific data
 */
static void _create_data_command (frame_data_command_t * frame_data_command, cn56_command_t cmd, void * cn56_add_config) {
    frame_data_command->time = 0x00;
    size_t temp_size = 0;
    switch (cmd)
    {
    case cn56_cmd_getFirmware:
        frame_data_command->size = 0;
        frame_data_command->cmd = 0x0a;
        break;
    case cn56_cmd_activateGreenLed:
        frame_data_command->size = 1;
        frame_data_command->cmd = 0x24;
        frame_data_command->data[0] = 0x02;
        break;
    case cn56_cmd_activateRedLed:
        frame_data_command->size = 1;
        frame_data_command->cmd = 0x24;
        frame_data_command->data[0] = 0x01;
        break;
    case cn56_cmd_activateBothLed:
        frame_data_command->size = 1;
        frame_data_command->cmd = 0x24;
        frame_data_command->data[0] = 0x03;
        break;
    case cn56_cmd_turnOffLed:
        frame_data_command->size = 1;
        frame_data_command->cmd = 0x24;
        frame_data_command->data[0] = 0x00;
        break;
    case cn56_cmd_configWiegand:
        frame_data_command->size = 13;
        frame_data_command->cmd = 0x18;
        _add_config(frame_data_command, (cn56_config_wiegand_t *) cn56_add_config);
        break;
    case cn56_cmd_runWiegandMode:
        frame_data_command->size = 0;
        frame_data_command->cmd = 0x1d;
        break;
    case cn56_cmd_autoRead:
        frame_data_command->size = 1;
        frame_data_command->cmd = 0x12;
        frame_data_command->data[0] = (((cn56_config_wiegand_t *) cn56_add_config)->set_wiegand == 0x01)? 0x12 : 0x13;
        break;
    case cn56_cmd_writeMifare:
        frame_data_command->size = 23;
        frame_data_command->cmd = 0x91;
        frame_data_command->data[0] = 0x01;
        frame_data_command->data[1] = 0x01;
        frame_data_command->data[2] = ((cn56_config_write_t *) cn56_add_config)->blk; // N° blk
        frame_data_command->data[3] = ((cn56_config_write_t *) cn56_add_config)->uid[0];
        frame_data_command->data[4] = ((cn56_config_write_t *) cn56_add_config)->uid[1];
        frame_data_command->data[5] = ((cn56_config_write_t *) cn56_add_config)->uid[2];
        frame_data_command->data[6] = ((cn56_config_write_t *) cn56_add_config)->uid[3];
        memcpy(frame_data_command->data + 7, ((cn56_config_write_t *) cn56_add_config)->data, sizeof(((cn56_config_write_t *) cn56_add_config)->data));
        break;
    case cn56_cmd_readMifare:
        frame_data_command->size = 7;
        frame_data_command->cmd = 0x90;
        frame_data_command->data[0] = 0x01;
        frame_data_command->data[1] = 0x01;
        frame_data_command->data[2] = ((cn56_config_read_t *) cn56_add_config)->blk;
        frame_data_command->data[3] = ((cn56_config_read_t *) cn56_add_config)->uid[0];
        frame_data_command->data[4] = ((cn56_config_read_t *) cn56_add_config)->uid[1];
        frame_data_command->data[5] = ((cn56_config_read_t *) cn56_add_config)->uid[2];
        frame_data_command->data[6] = ((cn56_config_read_t *) cn56_add_config)->uid[3];
        break;
    case cn56_cmd_Cext:
        frame_data_command->size = 1;
        frame_data_command->cmd = 0x29;
        frame_data_command->data[0] = ((cn56_config_delay_t *) cn56_add_config)->delay;
        break;
    case cn56_cmd_setIP:
        frame_data_command->size = 12;
        frame_data_command->cmd = 0x2c;
        frame_data_command->data[0] = ((cn56_config_ip_t *) cn56_add_config)->ip_1;
        frame_data_command->data[1] = ((cn56_config_ip_t *) cn56_add_config)->ip_2;
        frame_data_command->data[2] = ((cn56_config_ip_t *) cn56_add_config)->ip_3;
        frame_data_command->data[3] = ((cn56_config_ip_t *) cn56_add_config)->ip_4;
        frame_data_command->data[4] = 0xff;
        frame_data_command->data[5] = 0xff;
        frame_data_command->data[6] = 0xff;
        frame_data_command->data[7] = 0x00;
        frame_data_command->data[8] = 0xc0;
        frame_data_command->data[9] = 0xa8;
        frame_data_command->data[10] = 0x01;
        frame_data_command->data[11] = 0x64;
        break;
    case cn56_cmd_reset:
        frame_data_command->size = 0;
        frame_data_command->cmd = 0x19;
        break;
    case cn56_cmd_storeKey:
        frame_data_command->size = 8;
        frame_data_command->cmd = 0x47;
        frame_data_command->data[0] = ((cn56_config_load_key_t *)cn56_add_config)->key; // 0x60 = KeyA 0x61 = KeyB
        frame_data_command->data[1] = ((cn56_config_load_key_t *)cn56_add_config)->sector; // N° sector
        frame_data_command->data[2] = ((cn56_config_load_key_t *)cn56_add_config)->value[0];
        frame_data_command->data[3] = ((cn56_config_load_key_t *)cn56_add_config)->value[1];
        frame_data_command->data[4] = ((cn56_config_load_key_t *)cn56_add_config)->value[2];
        frame_data_command->data[5] = ((cn56_config_load_key_t *)cn56_add_config)->value[3];
        frame_data_command->data[6] = ((cn56_config_load_key_t *)cn56_add_config)->value[4];
        frame_data_command->data[7] = ((cn56_config_load_key_t *)cn56_add_config)->value[5];
        break;
    case cn56_cmd_readBlk:
        frame_data_command->size = 2;
        frame_data_command->cmd = 0x41;
        frame_data_command->data[0] = ((cn56_config_read_t *)cn56_add_config)->blk;
        frame_data_command->data[1] = 0x01;
        break;
    case cn56_cmd_modifKey:
        frame_data_command->size = 18;
        frame_data_command->cmd = 0x42;
        frame_data_command->data[0] = (uint8_t)((((cn56_config_modif_key_t *)cn56_add_config)->sector *4) + 3);
        frame_data_command->data[1] = 0x01;
        temp_size = sizeof(((cn56_config_modif_key_t *)cn56_add_config)->keyA);
        memcpy(frame_data_command->data + 2, ((cn56_config_modif_key_t *)cn56_add_config)->keyA, temp_size);
        //temp_size += sizeof(((cn56_config_modif_key_t *)cn56_add_config)->acls);
        memcpy(frame_data_command->data + 2 + temp_size, ((cn56_config_modif_key_t *)cn56_add_config)->acls, sizeof(((cn56_config_modif_key_t *)cn56_add_config)->acls));
        temp_size += sizeof(((cn56_config_modif_key_t *)cn56_add_config)->acls);
        memcpy(frame_data_command->data + 2 + temp_size, ((cn56_config_modif_key_t *)cn56_add_config)->keyB, sizeof(((cn56_config_modif_key_t *)cn56_add_config)->keyB));
        break;
    case cn56_cmd_writeBlk:
        frame_data_command->size = 18;
        frame_data_command->cmd = 0x42;
        frame_data_command->data[0] = ((cn56_config_write_block_t *)cn56_add_config)->blk;
        frame_data_command->data[1] = 0x01;
        memcpy(frame_data_command->data + 2, ((cn56_config_write_block_t *)cn56_add_config)->data, CN56_MAX_WRITE);

        break;
    case cn56_cmd_pn_inListPassive:
        frame_data_command->size = 3;
        frame_data_command->cmd = 0x3d;
        frame_data_command->data[0] = 0x00;
        frame_data_command->data[1] = 0x02;
        frame_data_command->data[2] = 0x00;
        break;
    case cn56_cmd_pn_selectTarget:
        frame_data_command->size = 1;
        frame_data_command->cmd = 0x3e;
        frame_data_command->data[0] = 0x01;
        break;
    case cn56_cmd_loadKey:
        frame_data_command->size = 6;
        frame_data_command->cmd = 0x45;
        frame_data_command->data[0] = ((cn56_config_load_key_t *)cn56_add_config)->value[0];
        frame_data_command->data[1] = ((cn56_config_load_key_t *)cn56_add_config)->value[1];
        frame_data_command->data[2] = ((cn56_config_load_key_t *)cn56_add_config)->value[2];
        frame_data_command->data[3] = ((cn56_config_load_key_t *)cn56_add_config)->value[3];
        frame_data_command->data[4] = ((cn56_config_load_key_t *)cn56_add_config)->value[4];
        frame_data_command->data[5] = ((cn56_config_load_key_t *)cn56_add_config)->value[5];
        break;
    case cn56_cmd_authenticate:
        frame_data_command->size = 6;
        frame_data_command->cmd = 0x40;
        frame_data_command->data[0] = ((cn56_config_authenticate_t *)cn56_add_config)->key;
        frame_data_command->data[1] = ((cn56_config_authenticate_t *)cn56_add_config)->uid[0];
        frame_data_command->data[2] = ((cn56_config_authenticate_t *)cn56_add_config)->uid[1];
        frame_data_command->data[3] = ((cn56_config_authenticate_t *)cn56_add_config)->uid[2];
        frame_data_command->data[4] = ((cn56_config_authenticate_t *)cn56_add_config)->uid[3];
        frame_data_command->data[5] = (uint8_t)(((cn56_config_authenticate_t *)cn56_add_config)->sec * 4);
        break;
    default:
        printf("Not available Command\n");
        frame_data_command->size = 1;
        frame_data_command->cmd = 0x00;
        frame_data_command->data[0] = 0x00;
        break;
    }
}

/**
 * _init_frame_command - Allocate enough memory to the frame thanks to the data to be  send
 *  Initiate static fields: STX, DADD, ETX And data
 * 
 * @param frame which will be send to the reader
 * @param frame_data data to add to the frame
 */
static void _init_frame_command (cn56_frame_t * frame, frame_data_command_t * frame_data) {
    frame->buffer[0] = 0x02;                                            /* STX */
    frame->buffer[1] = 0x00;                                            /* SEQ */
    frame->buffer[2] = 0x00;                                            /* DADD */
    frame->buffer[3] = frame_data->cmd;                                 /* CMD */
    frame->buffer[4] = (uint8_t)(frame_data->size + sizeof(uint8_t));   /* DATA_LENGTH */
    frame->buffer[5] = frame_data->time;                                /* TIME */
    memcpy(&frame->buffer[6], frame_data->data, frame_data->size);      /* DATA */
    frame->buffer[frame->size - 2] = 0x00;                              /* BCC */
    frame->buffer[frame->size - 1] = 0x03;                              /* ETX */
}

/**
 * _increment_seq - increase the SEQ field
 * 
 * @param frame which will get the Seq field
 * @param seq sequence to increase
 */
static void _increment_seq (cn56_frame_t * frame, uint8_t seq) {
    if (seq >= 0xf0) {
        seq =  0x80;
    } else {
        seq = (uint8_t)(seq + 0x10);
    }
    frame->buffer[1] = seq;
}

/**
 * _calc_bbc - update the BCC field. It use the EOR operation between all fields without STX & ETX
 *  EOR = Exclusive OR 
 * 
 * @param frame which will get the BBC field   
 */
static void _calc_bbc (cn56_frame_t * frame) {
    frame->buffer[frame->size-2] = frame->buffer[1] ^ frame->buffer[2] ^ frame->buffer[3] ^ frame->buffer[4] ^ frame->buffer[5];
    for (int n=0; n < (int)frame->buffer[4]-1; n++) {
        frame->buffer[frame->size-2] ^= frame->buffer[6 + n];
    }
}


/**
 * _create_frame_command - Create the frame to be send to the Reader
 * 
 * @param cmd The command to send
 * @param frame the frame which hold infos to be send
 * @param cn56_add_config configuration option if the cmd need specific data
 * @param lastSeq Last sequence send to the reader
 */
static void  _create_frame_command (cn56_command_t cmd, cn56_frame_t *frame, void * cn56_add_config, uint8_t lastSeq) {
    frame_data_command_t data;

    _create_data_command (&data, cmd, cn56_add_config);
    frame->size = 8 + data.size;

    _init_frame_command(frame, &data);
    _increment_seq(frame, lastSeq);
    _calc_bbc(frame);
}

/**
 * _is_seq - compare the seq with the SEQ field in frame
 * 
 * @param seq: sequence to compare
 * @param buff: buffer of the frame
 * @return true or false
 */
static bool _is_seq(uint8_t seq, uint8_t * buff) {
    return (seq == GETSEQ(buff));
}

/**
 * _is_success - read the response STATUS 
 * 
 * @param buff buffer of the frame
 * @return true or false
 */
static bool _is_success(uint8_t * buff) {
    return ((GETSTATUS(buff) == 0x00)? true: false);
}

/**
 * _encrypt_data - Encrypt a data of length len with the uuid of the badge
 * 
 * @param buffer Where data has to be encrypt
 * @param buffer_len Length of the data to encrypt
 * @param uuid Unique ID of the badge
 * @param uuid_len Length of th uuid
 * @return 0 in success, <0 in error
 */
static int _encrypt_data(uint8_t *buffer, uint8_t buffer_len, const uint8_t *uuid, uint8_t uuid_len) {
    if (!buffer || !uuid) return -CN56_ERROR_NULL_ARGS;
    int ret = 0;

    // Create the key
    uint8_t *key = (uint8_t *)calloc(buffer_len, sizeof(uint8_t));
    for (int idx = 0; idx < buffer_len; idx ++) {
        key[idx] = uuid[(idx+2) % uuid_len];
    }

    // Encrypt the buffer
    for (int idx = 0; idx < buffer_len; idx++) {
        buffer[idx] = key[idx] ^ buffer[idx];
    }

    free(key);
    return ret;
}

/**
 * _decrypt_data - Decrypt a data of length len with the uuid of the badge
 * 
 * @param buffer Where data has to be decrypt
 * @param buffer_len length of the data to decrypt
 * @param uuid Unique ID of the badge
 * @param uuid_len Length of th uuid
 * @return 0 in success, <0 in error
 */
static int _decrypt_data(uint8_t *buffer, uint8_t buffer_len, const uint8_t *uuid, uint8_t uuid_len) {
    if (!buffer || !uuid) return -CN56_ERROR_NULL_ARGS;
    int ret = 0;

    // Create the key
    uint8_t *key = (uint8_t *)calloc(buffer_len, sizeof(uint8_t));
    for (int idx = 0; idx < buffer_len; idx ++) {
        key[idx] = uuid[(idx+2) % uuid_len];
    }

    // Decrypt the buffer
    for (int idx = 0; idx < buffer_len; idx++) {
        buffer[idx] = key[idx] ^ buffer[idx];
    }

    free(key);
    return ret;
}

/**
 * _cn56_verify_response - Verify of response of a command is a success
 * 
 * @param reader ptr to the reader context
 * @return 0 in success, <0 in error
 */
static int _cn56_verify_response(cn56_reader_t *reader) {
    // Check the sequence
    if (!_is_seq(reader->seq, reader->buffer)) {
        return -CN56_ERROR_BAD_SEQ;
    }

    // Check the success in respons
    if (!_is_success(reader->buffer)) {
        return -CN56_ERROR_NOT_SUCCESS;
    }
    return 0;
}

/**
 * _cn56_catch_response - Catch and verify the respons from the buffer
 * 
 * @param reader ptr to the reader context
 * @param data where the response will be stores. data->buffer need to be free in success !!!!
 * @param response_len gives the response length
 * @return negative int if error, positive if status, 0 if success
 */
static int _cn56_catch_response(cn56_reader_t *reader, data_frame_t* data, int *response_len) {
    int ret = _cn56_verify_response(reader);
    if (ret < 0) {
        return ret;
    }

    *response_len = GETDATALENGTH(reader->buffer) -1;
    if (*response_len < 0) {
        return -CN56_ERROR_NO_DATA;
    }
    data->buffer = (uint8_t *) calloc(*response_len, sizeof(uint8_t));
    data->size = *response_len;

    // Get Data
    memcpy(data->buffer, GETDATA(reader->buffer), data->size);

    return 0;
}

/**
 * _cn56_rcv - wait epoll event and call recv from tcp socket and store response in reader->buffer
 *
 * @param reader pointer to the specific reader
 * @param timeout time to wait the response
 * 
 * @return a size of response, error if <0
 */
static size_t _cn56_rcv(cn56_reader_t *reader, int timeout) {
    struct epoll_event events[1];
    int event_count = 0;
    int count  = 0;
    ssize_t size = 0;

    // Wait EPOLLIN Event
    event_count = epoll_wait(reader->cmd_epoll_fd, events, 1, timeout);
    for (int index = 0; index < event_count; index++) {
        if (events[index].events & EPOLLIN) {
            size = recv(reader->sockfd, reader->buffer, MAX_CN56_BUFFER_LENGTH - 1, 0);

            while (size == EAGAIN && count <= 10) {
                fprintf(stdout, "busy read, retry\n");
                size = recv(reader->sockfd, reader->buffer, MAX_CN56_BUFFER_LENGTH - 1, 0);
                count ++;
                usleep(1000);
            }
            if (size <= 0) {
                return -CN56_ERROR_READ;
            }
            if (GETSEQ(reader->buffer) != reader->seq) {
                // Try to listen the next response:
                event_count = epoll_wait(reader->cmd_epoll_fd, events, 1, 1000);
                for (int index = 0; index < event_count; index++) {
                    if (events[index].events & EPOLLIN) {
                        size = recv(reader->sockfd, reader->buffer, MAX_CN56_BUFFER_LENGTH - 1, 0);
                        if (size <= 0) {
                            return -CN56_ERROR_READ;
                        }
                    }
                }
            }
        }
    }
    if(event_count == 0) {
        return -CN56_ERROR_TIMEOUT;
    }
    return size;
}

/**
 * _cn56_send_command_async - Send command to the specific reader
 *
 * @param reader pointer to the specific reader
 * @param cmd cn56 command to send
 * @param cn56_add_config specific config for some command
 * @return a CN56Error
 */
static int _cn56_send_command_async(cn56_reader_t *reader, cn56_command_t cmd, void *cn56_add_config) {
    ssize_t size = 0;
    cn56_frame_t frame_cmd;
    int count = 0;

    if (reader->state != CN56_STATE_CONNECTED && reader->action != CN56_ACTION_RESET) {
            reader->error = "Reader is disconnected!";
        return -CN56_ERROR_DISCONNECTED;
    }

    _create_frame_command(cmd, &frame_cmd, cn56_add_config, reader->seq);

    // Update the SEQ of the last cmd
    reader->seq = GETSEQ(frame_cmd.buffer);

    // Send command in TCP
    ssize_t write_ret =0;
    do {
        write_ret = write(reader->sockfd, &(frame_cmd.buffer[size]), frame_cmd.size - size);
        if (write_ret <= 0) {
            if (write_ret == EAGAIN) {
                count ++;
                usleep(1000);
                write_ret = 0;
            } else {
                fprintf(stderr, "Write() error : %s\n", strerror(errno));
                reader->state = CN56_STATE_BAD;
                reader->error = "Failed to write on the tcp socket!";
                return -CN56_ERROR_WRITE;
            }
        }
        size += write_ret;
    }while ((int) size < frame_cmd.size && count <= 10);

    return 0;
}

/**
 * _cn56_send_command - Send command to the specific reader and wait the response
 *
 * @param reader pointer to the specific reader
 * @param cmd cn56 command to send
 * @param cn56_add_config specific config for some command
 * @param data Where the respons will be stored need to free data->buffer if success
 * @param response_len store the length of the response if success
 * @param timeout delay to wait before the response
 * @return a CN56Error
 */
static int _cn56_send_command(cn56_reader_t *reader, cn56_command_t cmd, void *cn56_add_config, data_frame_t *data, int *response_len, int timeout) {
    ssize_t size = 0;
    cn56_frame_t frame_cmd;
    int count = 0;
    int ret =0;

    if (reader->state != CN56_STATE_CONNECTED) {
            reader->error = "Reader is disconnected!";
        return -CN56_ERROR_DISCONNECTED;
    }

    _create_frame_command(cmd, &frame_cmd, cn56_add_config, reader->seq);

    // Update the SEQ of the last cmd
    reader->seq = GETSEQ(frame_cmd.buffer);

    // Send command in TCP
    ssize_t write_ret =0;
    do {
        write_ret = write(reader->sockfd, &(frame_cmd.buffer[size]), frame_cmd.size - (int)size);
        if (write_ret < 0) {
            if (write_ret == EAGAIN) {
                count ++;
                usleep(1000);
                write_ret = 0;
            } else {
                fprintf(stderr, "[%s] Write() error : %s\n", reader->uid, strerror(errno));
                reader->state = CN56_STATE_BAD;
                reader->error = "Failed to write on the tcp socket!";
                ret = -CN56_ERROR_WRITE;
                goto OnExit;
            }
        }
        size += write_ret;
    }while ((int) size < frame_cmd.size && count <= 10);

    // Wait the response in TCP
    ret =(int) _cn56_rcv(reader, timeout);
    if (ret < 0) {
        if (ret != -CN56_ERROR_TIMEOUT) ret = -CN56_ERROR_READ;
        goto OnExit;
    }

    // Catch the data and verify
    ret = _cn56_catch_response(reader, data, response_len);

OnExit:
    return ret;
}

/**
 * _authenticate - authenticate to be able to modify keys or Read sector of a badge folowing this step: 
 *      PN_InListPassive -> PN_SelectTargets -> LoadKey -> Authenticate
 * 
 * @param reader pointer to reader object
 * @param key  with which key : key object
 * @param sec  with sector to authenticate
 * @param uid badge's uid caught
 * @return 0 in success, <0 in error
 */
static int _authenticate(cn56_reader_t *reader, const cn56_key_t *key, uint8_t sec, uint8_t **uid) {
    int ret = 0;
    int response_len = 0;
    cn56_action_e old_action = reader->action;
    reader->action = CN56_ACTION_AUTHENTICATION;
    data_frame_t frame_data = {0};

    // Check state
    if (!reader || !*uid) return -CN56_ERROR_NULL_ARGS;
    if (reader->state != CN56_STATE_CONNECTED) return -CN56_ERROR_DISCONNECTED;

    ret = _cn56_send_command(reader, cn56_cmd_pn_inListPassive, NULL, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret < 0) {
        reader->error = "[Authentication] Error to listPassive";
        goto OnExit;
    }
    if (response_len <= 0) {
        reader->error = "[Authentication] Error to get uuid from listPassive command";
        goto OnExit;
    }
    for (int index = 0; index < 4; index ++) {
        (*uid)[index] = frame_data.buffer[frame_data.size - 4 + index];
    }
    free(frame_data.buffer);

    // PN_SelectTargets
    ret = _cn56_send_command(reader, cn56_cmd_pn_selectTarget, NULL, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret < 0) {
        if (ret != -CN56_ERROR_NO_DATA) {
            reader->error = "[Authentication] Error to selectTarget";
            goto OnExit;
        }
    } else {
        free(frame_data.buffer);
    }

    // LoadKey
    cn56_config_load_key_t config_load_key = {
        .key = (key->idx == 0) ? 0x60 : 0x61,
        .sector = sec,
        .value = {key->value[0], key->value[1], key->value[2], key->value[3], key->value[4], key->value[5]},
    };
    ret = _cn56_send_command(reader, cn56_cmd_loadKey, &config_load_key, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret < 0) {
        if (ret != -CN56_ERROR_NO_DATA) {
            reader->error = "[Authentication] Error to LoadKey";
            goto OnExit;
        }
    } else {
        free(frame_data.buffer);
    }

    //Authenticate
    cn56_config_authenticate_t config_authenticate = {
        .key = (key->idx == 0) ? 0x60 : 0x61,
        .uid = {(*uid)[0], (*uid)[1], (*uid)[2], (*uid)[3]},
        .sec = sec,
    };
    ret = _cn56_send_command(reader, cn56_cmd_authenticate, &config_authenticate, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret < 0) {
        if (ret != -CN56_ERROR_NO_DATA) {
            reader->error = "[Authentication] Error to authenticate";
            ret = -CN56_ERROR_AUTHENTICATE;
            goto OnExit;
        }
    } else {
        free(frame_data.buffer);
    }

    // SAve the authentication
    reader->authenticate.isAuthenticate = true;
    reader->authenticate.sec = sec;
    reader->authenticate.key = (cn56_key_t *) key;
    memcpy(reader->authenticate.uuid, *uid, 4*sizeof(uint8_t));

OnExit:
    reader->action = old_action;
    return ret;
}

void sigpipe_handler(int unused) {
    fprintf(stdout, " ~~~ SIGPIPE CAUGHT with argument: %i ~~~\n", unused);
}

/** 
 * _cn56_thread_func - Take care about the epoll event of the cn56 reader
 */
static void *_cn56_thread_func(void *ctx) {
    if (!ctx) pthread_exit(NULL);

    cn56_reader_t * reader = (cn56_reader_t *)ctx;
    struct epoll_event events[MAX_EVENTS];
    int event_count = 0;
    int ret = 0;

    sigaction(SIGPIPE, (struct sigaction*) sigpipe_handler, NULL);

    while (1) {
        event_count = 0;
        event_count = epoll_wait(reader->epoll_fd, events, MAX_EVENTS, -1);
        //fprintf(stdout, "[%s] Event occurred on socket\n", reader->uid);
        if (reader->action == CN56_ACTION_AUTHENTICATION
            || reader->action == CN56_ACTION_TRAILER
            || reader->action == CN56_ACTION_READ
            || reader->action == CN56_ACTION_WRITE) {
            //usleep(50000);
            continue;
        }
        //pthread_mutex_lock(&reader->mutex);
        for (int index = 0; index < event_count; index++) {
            if (events[index].events & EPOLLRDHUP) {
                //fprintf(stdout, "[%s] EPOLLRDHUP \n", reader->uid);
                
                //if (reader->state == CN56_STATE_CONNECTED) reader->state = CN56_STATE_DISCONNECTED;
                //close(reader->sockfd);
                //continue;
                goto OnExit;
            }
            if (events[index].events & EPOLLWAKEUP) {
                //printf(" EPOLLWAKEUP \n");
            }
            if (events[index].events & EPOLLERR) {
                //printf(" EPOLLERR \n");
                if (reader->state == CN56_STATE_CONNECTED) reader->state = CN56_STATE_DISCONNECTED;
                close(reader->sockfd);
                continue;
            }
            if (events[index].events & EPOLLET) {
                //printf(" EPOLLET \n");
            }
            if (events[index].events & EPOLLOUT) {
                // fprintf(stdout, "[%s] EPOLLOUT \n", reader->uid);
                if (reader->state == CN56_STATE_DISCONNECTED) {
                    // test if the connection is well
                    uint8_t test[1] = {0};
                    ret = (int) write(reader->sockfd, test, 1);
                    if (ret >= 0) {
                        reader->state = CN56_STATE_CONNECTED;
                        if (reader->functions_handler->connection_handle) {
                            reader->functions_handler->connection_handle(CN56_STATE_CONNECTED, reader->user_data_connection);
                        } else {
                            fprintf(stderr, "Missing connection handler\n");
                        }
                    }
                }
            }
            usleep(1000);
        }
    }
OnExit:
    reader->state = CN56_STATE_DISCONNECTED;
    if (epoll_ctl(reader->cmd_epoll_fd, EPOLL_CTL_DEL, reader->sockfd, NULL)) {
        perror("Failed to del ctl epoll_cmd");
    }
    close(reader->sockfd);
    if (reader->functions_handler->connection_handle) {
        reader->functions_handler->connection_handle(CN56_STATE_DISCONNECTED, reader->user_data_connection);
    }
    pthread_exit(NULL);
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PUBLIC FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

int cn56_init(cn56_reader_t *reader, const char *reader_addr, uint16_t cli_port, const char *uid) {
    int ret = 0;

    if (!reader) {
        return -1;
    }

    //Initiate some reader data
    reader->uid = uid;
    reader->port = (uint16_t) 2000;
    reader->port_cli = cli_port;
    reader->state = CN56_STATE_DISCONNECTED;
    reader->action = CN56_ACTION_UNKNOWN;
    reader->seq = 0x80;
    reader->thread_id = 0;
    reader->user_data_connection = NULL;
    reader->user_data_response = NULL;
    reader->user_data_bad_authent = NULL;
    // Allocate memory for handler functions
    reader->functions_handler = (cn56_handlers_t *) calloc(1, sizeof(cn56_handlers_t));
    if (!reader->functions_handler) {
        perror("Failed to allocate handler functions");
        return -1;
    }

    // Initiate the authentication
    reader->authenticate.isAuthenticate = false;
    reader->authenticate.sec = 0;
    reader->authenticate.key = NULL;
    memset(reader->authenticate.uuid, 0, sizeof(uint8_t));

    // Initiate reader server addr
    reader->srvaddr.sin_family = AF_INET;
    reader->srvaddr.sin_addr.s_addr = inet_addr(reader_addr);
    reader->srvaddr.sin_port = htons(reader->port);
    
    // Alloc buffer
    reader->buffer = (uint8_t *)calloc(1 * MAX_CN56_BUFFER_LENGTH, sizeof(uint8_t));

    // Initiate a epoll socket for cmd response
    reader->epoll_fd = epoll_create1(0);
    if (reader->epoll_fd < 0) {
        perror("Failed to create epoll");
        ret = -1;
        goto OnErrorExit;
    }
    reader->cmd_epoll_fd = epoll_create1(0);
    if (reader->cmd_epoll_fd < 0) {
        perror("Failed to create cmd_epoll");
        ret = -1;
        goto OnErrorExit;
    }
    ret = pthread_mutex_init(&reader->mutex, NULL);
    if (ret < 0) {
        perror("Failed to initiate mutext");
        goto OnErrorExit;

    }

    goto OnExit;

OnErrorExit:
    free(reader->functions_handler);

OnExit:
    return ret;
}

int cn56_add_connection_handler(cn56_reader_t *reader, cn56_connection_handler_t handler, void *user_data) {
    if(!reader->functions_handler && !handler) {
        return -CN56_ERROR_NULL_ARGS;
    } else {
        reader->functions_handler->connection_handle = handler;
        reader->user_data_connection = user_data;
        return 0;
    }
}

int cn56_add_response_handler(cn56_reader_t *reader, cn56_response_handler_t handler, void *user_data) {
    if(!reader->functions_handler && !handler) {
        return -CN56_ERROR_NULL_ARGS;
    } else {
        reader->functions_handler->response_handle = handler;
        reader->user_data_response = user_data;
        return 0;
    }
}

int cn56_add_bad_authent_handler(cn56_reader_t *reader, cn56_bad_authent_handler_t handler, void *user_data) {
    if(!reader->functions_handler && !handler) {
        return -CN56_ERROR_NULL_ARGS;
    } else {
        reader->functions_handler->bad_authent_handle = handler;
        reader->user_data_bad_authent = user_data;
        return 0;
    }
}

// void sigpipe_handler(int unused) {
//     fprintf(stdout, " ~~~ SIGPIPE CAUGHT with argument: %i ~~~\n", unused);
// }

int cn56_connect_async_tcp(cn56_reader_t *reader) {
    int flags = 0;
    int optval = 1;
    socklen_t optlen = sizeof(optval);
    int ret = 0;

    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;
    // if (reader->state == CN56_STATE_CONNECTED) return -CN56_ERROR_ALREADY_CONNECTED;
    
    // Create the socket
    reader->sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (reader->sockfd < 0) {
        perror("socket failed");
        return -CN56_ERROR_SOCKET;
    }

    // Set the Socket in no-blocking
    flags = fcntl(reader->sockfd, F_GETFD);
    if (fcntl(reader->sockfd, F_SETFL, flags | O_NONBLOCK) == -1) {
        perror("socket fcntl failed");
        ret = -CN56_ERROR_SOCKET;
        goto OnSocketErrorExit;
    }

    // Set Keepalive option
    if (setsockopt(reader->sockfd, SOL_SOCKET, SO_KEEPALIVE, &optval, optlen) < 0) {
        perror("set socket option keepalive failed");
        ret = -CN56_ERROR_SOCKET;
        goto OnSocketErrorExit;
    }

    // Set Keepidle option
    optval = 10;
    if (setsockopt(reader->sockfd, SOL_TCP, TCP_KEEPIDLE, &optval, optlen) < 0) {
        perror("set socket option keepidle failed");
        ret = -CN56_ERROR_SOCKET;
        goto OnSocketErrorExit;
    }

    // Initiate epoll event
    struct epoll_event event = {0};
    struct epoll_event event_cmd = {0};
    //event.events = EPOLLHUP | EPOLLIN | EPOLLWAKEUP | EPOLLERR | EPOLLET | EPOLLOUT | EPOLLRDHUP;
    event.events = /*EPOLLIN|*/EPOLLRDHUP|EPOLLET|EPOLLOUT;
    event.data.fd = reader->sockfd;

    if (epoll_ctl(reader->epoll_fd, EPOLL_CTL_ADD, reader->sockfd, &event)) {
        perror("Failed to ctl epoll");
        ret = -CN56_ERROR_EPOLL;
        goto OnEpollErrorExit;
    }
    event_cmd.events = EPOLLIN;
    event_cmd.data.fd = reader->sockfd;
    if (epoll_ctl(reader->cmd_epoll_fd, EPOLL_CTL_ADD, reader->sockfd, &event_cmd)) {
        perror("Failed to ctl epoll_cmd");
        ret = -CN56_ERROR_EPOLL;
        goto OnEpollErrorExit;
    }

    // Create thread which wait on Epoll event
    ret = pthread_create(&reader->thread_id, NULL, _cn56_thread_func, reader);
    if (ret < 0) {
        perror("Error in thread creation");
        ret = -CN56_ERROR_CREATION_THREAD;
        goto OnEpollErrorExit;
    }

    // Connect to the reader
    ret = connect(reader->sockfd, (struct sockaddr *)&reader->srvaddr, sizeof(reader->srvaddr));
    if (ret < 0) {
        if (errno == EINPROGRESS) {
            // printf("\nConnecting ...\n");
        } else {
            perror("Failed to connect!");
            ret = -CN56_ERROR_CONNECT;
        }
    } else { 
        printf("\nConnected !\n");
        reader->state = CN56_STATE_CONNECTED;
    }


    //sigaction(SIGPIPE, (struct sigaction*) sigpipe_handler, NULL);

    ret = 0;
    goto Exit;

OnEpollErrorExit:
    close(reader->epoll_fd);

OnSocketErrorExit:
    close (reader->sockfd);

Exit:
    return ret;
}

int cn56_disconnect_tcp(cn56_reader_t *reader) {
    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;
    // if (reader->state == CN56_STATE_DISCONNECTED) return -CN56_ERROR_DISCONNECTED;

    if (close(reader->sockfd) < 0) {
        perror("Failed to close socket");
        return -CN56_ERROR_CLOSED_SOCKET;
    }
    pthread_cancel(reader->thread_id);
    pthread_join(reader->thread_id, NULL);

    reader->state = CN56_STATE_DISCONNECTED;
    return 0;
}

int cn56_get_uuid_badge(cn56_reader_t * reader, uint8_t **uuid) {
    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;

    data_frame_t frame_data = {0};
    int ret = 0;
    int response_len = 0;

    ret = _cn56_send_command(reader, cn56_cmd_pn_inListPassive, NULL, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);

    if (ret < 0) {
        if (ret != -CN56_ERROR_NOT_SUCCESS)
            reader->error = "[cn56_get_uuid_badge] Error to listPassive";
        return ret;
    }
    if (response_len <= 0) {
        reader->error = "[cn56_get_uuid_badge] No Data in response";
        return -CN56_ERROR_NO_DATA;
    }
    *uuid = (uint8_t *) calloc(4, sizeof(uint8_t));
    for (int index = 0; index < 4; index ++) {
        (*uuid)[index] = frame_data.buffer[frame_data.size - 4 + index];
    }

    free (frame_data.buffer);

    return 0;
}

int cn56_green_led(cn56_reader_t *reader) {
    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;
    // if (reader->state != CN56_STATE_CONNECTED) return -CN56_ERROR_DISCONNECTED;
    int ret  = 0;
    int response_len = 0;
    data_frame_t frame_data = {0};
    reader->action = CN56_ACTION_LED;
    
    // Send command command
    ret = _cn56_send_command(reader, cn56_cmd_activateGreenLed, NULL, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret < 0) {
        reader->error = "Error to send green LED command";
    } else {
        free (frame_data.buffer);
    }
       
    return ret;
}

int cn56_red_led(cn56_reader_t *reader) {
    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;
    // if (reader->state != CN56_STATE_CONNECTED) return -CN56_ERROR_DISCONNECTED;
    int ret  = 0;
    int response_len = 0;
    data_frame_t frame_data = {0};
    reader->action = CN56_ACTION_LED;

    // Send command command
    ret = _cn56_send_command(reader, cn56_cmd_activateRedLed, NULL, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret < 0) {
        reader->error = "Error to send green LED command";
    } else {
        free (frame_data.buffer);
    }
       
    return ret;
}

int cn56_no_led(cn56_reader_t *reader) {
    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;
    // if (reader->state != CN56_STATE_CONNECTED) return -CN56_ERROR_DISCONNECTED;
    int ret  = 0;
    int response_len = 0;
    data_frame_t frame_data = {0};
    reader->action = CN56_ACTION_LED;

    // Send command command
    ret = _cn56_send_command(reader, cn56_cmd_turnOffLed, NULL, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret < 0) {
        reader->error = "Error to send green LED command";
    } else {
        free (frame_data.buffer);
    }

    return ret;
}

int cn56_activate_ext(cn56_reader_t *reader) {
    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;
    // if (reader->state != CN56_STATE_CONNECTED) return -CN56_ERROR_DISCONNECTED;
    int ret = 0;
    int response_len = 0;
    data_frame_t frame_data = {0};

    // Send command command
    cn56_config_delay_t config_delay = {
        .delay = 0x04,
    };
    ret = _cn56_send_command(reader, cn56_cmd_Cext, &config_delay, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret < 0) {
        reader->error = "Error to send activate external command";
    } else {
        free (frame_data.buffer);
    }

    return ret;
}

int cn56_free(cn56_reader_t *reader) {
    free(reader->functions_handler);
    free(reader->buffer);
    free(reader);
    return 0;
}

int cn56_read_mifare(cn56_reader_t *reader, const cn56_key_t *key, uint8_t sec, uint8_t blk, uint8_t len, uint8_t *response) {
    // Check args
    if (!reader || !key) return -CN56_ERROR_NULL_ARGS;
    // if (reader->state != CN56_STATE_CONNECTED) return -CN56_ERROR_DISCONNECTED;
    int ret = 0;
    int response_len = 0;
    uint8_t *uid = (uint8_t *) calloc(4, sizeof(uint8_t));
    data_frame_t frame_data = {0};

    pthread_mutex_lock(&reader->mutex);

    // Authenticate
    if (sec== 0 && blk > 4) {
        sec = (uint8_t)((blk - blk%4)/4);
    }
    if (reader->authenticate.isAuthenticate
        && reader->authenticate.sec == sec
        && !strcmp(reader->authenticate.key->uid, key->uid)) {
    } else {
        ret = _authenticate(reader, key, sec, &uid);
        if (ret <0) {
            int auth_count = 0;
            while (ret == -CN56_ERROR_AUTHENTICATE && auth_count < 3) {
                ret = _authenticate(reader, key, sec, &uid);
                auth_count ++;
            }
            if (ret < 0)
                goto OnExit;
            fprintf(stdout, "read authent success after %i try\n", auth_count);
        }
    }
    cn56_config_read_t config_read = {0};
    config_read.blk = (uint8_t)(sec*4 + blk);

    // Read command
    ret = _cn56_send_command(reader, cn56_cmd_readBlk, &config_read, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret < 0) {
        if (ret != -CN56_ERROR_NO_DATA) {
            reader->error = "[cn56_read_mifare] Error to send read command";
            goto OnExit;
        }
    }

    // Split on the response length
    if (len == 0) {
        len = CN56_MAX_WRITE;
        memcpy(response, &frame_data.buffer[0], len);
    } else if (len < response_len) {
        memcpy(response, &frame_data.buffer[response_len - len], len);
    } else {
        len = (uint8_t) response_len;
        memcpy(response, &frame_data.buffer[len], len);
    }

    // Decrypt data to read
    ret = _decrypt_data(response,  len, reader->authenticate.uuid, 4);
    if (ret < 0) {
        goto OnExit;
    }
    free(frame_data.buffer);
    
OnExit:
    pthread_mutex_unlock(&reader->mutex);
    free(uid);
    return ret;
}

int cn56_write_mifare(cn56_reader_t *reader, const cn56_key_t *key, uint8_t sec, uint8_t blk, uint8_t len, uint8_t *buffer) {
    int ret = 0;
    uint8_t *uid = (uint8_t *) calloc(4, sizeof(uint8_t));
    uint8_t *tmp_buffer = (uint8_t *) calloc(len, sizeof(uint8_t));
    data_frame_t frame_data = {0};
    int response_len = 0;
    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;
    if (reader->state != CN56_STATE_CONNECTED) return -CN56_ERROR_DISCONNECTED;
    //reader->action = CN56_ACTION_WRITE;

    pthread_mutex_lock(&reader->mutex);

    if((blk == 0 && sec ==0) || blk%4 == 3) return -CN56_ERROR_WRONG_BLK;

    // Authenticate
    if (sec == 0 && blk > 4) {
        sec = (uint8_t)((blk - blk%4)/4);
    }
    if (reader->authenticate.isAuthenticate
        && reader->authenticate.sec == sec
        && !strcmp(reader->authenticate.key->uid, key->uid)) {
    } else {
        ret = _authenticate(reader, key, sec, &uid);
        if (ret <0) {
            int auth_count = 0;
            while (ret == -CN56_ERROR_AUTHENTICATE && auth_count < 3) {
                ret = _authenticate(reader, key, sec, &uid);
                auth_count ++;
            }
            if (ret < 0) {
                goto OnExit;
            }
            fprintf(stdout, "write authent success after %i try\n", auth_count);
        }
    }

    // Encrypt de data
    memcpy(tmp_buffer, buffer, len);
    ret = _encrypt_data(tmp_buffer, len, reader->authenticate.uuid, 4);
    if (ret < 0) {
        goto OnExit;
    }

    // Can write maximum of CN56_MAX_WRITE character
    cn56_config_write_block_t config_write = {0};
    config_write.blk = (uint8_t) (sec*4 + blk);
    if (len < CN56_MAX_WRITE) {
        for (int index = 0; index < len; index ++) {
            config_write.data[CN56_MAX_WRITE - len + index] = tmp_buffer[index];
        }
    } else {
        for (int index = 0; index < CN56_MAX_WRITE; index ++) {
            config_write.data[index] = tmp_buffer[index];
        }
        len = CN56_MAX_WRITE;
    }

    // Write command
    ret = _cn56_send_command(reader, cn56_cmd_writeBlk, &config_write, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret < 0) {
        if (ret != -CN56_ERROR_NO_DATA) {
            reader->error = "[cn56_write_mifare] Error to send write command";
            goto OnExit;
        }
    }
    free(frame_data.buffer);

OnExit:
    pthread_mutex_unlock(&reader->mutex);
    free(tmp_buffer);
    free(uid);
    return ret;
}

int cn56_store_key(cn56_reader_t * reader, uint8_t key, uint8_t sector, uint8_t *value) {
    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;
    if (reader->state != CN56_STATE_CONNECTED) return -CN56_ERROR_DISCONNECTED;

    if (sector > 15 || (key != 0x60 && key != 0x61)) {
        return -CN56_ERROR_WRONG_ARGS;
    }
    cn56_config_load_key_t config_load_key = {
        .key = key,
        .sector = sector,
        .value= {value[0], value[1], value[2], value[3], value[4], value[5]},
    };
    return _cn56_send_command_async(reader, cn56_cmd_storeKey, &config_load_key);
}

int cn56_set_config(cn56_reader_t * reader, cn56_config_wiegand_t config) {
    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;
    if (reader->state != CN56_STATE_CONNECTED) return -CN56_ERROR_DISCONNECTED;
    data_frame_t frame_data = {0};
    int response_len = 0;

    // Send command
    int ret = _cn56_send_command(reader, cn56_cmd_configWiegand, &config, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret < 0) {
        reader->error = "Error to send set config";
    } else {
        free (frame_data.buffer);
    }
    return ret;
}

int cn56_trailer(cn56_reader_t * reader, uint8_t sec, const cn56_key_t *key, cn56_config_modif_key_t config_modif_key) {
    int ret = 0;
    data_frame_t frame_data = {0};
    int response_len = 0;
    uint8_t *uid = (uint8_t *) calloc(4, sizeof(uint8_t));
    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;
    if (reader->state != CN56_STATE_CONNECTED) return -CN56_ERROR_DISCONNECTED;
    if (!key) return -CN56_ERROR_NULL_ARGS;

    // Authenticate
    ret = _authenticate(reader, key, sec, &uid);
    if (ret <0) {
        goto OnErrorExit;
    }

    // Construct the modif config object
    // cn56_config_modif_key_t config_modif_key = {
    //     .keyA = {trailer->keyA->value[0], trailer->keyA->value[1], trailer->keyA->value[2], trailer->keyA->value[3], trailer->keyA->value[04], trailer->keyA->value[5]},
    //     .keyB = {trailer->keyB->value[0], trailer->keyB->value[1], trailer->keyB->value[2], trailer->keyB->value[3], trailer->keyB->value[04], trailer->keyB->value[5]},
    //     .acls = {trailer->acls[0], trailer->acls[1], trailer->acls[2], trailer->acls[3]},
    //     .sector = sec,
    // };

    // Send command pre-modify
    cn56_config_read_t config_read = {0};
    config_read.blk = (uint8_t)(sec*4);
    ret = _cn56_send_command(reader, cn56_cmd_readBlk, &config_read, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret <0) {
        if (ret != -CN56_ERROR_NO_DATA) {
            goto OnErrorExit;
        }
    } else {
        free(frame_data.buffer);
    }

    // Send command modify
    ret = _cn56_send_command(reader, cn56_cmd_modifKey, &config_modif_key, &frame_data, &response_len, CN56_TIMEOUT_RECEIVE_MS);
    if (ret <0) {
        if (ret != -CN56_ERROR_NO_DATA) {
            goto OnErrorExit;
        }
    } else {
        free(frame_data.buffer);
    }
    free(uid);
    return 0;

OnErrorExit:
    free(uid);
    return ret;
}

int cn56_reset(cn56_reader_t *reader) {
    int ret = 0;
    // Check args
    if (!reader) return -CN56_ERROR_NULL_ARGS;

    int sockfd;
    ssize_t size;
    struct sockaddr_in     servaddr;
    cn56_frame_t frame_cmd;

    // Close the last tcp connection
    ret = cn56_disconnect_tcp(reader);
    if (ret < 0) {
        fprintf(stderr, "Something went wrong during connection: err=%i\n", ret);
        // return ret;
    }
    reader->state = CN56_STATE_DISCONNECTED;

    // Creating socket file descriptor
    if ( (sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0 ) {
        perror("socket creation failed");
        exit(EXIT_FAILURE);
    }
    memset(&servaddr, 0, sizeof(servaddr));
      
    // Filling server information
    servaddr.sin_family = AF_INET;
    servaddr.sin_port = reader->srvaddr.sin_port;
    servaddr.sin_addr.s_addr = reader->srvaddr.sin_addr.s_addr;

    _create_frame_command(cn56_cmd_reset, &frame_cmd, NULL, reader->seq);

    // Update the SEQ of the last cmd
    reader->seq = GETSEQ(frame_cmd.buffer);

    size = sendto(sockfd, frame_cmd.buffer
                            , frame_cmd.size
                            , MSG_WAITALL
                            , (const struct sockaddr *) &servaddr
                            , sizeof(servaddr));
    if (size <= 0) {
        fprintf(stderr, "size return: %ld\n", size);
        return -CN56_ERROR_WRITE;
    }

    sleep(2);
    close (sockfd);

    return ret;
}

const char *cn56_error_msg(cn56_reader_t *reader) {
    return (reader->error);
}
