/*
 * Copyright (C) 2020 - 2021 "IoT.bzh"
 *
 * Author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

// Standard includes
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>

// Include for thread and epoll
#include <pthread.h>
#include <sys/epoll.h>
#include <unistd.h>

// Includes for Udp/tcp connection
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#define MAX_EVENTS 5
#define CN56_MAX_WRITE 16
#define CN56_TIMEOUT_RECEIVE_MS 1000

#define GETSTX(x) ((x)[0])
#define GETSEQ(x) ((x)[1])
#define GETDADD(x) ((x)[2])
#define GETDATALENGTH(x) ((x)[3])
#define GETSTATUS(x) ((x)[4])
#define GETDATA(x) (&(x)[5])
#define GETBCC(x) ((x)[5 + GETDATALENGTH(x) -1])
#define GETETX(x) ((x)[6 + GETDATALENGTH(x) -1])

#define MAX_CN56_BUFFER_LENGTH 80

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            ENUM
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * Option - Enum which List of different command available to be send to the CN56 reader badges
 */
typedef enum cn56_command_e {
    cn56_cmd_getFirmware,
    cn56_cmd_activateRedLed,
    cn56_cmd_activateGreenLed,
    cn56_cmd_activateBothLed,
    cn56_cmd_turnOffLed,
    cn56_cmd_configWiegand,
    cn56_cmd_runWiegandMode,
    cn56_cmd_autoRead,
    cn56_cmd_writeMifare,
    cn56_cmd_readMifare,
    cn56_cmd_Cext,
    cn56_cmd_setIP,
    cn56_cmd_reset,
    cn56_cmd_storeKey,
    cn56_cmd_readBlk,
    cn56_cmd_writeBlk,
    cn56_cmd_modifKey,
    cn56_cmd_pn_inListPassive,
    cn56_cmd_pn_selectTarget,
    cn56_cmd_loadKey,
    cn56_cmd_authenticate
}cn56_command_t;

/**
 * Option - Enum which List of different kind of error
 */
typedef enum cn56_error_e {
    CN56_ERROR_DISCONNECTED=1,
    CN56_ERROR_ALREADY_CONNECTED,
    CN56_ERROR_NULL_ARGS,
    CN56_ERROR_SOCKET,
    CN56_ERROR_EPOLL,
    CN56_ERROR_CREATION_THREAD,
    CN56_ERROR_CONNECT,
    CN56_ERROR_WRITE,
    CN56_ERROR_READ,
    CN56_ERROR_CLOSED_SOCKET,
    CN56_ERROR_NOT_SUCCESS,
    CN56_ERROR_BAD_SEQ,
    CN56_ERROR_WRONG_ARGS,
    CN56_ERROR_NO_DATA,
    CN56_ERROR_WRONG_BLK,
    CN56_ERROR_TIMEOUT,
    CN56_ERROR_AUTHENTICATE
}cn56_error_t;

/**
 * Action - Enum which List of different kind of action
 */
typedef enum {
    CN56_ACTION_UNKNOWN=0,
    CN56_ACTION_LED,
    CN56_ACTION_READ,
    CN56_ACTION_WRITE,
    CN56_ACTION_RESET,
    CN56_ACTION_AUTHENTICATION,
    CN56_ACTION_TRAILER,
    CN56_ACTION_EXT
} cn56_action_e;

/**
 * State - Enum which List of different kind of state
 */
typedef enum {
    CN56_STATE_CONNECTED=0,
    CN56_STATE_DISCONNECTED,
    CN56_STATE_BAD
} cn56_state_e;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            STRUCT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

typedef struct cn56_handlers_s cn56_handlers_t;
typedef struct cn56_thread_s cn56_thread_t;

typedef struct {
    const char *uid;
    int idx;    // idx 0 = KeyA // idx 1 == KeyB
    uint8_t *value;
    uint8_t len;
} cn56_key_t;


typedef struct {
    bool isAuthenticate;
    uint8_t sec;
    cn56_key_t * key;
    uint8_t uuid[4];
} cn56_authenticate_t;

/**
 * cn56_reader_t - Hold all cn56 reader information
 */
typedef struct {
    const char *uid;
    struct sockaddr_in srvaddr;
    int sockfd;
    uint16_t port;
    uint16_t port_cli;
    uint8_t seq;
    cn56_state_e state;
    cn56_action_e action;
    pthread_t thread_id;
    cn56_handlers_t *functions_handler;
    int epoll_fd;
    int cmd_epoll_fd;
    uint8_t *buffer;
    const char *error;
    pthread_mutex_t mutex;
    cn56_authenticate_t authenticate;
    void *user_data_connection;
    void *user_data_response;
    void *user_data_bad_authent;
} cn56_reader_t;

/**
 * cn56_thread_t - Hold data for thread
 */
typedef struct cn56_thread_s {
    pthread_t threadID;
    cn56_reader_t *readerCtx;
    int epoll_fd;
} cn56_thread_t;

/**
 * cn56_frame_t
 * 
 *  Represent the frame to be send  :
 *  ------------------------------------------------------------------------ 
 *  | STX | SEQ | DADD | CMD | DATA_LENGTH | TIME | DATA[0..n] | BCC | ETX |
 *  ------------------------------------------------------------------------
 * 
 *  Or the response :
 *  --------------------------------------------------------------------
 *  | STX | SEQ | DADD | DATA_LENGTH | STATUS | DATA[0..n] | BCC | ETX |
 *  --------------------------------------------------------------------
 */
typedef struct
{
    int size;
    uint8_t buffer[MAX_CN56_BUFFER_LENGTH];
}cn56_frame_t;

/**
 * cn56_config_wiegand_t - struct define all field of configuration command
 */
typedef struct
{
    uint8_t block_to_read;
    uint8_t csn_reader;
    uint8_t beep_led_enable;
    uint8_t led_buzzer_controle;
    uint8_t reader_enable;
    uint8_t request_idle;
    uint8_t uart_out_enable;
    uint8_t open_wiegand;
    uint8_t set_wiegand;
} cn56_config_wiegand_t;

/**
 * cn56_config_delay_t - struct define the delay to put for some command:
 *          - Activate C_Ext;
 */
typedef struct
{
    uint8_t delay;
}cn56_config_delay_t;

/**
 * cn56_config_write_t - struct define the Data to write on MiFare Card
 *  example: to write "feed beef"
 *      data = [fe eb de ef]
 */
typedef struct
{
    uint8_t blk;
    uint8_t uid[4];
    uint8_t data[CN56_MAX_WRITE];
}cn56_config_write_t;

/**
 * cn56_config_ip_t - struct define the address IP in hexa for command setIp
 *  example: 192.168.1.100
 *      ip_1 = 0xc0 (192)
 *      ip_2 = 0xa8 (168)
 *      ip_3 = 0x01 (1)
 *      ip_4 = 0x64 (100)
 */
typedef struct
{
    uint8_t ip_1;
    uint8_t ip_2;
    uint8_t ip_3;
    uint8_t ip_4;
}cn56_config_ip_t;

typedef struct cn56_config_read_s {
    uint8_t blk;
    uint8_t uid[4]; // uid of the badge
}cn56_config_read_t;

typedef struct cn56_config_load_key_s {
    uint8_t key;
    uint8_t sector;
    uint8_t value[6];
}cn56_config_load_key_t;

typedef struct cn56_config_authenticate_s {
    uint8_t key; // Should be 0x60 for keyA // 0x61 for keyB
    uint8_t uid[4]; // uid of the badge
    uint8_t sec;
}cn56_config_authenticate_t;

typedef struct cn56_config_modif_key_s {
    uint8_t sector;
    uint8_t keyA[6];
    uint8_t acls[4];
    uint8_t keyB[6];
}cn56_config_modif_key_t;

typedef struct cn56_config_write_block_s {
    uint8_t blk;
    uint8_t data[CN56_MAX_WRITE];
}cn56_config_write_block_t;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            HANDLER TYPEDEF
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * cn56_connection_handler_t - handler function call when reade is connected
 * 
 * @param state State can be CN56_STATE_CONNECTED or CN56_STATE_DISCONNECTED
 * @param user_data pointer to the user data
 */
typedef void (*cn56_connection_handler_t) (cn56_state_e state, void * user_data);

/**
 * cn56_response_handler_t - handler function call when data has been read
 * 
 * @param uuid The Unique ID of the reader
 * @param cmd_uid Uid of the command
 * @param response Buffer containing response
 * @param response_len Length of the response
 * @param user_data pointer to the user data
 */
typedef void (*cn56_response_handler_t) (uint8_t *uuid, const char * cmd_uid, uint8_t * response, int response_len, void * user_data);

/**
 * cn56_bad_authent_handler_t - handler function call when badge hasn't a good access key
 * 
 * @param uuid The Unique ID of the reader
 * @param user_data pointer to the user data
 */
typedef void (*cn56_bad_authent_handler_t) (uint8_t *uuid, void * user_data);

struct cn56_handlers_s {
    cn56_connection_handler_t connection_handle;
    cn56_response_handler_t response_handle;
    cn56_bad_authent_handler_t bad_authent_handle;
};

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PUBLIC FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * cn56_init - Initiate the cn56 reader object
 * 
 * @param pointer to allocated reader object
 * @param reader_addr Address of the reader (port always has to be 2000)
 * @param cli_port Port from where th TCP connection has to start
 * @param uid Unique Id of the reader
 * 
 * @return 0 if success, <0 otherwise
 */
int cn56_init(cn56_reader_t *reader, const char *reader_addr, uint16_t cli_port, const char *uid);

/**
 * cn56_free - Free all allocates data into reader
 * 
 * \param reader Reader context
 * 
 * \return 0 in success, <0 otherwise
*/
int cn56_free(cn56_reader_t *reader);

/**
 * cn56_add_connection_handler - Add a handler for the Connection process
 * 
 * @param reader Reader context
 * @param handler connection handler
 * @param user_data pointer to user data that will be accessible in handler function
 * 
 * @return 0 in success, <0 otherwise
*/
int cn56_add_connection_handler(cn56_reader_t *reader, cn56_connection_handler_t handler, void *user_data);

/**
 * cn56_add_response_handler - Add a handler for the Response process
 * 
 * @param reader Reader context
 * @param handler response handler
 * @param user_data pointer to user data that will be accessible in handler function
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_add_response_handler(cn56_reader_t *reader, cn56_response_handler_t handler, void *user_data);

/**
 * cn56_add_bad_authent_handler - Add a handler for the Bad authent process
 * 
 * @param reader Reader context
 * @param handler response handler
 * @param user_data pointer to user data that will be accessible in handler function
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_add_bad_authent_handler(cn56_reader_t *reader, cn56_bad_authent_handler_t handler, void *user_data);


/**
 * cn56_connect_tcp - Start tcp connection
 * 
 * @param reader Reader context
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_connect_async_tcp(cn56_reader_t *reader);

/** cn56_get_uuid_badge - Get the uuid of a badge, if no badge get the error CN56_ERROR_NOT_SUCCESS
 * 
 * @param reader to the reader context
 * @param uuid the uuid allocated, need to be free in success
 * 
 * @return 0 in success, <0 in error
 */
int cn56_get_uuid_badge(cn56_reader_t * reader, uint8_t **uuid);

/**
 * cn56_disconnect_tcp - Stop tcp connection
 * 
 * @param reader Reader context
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_disconnect_tcp(cn56_reader_t *reader);

/**
 * cn56_green_led - Manage green LED
 * 
 * @param reader Reader context
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_green_led(cn56_reader_t *reader);

/**
 * cn56_red_led - Manage red LED
 * 
 * @param reader Reader context
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_red_led(cn56_reader_t *reader);

/**
 * cn56_no_led - Turn off LED
 * 
 * @param reader Reader context
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_no_led(cn56_reader_t *reader);

/**
 * cn56_activate_ext - Activate the external command (open door)
 * 
 * @param reader Reader context
 * @return 0 in success, <0 otherwise
 *
 */
int cn56_activate_ext(cn56_reader_t *reader);

/**
 * cn56_read_mifare - Authenticate & Read Mifare Badge
 * 
 * @param reader Poiter to Reader context
 * @param key Key to use for read, if no-one, use the default key
 * @param sec Sector where to read
 * @param blk block to read
 * @param len Length of data to read
 * @param response Response returned by the reader
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_read_mifare(cn56_reader_t *reader, const cn56_key_t *key, uint8_t sec, uint8_t blk, uint8_t len, uint8_t *response);

/**
 * cn56_write_mifare - Authenticate & Write Mifare Badge
 * 
 * @param reader Reader context
 * @param key Key to use for read, if no-one, use the default key
 * @param sec Sector where to write
 * @param blk block to write
 * @param len Data length to write
 * @param buffer Data to write
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_write_mifare(cn56_reader_t *reader, const cn56_key_t *key, uint8_t sec, uint8_t blk, uint8_t len, uint8_t *buffer);

/**
 * cn56_store_key - Store Key to EE
 * 
 * @param reader Reader context
 * @param key: 0x60 = Key A / 0x61 = KeyB
 * @param sector: from 0 to 15
 * @param value: Value of the key, must be a array of 6 uint8_t
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_store_key(cn56_reader_t * reader, uint8_t key, uint8_t sector, uint8_t *value);

/**
 * cn56_set_config - Set the Wiegand configuration
 * 
 * @param reader Reader context
 * @param config: Wiegand Configuration
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_set_config(cn56_reader_t * reader, cn56_config_wiegand_t config);

/**
 * cn56_trailer - Modify Key
 * 
 * @param reader Pointer to reader context
 * @param sec Sector where modify keys
 * @param key The key to use to updates keys
 * @param trailer trailer object containing information about new keys
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_trailer(cn56_reader_t * reader, uint8_t sec, const cn56_key_t *key, cn56_config_modif_key_t config_modif_key);
// int cn56_trailer(cn56_reader_t * reader, uint8_t sec, const cn56_key_t *key, cn56_trailer_t *trailer);

/**
 * cn56_reset - Reset the badge reader
 *  It will also call the disconnect and free function.
 *  You don't need to disconnect.
 * 
 * @param reader Reader context
 * 
 * @return 0 in success, <0 otherwise
 */
int cn56_reset(cn56_reader_t *reader);

/**
 * cn56_error_msg - return a string of the reader error
 * 
 * @param reader Reader ot print the error
 * @return the reader's error in string
 */
const char *cn56_error_msg(cn56_reader_t *reader);
