/*
 * Copyright (C) 2020 - 2021 "IoT.bzh"
 *
 * Author: Valentin Lefebvre <valentin.lefebvre@iot.bzh>
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#define _GNU_SOURCE

#define MAX_READ_BLOCK_VALUE 16

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "badge-reader-binding.h"
#include "cn56-core.h"
#include <ctl-plugin.h>
#include <errno.h>

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PLUGIN REGISTRATION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

CTLP_CAPI_REGISTER("cn56-reader");
//extern int test_cb(afb_api_t api, const char *reader_ui);

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            STRUCT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * br_cn56_args_cmd_t - Structure of the command arguments from request
 * @param key Uid of the key to use
 * @param sec N° sector
 * @param blk N° block
 * @param dlen Data length
 * @param data Data
 */
typedef struct {
    reader_key_t *key;
    uint8_t sec;
    uint8_t blk;
    ulong dlen;
    uint8_t *data;
} br_cn56_args_cmd_t;

/**
 * br_cn56_ctx_t - Structure of the reader plugin context
 * @param cn56_reader CN56 Reader information
 * @param br_reader Binding reader information
 * @param name Reader's name
 * @param config Config parsed
 */
typedef struct
{
    cn56_reader_t * cn56_reader;
    reader_t * br_reader;
    char *reader_name;
    char * info;
} br_cn56_ctx_t;

/**
 * data_frame_t - Structure of the data frame
 * @param buffer data buffer
 * @param size size of the buffer
 */
typedef struct
{
    uint8_t *buffer;
    int size;
} data_frame_t;

/**
 * br_cn56_status_e - list of the status
 */
typedef enum
{
    BR_CN56_NO_STATUS = 0,
    BR_CN56_STATUS_WRITE,
    BR_CN56_STATUS_READ,
    BR_CN56_STATUS_AUTOREAD,
    BR_CN56_STATUS_ERROR
} br_cn56_status_e;

/**
 * br_cn56_error_e - List of errors
 */
typedef enum
{
    BR_CN56_ERROR_NOT_SUCCESS = 100,
    BR_CN56_ERROR_INIT,
    BR_CN56_ERROR_CONFIG,
    BR_CN56_ERR_BAD_SEQ
} br_cn56_error_e;

/**
 * br_cn56_error_e - List of warning
 */
typedef enum
{
    BR_CN56_WARNING_NO_BADGE = 10
} br_cn56_warning_e;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            CALLBACK FUNCTION 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * _br_cn56_stop_led_cb - send turnoff led command
 * 
 * @param signal signal of the scheduler called
 * @param ctx Context containing reader data
 */
static void _br_cn56_stop_led_cb(int signal, void *ctx) {
    if(!ctx)
        return;
    if (signal)
        return;
    int ret = 0;
    reader_t *reader = (reader_t *)ctx;
    cn56_reader_t *cn56_reader = ((br_cn56_ctx_t *) reader->ctx)->cn56_reader;

    ret = cn56_no_led(cn56_reader);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "[_br_cn56_stop_led_cb] Failed to turn off led for reader %s/%s", reader->fReader->uid, reader->uid);
    }
    reader->action = READER_ACTION_UNKNOWN;
    return;
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE FUNCTION 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * _connection_handler - Called when reader cn56 is connected through TCP
 * 
 * @param state State of the connection, can be CN56_STATE_CONNECTED / CN56_STATE_DISCONNECTED / CN56_STATE_BAD
 * @param user_data Context of the binding reader
 */
static void _connection_handler(cn56_state_e state, void * user_data) {
    reader_t * reader_ctx = (reader_t *) user_data;
    cn56_reader_t * cn56_reader = ((br_cn56_ctx_t *)reader_ctx->ctx)->cn56_reader;
    int ret = 0;

    if (state == CN56_STATE_CONNECTED) {
        AFB_API_NOTICE(reader_ctx->api, "[%s] Connected (_connection_handler)", cn56_reader->uid);
        reader_ctx->state = READER_STATE_CONNECTED;
        cn56_reader->state = CN56_STATE_CONNECTED;

        // Turn off led
        ret = cn56_no_led(cn56_reader);
        if (ret < 0) {
            AFB_API_DEBUG(reader_ctx->api, "[_connection_handler] Failed to turn off led for reader %s/%s, LED should already be off", reader_ctx->fReader->uid, reader_ctx->uid);
        }

        // Set the configuration
        cn56_config_wiegand_t config = {0};
        ret = wrap_json_unpack(reader_ctx->specificConfig, "{si si si si si si si si si}"
                    , "blockToRead", config.block_to_read
                    , "csnRead", config.csn_reader
                    , "beepLedEnable", config.beep_led_enable
                    , "ledBuzzerControle", config.led_buzzer_controle
                    , "readerEnable", config.reader_enable
                    , "requestIdle", config.request_idle
                    , "uartOutEnable", config.uart_out_enable
                    , "openWiegand", config.open_wiegand
                    , "setWiegand", config.set_wiegand);
        if (ret < 0) {
            AFB_API_ERROR(reader_ctx->api, "[_connection_handler] Failed to unpack config for reader %s/%s", reader_ctx->fReader->uid, reader_ctx->uid);
        }
        reader_ctx->specificConfig = json_object_get(reader_ctx->specificConfig);
        ret = cn56_set_config(cn56_reader, config);
        if (ret < 0) {
            AFB_API_DEBUG(reader_ctx->api, "[_connection_handler] Failed to set config for reader %s/%s", reader_ctx->fReader->uid, reader_ctx->uid);
        }
    } else {
        reader_ctx->state = READER_STATE_DISCONNECTED;
        cn56_reader->state = CN56_STATE_DISCONNECTED;
        AFB_API_NOTICE(reader_ctx->api, "[%s] Connection can't be established ... (_connection_handler)", cn56_reader->uid);

        // AFB_API_NOTICE(reader_ctx->api, "[%s] Reconnection ... (_connection_handler)", cn56_reader->uid);
        // ret = cn56_connect_async_tcp(cn56_reader);
        // if (ret < 0) {
        //     if (ret != -READER_ERROR_ALREADY_CONNECTED) {
        //         AFB_API_ERROR(reader_ctx->api, "An error occurred during the connection: %d\n", ret);
        //     } else {
        //         AFB_API_ERROR(reader_ctx->api, "Try to reconnect an already connected reader %s\n", cn56_reader->uid);
        //     }
        // }

    }
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PLUGIN FUNCTION POINTER 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * br_cn56_connect_ptr - Connect to the CN56 reader
 * 
 * @param reader pointer to reader data
 * @return 0 in success, negative int in error
 */
static int br_cn56_connect_ptr(reader_t *reader) {
    int ret = 0;

    // Do not try to connect if an instance is already in process
    if (reader->state == READER_STATE_CONNECTING || reader->state == READER_STATE_CONNECTED) return 0;

    // Catch reader context
    cn56_reader_t * cn56_reader = ((br_cn56_ctx_t *) reader->ctx)->cn56_reader;

    // Try to connect to the reader
    AFB_API_NOTICE(reader->api, "[%s] Connection ... (br_cn56_connect_ptr)", cn56_reader->uid);
    reader->state = READER_STATE_CONNECTING;
    ret = cn56_connect_async_tcp(cn56_reader);
    if (ret < 0) {
        if (ret != -READER_ERROR_ALREADY_CONNECTED) {
            AFB_API_ERROR(reader->api, "An error occurred during the connection: %d\n", ret);
        }
    }

    return ret;
}

/**
 * br_cn56_init_ptr - initiate reader data context and start timer
 * 
 * @param reader pointer to reader data
 * @param argsJ JSON args to config reader
 * @return 0 in success, negative int in error
 */
static int br_cn56_init_ptr(reader_t *reader, json_object *argsJ) {
    int ret = 0;
    cn56_reader_t *cn56_reader  = (cn56_reader_t *) calloc(1, sizeof(cn56_reader_t));
    br_cn56_ctx_t *cn56_ctx     = (br_cn56_ctx_t *) calloc(1, sizeof(br_cn56_ctx_t));
    //cn56_config_t *config       = (cn56_config_t *) calloc(1, sizeof(cn56_config_t));
    char * tcpAddress = NULL;
    int tcpPort = 0;

    // Parse Config in argsJ
    ret = wrap_json_unpack(argsJ, "{s?s s?s s?s s?s s?i}"
        , "uid", &cn56_reader->uid
        , "info", &cn56_ctx->info
        , "reader", &cn56_ctx->reader_name
        , "tcp_address", &tcpAddress
        , "tcp_port", &tcpPort
    );
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "[br_cn56_init_ptr] Unsupported arguments in JSON");
        ret = -BR_CN56_ERROR_CONFIG;
        goto OnErrorExit;
    }
    if (!cn56_reader->uid) cn56_reader->uid = cn56_ctx->reader_name;

    // Initiate the reader context
    ret = cn56_init(cn56_reader, tcpAddress, (uint16_t) tcpPort, cn56_reader->uid);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "An error occurred during the cn56 context initialization");
        ret = -BR_CN56_ERROR_INIT;
        goto OnErrorExit;
    }
    // Add some handlers
    ret = cn56_add_connection_handler(cn56_reader, _connection_handler, (void *)reader);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "An error occurred for adding the connection handler: %d\n", ret);
       return ret;
    }
    // ret = cn56_add_response_handler(cn56_reader, _response_handler, (void *)reader);
    // if (ret < 0) {
    //     AFB_API_ERROR(reader->api, "An error occurred for adding the response handler: %d\n", ret);
    //    return ret;
    // }
    // ret = cn56_add_bad_authent_handler(cn56_reader, _bad_authent_handler, (void *)reader);
    // if (ret < 0) {
    //     AFB_API_ERROR(reader->api, "An error occurred for adding the response handler: %d\n", ret);
    //    return ret;
    // }

    // Adding the cn56_reader and config to the context
    cn56_ctx->cn56_reader = cn56_reader;
    cn56_ctx->br_reader = reader;
    cn56_ctx->br_reader->state = READER_STATE_DISCONNECTED;
    cn56_ctx->br_reader->action = READER_ACTION_UNKNOWN;

    // Adding the context to the binding data
    reader->ctx = (void *)cn56_ctx;

    goto OnExit; 

OnErrorExit:
    // free(config);
    free(cn56_ctx);
    free(cn56_reader);

OnExit:
    return ret;
}

static int br_catch_uuid_ptr(reader_t *reader, uint8_t **uuid) {
    if (reader->state != READER_STATE_CONNECTED) {
        return -READER_ERROR_DISCONNECTED;
    }
    cn56_reader_t *cn56_reader = ((br_cn56_ctx_t *) reader->ctx)->cn56_reader;
    return cn56_get_uuid_badge(cn56_reader, uuid);
}

/**
 * br_cn56_read_ptr - send Read command 
 * 
 * @param reader pointer to reader data
 * @param inputJ JSON input for configure read command
 * @param outputJ JSON respons of the read command
 * @return 0 in success, negative int in error
 */
static int br_cn56_read_ptr(reader_t *reader, const reader_key_t *key, uint8_t sec, uint8_t blk, uint8_t len, uint8_t *response) {
    int ret = 0;
    //uint8_t *response = NULL;
    if (reader->state != READER_STATE_CONNECTED) {
        ret = -READER_ERROR_DISCONNECTED;
        goto OnExit;
    }
    
    cn56_reader_t *cn56_reader = ((br_cn56_ctx_t *) reader->ctx)->cn56_reader;
    
    // Send read command
    cn56_reader->authenticate.isAuthenticate = false;
    //response = (uint8_t *) calloc(len, sizeof(uint8_t));
    ret = cn56_read_mifare(cn56_reader, (const cn56_key_t *) key, sec, blk, len, response);
    if (ret < 0 ){
        AFB_API_ERROR(reader->api, "[br_cn56_read_ptr] Error occurred when sending read command");
        //free(response);
        goto OnExit;
    }
    AFB_API_DEBUG(reader->api, "ReadBadge cmd send successfully to reader=%s", reader->uid);

OnExit:
    //free (args_cmd.data);
    return ret;
}

/**
 * br_cn56_write_ptr - send Write command 
 * 
 * @param reader pointer to reader data
 * @param intputJ JSON input to config the send command
 * @return 0 in success, negative int in error
 */
static int br_cn56_write_ptr(reader_t *reader, const reader_key_t *key, uint8_t sec, uint8_t blk, uint8_t len, uint8_t *buffer) {
    int ret = 0;
    //uint8_t *response = NULL;
    if (reader->state != READER_STATE_CONNECTED) {
        return -READER_ERROR_DISCONNECTED;
    }
    cn56_reader_t *cn56_reader = ((br_cn56_ctx_t *) reader->ctx)->cn56_reader;
    
    // Send write command
    cn56_reader->authenticate.isAuthenticate = false;
    ret = cn56_write_mifare(cn56_reader, (const cn56_key_t *) key, sec, blk, len, buffer);
    if (ret < 0 ){
        AFB_API_ERROR(reader->api, "[br_cn56_write_ptr] Error occurred when sending write command : %i", ret);
        goto OnExit;
    }
    AFB_API_DEBUG(reader->api, "[br_cn56_write_ptr] Command write send successfully to reader=%s", reader->uid);

OnExit:
    return ret;
}

/**
 * br_cn56_activate_ext_ptr - Send the Activate CommandExt and manage LED
 * 
 * @param reader pointer to reader data
 * @return 0 in success, negative int in error
 */
static int br_cn56_activate_ext_ptr(reader_t *reader) {
    int ret = 0;
    if (reader->state != READER_STATE_CONNECTED) {
        ret = -READER_ERROR_DISCONNECTED;
        goto OnExit;
    }
    cn56_reader_t *cn56_reader = ((br_cn56_ctx_t *) reader->ctx)->cn56_reader;

    // Activate Green LED
    ret = cn56_green_led(cn56_reader);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "[br_cn56_activate_ext_ptr] Failed to send green led command to reader %s/%s",
            reader->fReader->uid,
            reader->uid);
        goto OnExit;
    }
    reader->action = READER_ACTION_EXT;

    // External Command
    ret = cn56_activate_ext(cn56_reader);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "[br_cn56_activate_ext_ptr] Failed to send external command to reader %s/%s",
            reader->fReader->uid,
            reader->uid);
        goto OnExit;
    }

    // Schedule the end of the process: Turn off led and reactivate autoRead mode
    ret = afb_api_queue_job(reader->api, _br_cn56_stop_led_cb, reader, NULL, -5000);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "[activateExtCB] Failed to schedule top led for reader %s/%s",
            reader->fReader->uid,
            reader->uid);
        goto OnExit;
    }

    AFB_API_DEBUG(reader->api, "[activateExtCB] cmd send successfully to reader=%s", reader->uid);

OnExit:
    return ret;
}

/**
 * br_cn56_trailer_ptr -Se,d trailer command
 * 
 * @param reader pointer to reader data
 * @return 0 in success, negative int in error
 */
static int br_cn56_trailer_ptr(reader_t *reader, const reader_key_t *key, uint8_t sec, reader_trailer_t *trailer) {
    int ret = 0;
    if (reader->state != READER_STATE_CONNECTED) {
        ret = -READER_ERROR_DISCONNECTED;
        goto OnExit;
    }
    cn56_config_modif_key_t config_modif_key = {
        .keyA = {trailer->keyA->value[0], trailer->keyA->value[1], trailer->keyA->value[2], trailer->keyA->value[3], trailer->keyA->value[04], trailer->keyA->value[5]},
        .keyB = {trailer->keyB->value[0], trailer->keyB->value[1], trailer->keyB->value[2], trailer->keyB->value[3], trailer->keyB->value[04], trailer->keyB->value[5]},
        .acls = {trailer->acls[0], trailer->acls[1], trailer->acls[2], trailer->acls[3]},
        .sector = sec,
    };
    cn56_reader_t *cn56_reader = ((br_cn56_ctx_t *) reader->ctx)->cn56_reader;

    ret = cn56_trailer(cn56_reader, sec, (const cn56_key_t *) key, config_modif_key);

OnExit:
    return ret;
}


/**
 * br_cn56_error_ptr - Error process -> blink red led one time
 * 
 * @param reader pointer to reader data
 * @return 0 in success, negative int in error
 */
static int br_cn56_error_ptr(reader_t *reader) {
    int ret = 0;
    ret = cn56_red_led(((br_cn56_ctx_t *) reader->ctx)->cn56_reader);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "[br_cn56_error_ptr] Failed to send red led command");
    } else {
        // Schedule to turn off led an reactivate the autoread if needed
        ret = afb_api_queue_job(reader->api, _br_cn56_stop_led_cb, reader, NULL, -1000);
        //ret = afb_sched_post_job(NULL /*No group*/, 1000 /*delay MS*/, 0 /*No timeout*/, _br_cn56_stop_led_cb /*CB*/, reader /*ctx*/);
        if (ret < 0) {
            AFB_API_ERROR(reader->api, "[br_cn56_error_ptr] Failed to schedule job for reader %s/%s", reader->fReader->uid, reader->uid);
        }
    }
    return ret;
}

/**
 * br_cn56_subscribe_ptr - Start timer to send read command every time
 * 
 * @param reader pointer to reader data
 * @return 0 in success, negative int in error
 */
static int br_cn56_subscribe_ptr(reader_t *reader) {
    int ret = 0;
    //reader->status = BR_CN56_STATUS_AUTOREAD;
    return ret;
}

/**
 * br_cn56_unsubscribe_ptr - Stop timer for sending read command
 * 
 * @param reader pointer to reader data
 * @return 0 in success, negative int in error
 */
static int br_cn56_unsubscribe_ptr(reader_t *reader) {
    int ret = 0;
    //reader->status = BR_CN56_NO_STATUS;
    return ret;
}

/**
 * br_cn56_reset_ptr - Callback of the Reset command - Actually called when the event EPOLLERR is cached
 * 
 * @param reader pointer to reader data
 * @return 0 in success, negative int in error
 */
static int br_cn56_reset_ptr(reader_t *reader) {
    int ret = 0;

    ret = cn56_reset(((br_cn56_ctx_t *) reader->ctx)->cn56_reader);
    if (ret < 0) {
        AFB_API_ERROR(reader->api, "[br_cn56_reset_ptr] Cannot send reset command");
    }
    reader->state = READER_STATE_DISCONNECTED;
    return ret;
}

// functions callbacks
static family_reader_func_ptr_t cn56Readers[] = {
    {.uid = "doors", .info = "Readers use for opening doors", 
        .initPtr = br_cn56_init_ptr,
        .catchUuidPtr = br_catch_uuid_ptr,
        .connectPtr = br_cn56_connect_ptr,
        .readBadgePtr = br_cn56_read_ptr,
        .writeBadgePtr = br_cn56_write_ptr,
        .activateExtPtr = br_cn56_activate_ext_ptr,
        .errorPtr = br_cn56_error_ptr,
        .trailerBadgePtr = br_cn56_trailer_ptr,
        .subscribePtr = br_cn56_subscribe_ptr,
        .unSubscribePtr = br_cn56_unsubscribe_ptr,
        .onResetPtr = br_cn56_reset_ptr},
    {.uid = "saft", .info = "Readers use for activate Wifi"},
    {.uid = NULL} // must be NULL terminated
};

CTLP_ONLOAD(plugin, registryCB) {
    registerCbT callback = (registerCbT)registryCB;
    assert(callback);
    (*callback)(plugin->uid, cn56Readers);
    return 0;
}