/*
 * Copyright (C) 2021 "IoT.bzh"
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "sqldb.h"

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            GLOBAL VARIABLES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

// check with 'sqlite3 /xxx/reader.db .tables'
const char *sqlSchema =  
    "CREATE TABLE id_cards"
    "('Id' INTEGER PRIMARY KEY"
    ",'cardkey' TEXT NOT NULL"
    ",'control' TEXT NOT NULL"
    ",'name' TEXT NOT NULL"
    " ,UNIQUE ('cardkey')"
    ");";
    // end sqlSchema

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PRIVATE FUNCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

static bool _table_exists(sqldbData_t *sqldbData, char * tableName, char **errorMsg) {
    sqlite3_stmt *queryRqt=NULL;
    static char queryPattern[]=
        "SELECT 1 FROM sqlite_master WHERE type='table'"
        " AND name='%s'"
        ";";
    int err, rc;
    char *queryStr;
    int queryLen=asprintf (&queryStr, queryPattern , tableName);
    if (queryLen < 0) {
        sqlite3_finalize (queryRqt);
        return false;
    }

    err = sqlite3_prepare_v3 (sqldbData->dbFd, queryStr, queryLen, 0, &queryRqt, NULL);
    if (err != SQLITE_OK) {
        sqlite3_finalize (queryRqt);
        free(queryStr);
        return -1;
    };

    rc = sqlite3_step(queryRqt);
    if (rc == SQLITE_ROW) {
        sqlite3_finalize (queryRqt);
        free (queryStr);
        return true;
    } else if (rc == SQLITE_DONE) {
    } else {
        asprintf (errorMsg, "Fail to create SQLlite dbfile=%s", sqldbData->dbPath);
    }

    sqlite3_finalize (queryRqt);
    free (queryStr);
    return false;
}

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            PUBLIC FUNCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

sqldbData_t *sqldb_preinit() {
    sqldbData_t *sqldbData;
    sqldbData = (sqldbData_t *)malloc(sizeof(sqldbData_t));
    sqldbData->dbPath = (char*) calloc(254, sizeof(char));
    sqldbData->dbFd = NULL;
    return sqldbData;
}

int sqldb_init(sqldbData_t *sqldbData) {
    return 0;
}

int sqldb_create(sqldbData_t *sqldbData, char **errorMsg) {
    int rc;
    // db is already open
    if (sqldbData->dbFd) return 1;
    
    // check if DB exit;
    rc = sqlite3_open_v2(sqldbData->dbPath, &sqldbData->dbFd, SQLITE_OPEN_READONLY, NULL);

    // if db does not exist create it
    if (rc != SQLITE_OK) {
        rc = sqlite3_open_v2(sqldbData->dbPath, &sqldbData->dbFd, SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE, NULL);
        if (rc != SQLITE_OK) {
            asprintf (errorMsg, "Fail to create SQLlite dbfile=%s", sqldbData->dbPath);
            sqldbData->dbFd = NULL;
            return -1;
        }

        // populate db schema into newly create db
        rc = sqlite3_exec(sqldbData->dbFd, sqlSchema, NULL, 0, errorMsg);
        if (rc != SQLITE_OK) {
            asprintf (errorMsg, "Fail to execute comand SQLlite dbfile=%s", sqldbData->dbPath);
            remove (sqldbData->dbPath);
            sqldbData->dbFd = NULL;
            return -1;
        }
    } else {
        // close readonly DB and reopen RW
        sqlite3_close_v2(sqldbData->dbFd);
        rc = sqlite3_open_v2(sqldbData->dbPath, &sqldbData->dbFd, SQLITE_OPEN_READWRITE, NULL);
        if (rc != SQLITE_OK) {
            asprintf (errorMsg,"Fail to create SQLlite dbfile=%s", sqldbData->dbPath);
            sqldbData->dbFd = NULL;
            return -1;
        }

        // Check if table doesn't exists
        if ( ! _table_exists(sqldbData, "id_cards", errorMsg)) {
            if (!errorMsg && !*errorMsg) {
                // asprintf (errorMsg, "Fail to retreive table SQLlite dbfile=%s", sqldbData->dbPath);
                sqldbData->dbFd = NULL;
                return -1;
            }
            // populate db schema into newly create db
            rc = sqlite3_exec(sqldbData->dbFd, sqlSchema, NULL, 0, errorMsg);
            if (rc != SQLITE_OK) {
            asprintf (errorMsg, "Fail to execute comand SQLlite dbfile=%s", sqldbData->dbPath);
                remove (sqldbData->dbPath);
                sqldbData->dbFd = NULL;
                return -1;
            }
        }
    }
    return 0;
}

int sqldb_add(sqldbData_t *sqldbData, char *key, char *control, char *name, char **errorMsg) {
    sqlite3_stmt *queryRqt=NULL;
    static char queryPattern[]=
        " insert into id_cards(cardkey, control, name)"
        " values('%s', '%s', '%s')"
        ";";
    int err, status;
    char *queryStr;
    int queryLen=asprintf (&queryStr, queryPattern , key, control, name);
    if (queryLen < 0) {
        asprintf(errorMsg, "Failed to create query");
        sqlite3_finalize (queryRqt);
        return -SQLDB_ERROR_QUERY;
    }

    // compile request into byte code
    err= sqlite3_prepare_v3 (sqldbData->dbFd, queryStr, queryLen, 0, &queryRqt, NULL);
    if (err != SQLITE_OK) {
        asprintf(errorMsg, "Failed to prepare query: [SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt);
        free (queryStr);
        return -SQLDB_ERROR_QUERY;
    };

    status= sqlite3_step(queryRqt);
    if (status != SQLITE_DONE) {
        asprintf(errorMsg, "[SQL Error -> %i] %s", status, sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt);
        free (queryStr);
        return -SQLDB_ERROR_ADDING;
    };

    sqlite3_finalize (queryRqt);
    free (queryStr);
    return 0;
}

int sqldb_get(sqldbData_t *sqldbData, char *key, char **idKey, char **control, char **name, char **errorMsg) {
    sqlite3_stmt *queryRqt=NULL;
    static char queryPattern[]=
        " select id, control, name from id_cards where cardkey='%s'"
        ";";
    int err, status;
    char *queryStr;
    int queryLen=asprintf (&queryStr, queryPattern , key);
    if (queryLen < 0) {
        asprintf(errorMsg, "Failed to create query");
        sqlite3_finalize (queryRqt);
        return -SQLDB_ERROR_QUERY;
    }

    // compile request into byte code
    err= sqlite3_prepare_v3 (sqldbData->dbFd, queryStr, queryLen, 0, &queryRqt, NULL);
    if (err) {
        asprintf(errorMsg, "Failed to prepare query: [SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt);
        free (queryStr);
        return -SQLDB_ERROR_QUERY;
    };

    // Apply the request
    status = sqlite3_step(queryRqt);
    if (status == SQLITE_ROW) {
        asprintf(idKey, "%s", (char *) sqlite3_column_text(queryRqt, 0));
        asprintf(control, "%s", (char *) sqlite3_column_text(queryRqt, 1));
        asprintf(name, "%s", (char *) sqlite3_column_text(queryRqt, 2));
        free (queryStr);
        sqlite3_finalize (queryRqt);
        return 0;
    }
    asprintf(errorMsg, "Key is not found");
    free (queryStr);
    sqlite3_finalize (queryRqt);
    return -SQLDB_ERROR_NOT_FOUND;
}


int sqldb_update_control(sqldbData_t *sqldbData, char *key, char *control, char **errorMsg) {
    sqlite3_stmt *queryRqt=NULL;
    static char queryPattern[]=
        " UPDATE id_cards SET control = '%s' WHERE cardkey='%s'"
        ";";
    int err, status;
    char *queryStr;
    int queryLen=asprintf (&queryStr, queryPattern , control, key);
    if (queryLen < 0) {
        asprintf(errorMsg, "Failed to create query");
        sqlite3_finalize (queryRqt);
        return -SQLDB_ERROR_QUERY;
    }

    // compile request into byte code
    err= sqlite3_prepare_v3 (sqldbData->dbFd, queryStr, queryLen, 0, &queryRqt, NULL);
    if (err != SQLITE_OK) {
        asprintf(errorMsg, "Failed to prepare query: [SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt);
        free (queryStr);
        return -SQLDB_ERROR_QUERY;
    };

    status= sqlite3_step(queryRqt);
    if (status != SQLITE_DONE) {
        asprintf(errorMsg, "[SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt);
        free (queryStr);
        return -SQLDB_ERROR_UPDATE;
    };

    sqlite3_finalize (queryRqt);
    free (queryStr);
    return 0;

}

char ** sqldb_list(sqldbData_t *sqldbData, int *length, char **errorMsg) {

    // Get the number of row

    sqlite3_stmt *queryRqt_count=NULL;
    static char queryPattern_count[]=
        " SELECT COUNT(*) FROM id_cards"
        ";";
    int err = 0, status = 0;

    // compile request into byte code
    err = sqlite3_prepare_v3 (sqldbData->dbFd, queryPattern_count, sizeof(queryPattern_count), 0, &queryRqt_count, NULL);
    if (err != SQLITE_OK) {
        asprintf(errorMsg, "Failed to prepare query: [SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt_count);
        return NULL;
    };

    status= sqlite3_step(queryRqt_count);
    
    *length = sqlite3_column_int(queryRqt_count, 0);
    if (status != SQLITE_DONE && status != SQLITE_ROW) {
        asprintf(errorMsg, "[SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt_count);
        return NULL;
    }

    // Get the list

    sqlite3_stmt *queryRqt=NULL;
    static char queryPattern[]=
        " SELECT * FROM id_cards"
        ";";

    // compile request into byte code
    err = sqlite3_prepare_v3 (sqldbData->dbFd, queryPattern, sizeof(queryPattern), 0, &queryRqt, NULL);
    if (err != SQLITE_OK) {
        asprintf(errorMsg, "Failed to prepare query: [SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt);
        return NULL;
    };

    status= sqlite3_step(queryRqt);
    char **list_key = (char **)calloc(*length, sizeof(char*)); 

    int index  = 0;
    while (status == SQLITE_ROW && index < *length) {
        list_key[index] = (char *)calloc(1, sizeof(char*));
        asprintf(&list_key[index], "%s", (char *) sqlite3_column_text(queryRqt, 1));
        status= sqlite3_step(queryRqt);
        index ++;
    }
    if (status != SQLITE_DONE) {
        asprintf(errorMsg, "[SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt);
        for (int idx = 0; idx < index; idx ++) {
            free(list_key[idx]);
        }
        free (*list_key);
        return NULL;
    };

    sqlite3_finalize (queryRqt);
    return list_key;
}

int sqldb_delete(sqldbData_t *sqldbData, char *key, char **errorMsg) {
    sqlite3_stmt *queryRqt=NULL;
    static char queryPattern[]=
        " DELETE FROM id_cards WHERE cardkey='%s'"
        ";";
    int err, status;
    char *queryStr;
    int queryLen=asprintf (&queryStr, queryPattern, key);
    if (queryLen < 0) {
        asprintf(errorMsg, "Failed to create query");
        sqlite3_finalize (queryRqt);
        return -SQLDB_ERROR_QUERY;
    }

    // compile request into byte code
    err = sqlite3_prepare_v3 (sqldbData->dbFd, queryStr, queryLen, 0, &queryRqt, NULL);
    if (err != SQLITE_OK) {
        asprintf(errorMsg, "Failed to prepare query: [SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt);
        // free (queryStr);
        return -SQLDB_ERROR_QUERY;
    };

    status= sqlite3_step(queryRqt);
    if (status != SQLITE_DONE) {
        asprintf(errorMsg, "[SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt);
        // free (queryStr);
        return -SQLDB_ERROR_DELETE;
    };

    sqlite3_finalize (queryRqt);
    // free (queryStr);
    return 0;
}

int sqldb_delete_all(sqldbData_t *sqldbData, char **errorMsg) {
    sqlite3_stmt *queryRqt=NULL;
    static char queryPattern[]=
        " DELETE FROM id_cards"
        ";";
    int err, status;
    // char *queryStr;
    // int queryLen=asprintf (&queryStr, queryPattern);
    // if (queryLen < 0) {
    //     asprintf(errorMsg, "Failed to create query");
    //     sqlite3_finalize (queryRqt);
    //     return -1;
    // }

    // compile request into byte code
    err = sqlite3_prepare_v3 (sqldbData->dbFd, queryPattern, sizeof(queryPattern), 0, &queryRqt, NULL);
    if (err != SQLITE_OK) {
        asprintf(errorMsg, "Failed to prepare query: [SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt);
        // free (queryStr);
        return -SQLDB_ERROR_QUERY;
    };

    status= sqlite3_step(queryRqt);
    if (status != SQLITE_DONE) {
        asprintf(errorMsg, "[SQL Error] %s", sqlite3_errmsg(sqldbData->dbFd));
        sqlite3_finalize (queryRqt);
        // free (queryStr);
        return -SQLDB_ERROR_DELETE;
    };

    sqlite3_finalize (queryRqt);
    // free (queryStr);
    return 0;
}