/*
 * Copyright (C) 2021 "IoT.bzh"
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            DEFINE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
#ifndef SQLDB_H
#define SQLDB_H

#define _GNU_SOURCE
#define MAX_QUERIES         3
#define MAX_UUID_KEY_LENGTH 16
#define SQLDB_GROUP         "SQLBD"

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            INCLUDE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
#include <sqlite3.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            ENUM
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * sqldb_error_t - Enum which List of different kind of error
 */
typedef enum {
    SQLDB_ERROR_NOT_FOUND = 1,
    SQLDB_ERROR_ADDING,
    SQLDB_ERROR_GET,
    SQLDB_ERROR_UPDATE,
    SQLDB_ERROR_DELETE,
    SQLDB_ERROR_COUNT,
    SQLDB_ERROR_QUERY,
    SQLDB_ERROR_NULL_ARGS
}sqldb_error_t;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            STRUCT
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

typedef struct {
    char * dbPath;
    sqlite3 *dbFd;
} sqldbData_t;

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                            FUNCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

/**
 * sqldb_preinit - SQLDB structure preinitialization
 *
 * @return sqldb object preinitialized
 */
sqldbData_t *sqldb_preinit();

/**
 * sqldb_init - SQLDB structure initialization
 *
 * @param sqldbData SQLDB structure to initialize
 * @return 0 if success
 */
int sqldb_init(sqldbData_t *sqldbData);

/**
 * sqldb_create - Create database 
 *
 * @param sqldbData SQLDB data structure
 * @param errorMsg Message of the specific error, NEED TO BE FREE by the user in case of error
 * @return 0 if success, <0 if not
 */
int sqldb_create(sqldbData_t *sqldbData, char **errorMsg);

/**
 * sqldb_add - Add Id of card to database 
 *
 * @param sqldbData SQLDB data structure
 * @param key Id of the key to add
 * @param control Value of the control data (date)
 * @param name Value of the user name
 * @param errorMsg Message of the specific error, NEED TO BE FREE by the user in case of error
 * @return 0 if success, <0 if not
 */
int sqldb_add(sqldbData_t *sqldbData, char *key, char *control, char *name, char **errorMsg);

/**
 * sqldb_get - Get id of a Card in database if exists
 *
 * @param sqldbData SQLDB data structure
 * @param key Id of the key to retrieve
 * @param idKey Id of the key found, NEED TO BE FREE by the user in case of success
 * @param control control column found, NEED TO BE FREE by the user in case of success
 * @param name name column found, NEED TO BE FREE by the user in case of success
 * @param errorMsg Message of the specific error, NEED TO BE FREE by the user in case of error
 * @return 0 if success, <0 if not
 */
int sqldb_get(sqldbData_t *sqldbData, char *key, char **idKey, char **control, char **name, char **errorMsg);

/**
 * sqldb_update_control - Update the control value of a key in database
 *
 * @param sqldbData SQLDB data structure
 * @param key Id of the key to retrieve
 * @param control New value of controle to update
 * @param errorMsg Message of the specific error, NEED TO BE FREE by the user in case of error
 * @return 0 if success, <0 if not
 */
int sqldb_update_control(sqldbData_t *sqldbData, char *key, char *control, char **errorMsg);

/**
 * sqldb_list - Get the list of badges in database
 *
 * @param sqldbData SQLDB data structure
 * @param kist_key return list of kay allocated
 * @param length return length of the list
 * @param errorMsg Message of the specific error, NEED TO BE FREE by the user in case of error
 * @return 0 if success, <0 if not
 */
char ** sqldb_list(sqldbData_t *sqldbData, int *length, char **errorMsg);

/**
 * sqldb_delete - delete row from database
 *
 * @param sqldbData SQLDB data structure
 * @param key Key to remove
 * @param errorMsg Message of the specific error, NEED TO BE FREE by the user in case of error
 * @return 0 if success, <0 if not
 */
int sqldb_delete(sqldbData_t *sqldbData, char *key, char **errorMsg);

/**
 * sqldb_delete_all - delete all data in database
 *
 * @param sqldbData SQLDB data structure
 * @param errorMsg Message of the specific error, NEED TO BE FREE by the user in case of error
 * @return 0 if success, <0 if not
 */
int sqldb_delete_all(sqldbData_t *sqldbData, char **errorMsg);

#endif /* SQLDB_H */